//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    polygon2.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include "polygon2.h"
#include <cmath>

namespace qmpp {

  // -----------------------------------------------------------------
  // Method: tPolygon2()
  // 
  // Default constructor
  //
  // -----------------------------------------------------------------
  tPolygon2::tPolygon2() : m_size(0) {}

  // -----------------------------------------------------------------
  // Method: tPolygon2()
  // 
  // Constructor
  //
  // -----------------------------------------------------------------
  tPolygon2::tPolygon2(int n)
  {
    m_size = 0;
    m_verts.resize(n);
    m_ec.resize(n);
  }

  // -----------------------------------------------------------------
  // Method: tPolygon2()
  // 
  // Constructor
  //
  // -----------------------------------------------------------------
  tPolygon2::tPolygon2(const tPolygon2& q)
  {
    m_size = q.m_size;
    m_verts.resize(m_size);
    m_ec.resize(m_size);

    for (int i=0; i<m_size; i++) {
      m_verts[i] = q.m_verts[i];
      m_ec[i] = q.m_ec[i];
    }
  }

  // -----------------------------------------------------------------
  // Method: operator=()
  // 
  // Overload assignment operator
  //
  // -----------------------------------------------------------------
  tPolygon2&
  tPolygon2::operator=(const tPolygon2& q)
  {
    if (m_verts.size() > 0) m_verts.clear();
 
    m_size = q.m_size;
    m_verts.resize(m_size);
    m_ec.resize(m_size);

    for (int i=0; i<m_size; i++) {
      m_verts[i] = q.m_verts[i];
      m_ec[i] = q.m_ec[i];
    }

    return *this;
  }

  // -----------------------------------------------------------------
  // Method: findReflexPoint()
  // 
  // Returns the index of a reflex vertex in the polygon, if any.
  //
  // -----------------------------------------------------------------
  int
  tPolygon2::findReflexPoint() const throw (tExceptionObject)
  {
    for (int i=0; i<m_size; i++) if (isReflex(i)) return i;

    stringstream ss (stringstream::in | stringstream::out);
    ss << "Polygon has no reflex vertex";
    throw tExceptionObject(__FILE__,__LINE__,ss.str());
  }

  // -----------------------------------------------------------------
  // Method: getIndexOfPoint()
  // 
  // Returns  the index  of  a vertex  in  this polygon  with a  given
  // location.  If  such there is no  such vertex in  this polygon, an
  // exception is thrown.
  //
  // -----------------------------------------------------------------
  int
  tPolygon2::getIndexOfPoint(const Point2& p) const 
    throw (tExceptionObject)
  {
    for (int i=0; i<m_size; i++) {
      if (p == m_verts[i].getPoint()) return i;
    }

    stringstream ss (stringstream::in | stringstream::out);
    ss << "Polygon has no point at the given location";
    throw tExceptionObject(__FILE__,__LINE__,ss.str());
  }

  // -----------------------------------------------------------------
  // Method: getNumOfReflexPoints()
  // 
  // Returns the number of reflex vertices of this polygon.
  //
  // -----------------------------------------------------------------
  int
  tPolygon2::getNumOfReflexPoints() const
  {
    int counter = 0;
    for (int i=0; i<m_size; i++) {
      if (this->isReflex(i)) ++counter;
    }

    return counter;
  }

  // -----------------------------------------------------------------
  // Method: area()
  // 
  // Returns the area of this polygon.
  //
  // -----------------------------------------------------------------
  double
  tPolygon2::area() const
  {
    Polygon2 q;

    for (int i=0; i<m_size; i++) {
      q.push_back(m_verts[i].getPoint());
    }

    return q.area();
  }

  // -----------------------------------------------------------------
  // Method: cosangle()
  // 
  // Returns the cosine of the  angle defined by two given consecutive
  // edges of this polygon.
  //
  // -----------------------------------------------------------------
  double
  tPolygon2::cosangle(int i) const throw (tExceptionObject)
  {
    if (i >= m_size) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Out-of-range index of polygon vertex";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    Point2 p1;
    Point2 p2;
    Point2 p3;

    if (i == 0) {
      p1 = m_verts[m_size-1].getPoint();
      p2 = m_verts[0].getPoint();
      p3 = m_verts[1].getPoint();
    }
    else {
      p1 = m_verts[(i-1)].getPoint();
      p2 = m_verts[i].getPoint();
      p3 = m_verts[(i+1) % m_size].getPoint();
    }

    double x1 = p2.x() - p1.x();
    double y1 = p2.y() - p1.y();
    double n1 = sqrt((x1 * x1) + (y1 * y1));
    x1 /= n1;
    y1 /= n1;

    double x2 = p3.x() - p2.x();
    double y2 = p3.y() - p2.y();
    double n2 = sqrt((x2 * x2) + (y2 * y2));
    x2 /= n2;
    y2 /= n2;

    return ((x1*x2) + (y1*y2));
  }

  // -----------------------------------------------------------------
  // Method: isCW()
  // 
  // Returns  true if this  polygon is  clockwise oriented,  and false
  // otherwise.
  //
  // -----------------------------------------------------------------
  bool
  tPolygon2::isCW() const
  { return !this->isCCW(); }

  // -----------------------------------------------------------------
  // Method: isCCW()
  // 
  // Returns true  if this  polygon is counterclockwise  oriented, and
  // false otherwise.
  //
  // -----------------------------------------------------------------
  bool
  tPolygon2::isCCW() const
  {
    Polygon2 q;

    for (int i=0; i<m_size; i++) {
      q.push_back(m_verts[i].getPoint());
    }

    return q.is_counterclockwise_oriented();
  }

  // -----------------------------------------------------------------
  // Method: isConvex()
  // 
  // Returns true if this polygon is convex, and false otherwise.
  //
  // -----------------------------------------------------------------
  bool
  tPolygon2::isConvex() const
  {
    for (int i=0; i<m_size; i++) {
      if (isReflex(i)) return false;
    }

    return true;
  }

  // -----------------------------------------------------------------
  // Method: isReflex()
  // 
  // Returns   true  if   the   i-th  vertex   of   this  polygon   is
  // reflex. Otherwise, it returns false.
  //
  // -----------------------------------------------------------------
  bool
  tPolygon2::isReflex(int i) const
  {
    CGAL::Oriented_side right = (this->isCW()) ? 
      CGAL::ON_POSITIVE_SIDE : CGAL::ON_NEGATIVE_SIDE;

    Point2 p1 = m_verts[i].getPoint();
    Point2 p2 = m_verts[(i+1) % m_size].getPoint();
    Point2 p3 = m_verts[(i+m_size-1) % m_size].getPoint();
    Line2 l1(p1,p2);
    Line2 l2(p3,p1);

    // There  is a redundancy  here, but  we do  so to  avoid annoying
    // numerical problems.
    return ((l1.oriented_side(p3) == right) && 
	    (l2.oriented_side(p2) == right));
  }

  // -----------------------------------------------------------------
  // Method: isSimple()
  // 
  // Returns  true if this  polygon is  simple. Otherwise,  it returns
  // false.
  //
  // -----------------------------------------------------------------
  bool
  tPolygon2::isSimple() const
  {
    Polygon2 q;

    for (int i=0; i<m_size; i++) {
      q.push_back(m_verts[i].getPoint());
    }

    return q.is_simple();
  }

  // -----------------------------------------------------------------
  // Method: reverse()
  // 
  // Reverses the orientation of this polygon.
  //
  // -----------------------------------------------------------------
  void
  tPolygon2::reverse()
  {
    tIndVertex2 aux1;
    bool aux2;
    int size2 = m_size >> 1;

    for (int i=0; i<size2; i++) {
      aux1  = m_verts[i];
      m_verts[i] = m_verts[m_size-i-1];
      m_verts[m_size-i-1] = aux1;

      aux2  = m_ec[i];
      m_ec[i] = m_ec[m_size-i-1];
      m_ec[m_size-i-1] = aux2;
    }
  }

  // -----------------------------------------------------------------
  // Method: insert()
  // 
  // Inserts a new vertex into the list of vertices of this polygon.
  //
  // -----------------------------------------------------------------
  void tPolygon2::insert(int i, int id, const Point2& p)
  {
    m_verts[i].setVertexId(id);
    m_verts[i].setPoint(p);
    ++m_size;
  }

  // -----------------------------------------------------------------
  // Method: insert()
  // 
  // Inserts a new vertex into the list of vertices of this polygon.
  //
  // -----------------------------------------------------------------
  void 
  tPolygon2::insert(int i, const tIndVertex2& v)
  {
    m_verts[i] = v;
    ++m_size;
  }

  // -----------------------------------------------------------------
  // Method: shiftLeft()
  // 
  // Shifts the list  of vertices of this polygon  by "d" positions to
  // the left.  So,  the first vertex of the list  will be the current
  // (d+1)-th vertex.
  //
  // -----------------------------------------------------------------
  void 
  tPolygon2::shiftLeft(int d)
  {
    for (int i=0; i<d; i++) {
      tIndVertex2 aux1 = m_verts[0];
      bool aux2 = m_ec[0];
      for (int j=0; j<m_size-1; j++) {
	m_verts[j] = m_verts[j+1];
	m_ec[j] = m_ec[j+1];
      }

      m_verts[m_size-1] = aux1;
      m_ec[m_size-1] = aux2;
    }
  }

  // -----------------------------------------------------------------
  // Method: shiftRight()
  // 
  // Shifts the list  of vertices of this polygon  by "d" positions to
  // the right.  So, the first vertex  of the list will be the current
  // (n-d)-th  vertex, where  "n" is  the number  of vertices  in this
  // polygon.
  //
  // -----------------------------------------------------------------
  void 
  tPolygon2::shiftRight(int d)
  {
    for (int i=0; i<d; i++) {
      tIndVertex2 aux1 = m_verts[m_size-1];
      bool aux2 = m_ec[m_size-1];
      for (int j=m_size-1; j>0; j--) {
	m_verts[j] = m_verts[j-1];
	m_ec[j] = m_ec[j-1];
      }
      m_verts[0] = aux1;
      m_ec[0] = aux2;
    }
  }


};   // end namespace qmpp
