//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    complex2.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include "complex2.h"
#include <sstream>

namespace hedge {
  
  // -----------------------------------------------------------------
  // Method: tComplex2() 
  // 
  // Default constructor
  //
  // -----------------------------------------------------------------
  tComplex2::tComplex2() 
  {
    m_dual = tDualGraph::New();
    m_dual->setDual(Pointer(this));
  }

  // -----------------------------------------------------------------
  // Method: ~tComplex2() 
  // 
  // Destructor
  //
  // -----------------------------------------------------------------
  tComplex2::~tComplex2() 
  {
    while (!m_lverts.empty()) {
      vtPointer vaux = m_lverts.front();
      vaux->setAttribute(vaPointer(0));
      vaux->setHalfEdge(hePointer(0));
      vaux->setBoundHalfEdge(hePointer(0));
      m_ledges.pop_front();
    }

    while (!m_ledges.empty()) {
      egPointer eaux = m_ledges.front();
      hePointer he1 = eaux->getHalfEdge();
      hePointer he2 = eaux->getMate(he1);

      if (he1.isNotNull()) {
	he1->setVertex(vtPointer(0));
	he1->setEdge(egPointer(0));
	he1->setFace(fcPointer(0));
	he1->setNext(hePointer(0));
	he1->setPrev(hePointer(0));
      }
      
      if (he2.isNotNull()) {
	he2->setVertex(vtPointer(0));
	he2->setEdge(egPointer(0));
	he2->setFace(fcPointer(0));
	he2->setNext(hePointer(0));
	he2->setPrev(hePointer(0));
      }

      eaux->setHalfEdge1(hePointer(0));
      eaux->setHalfEdge2(hePointer(0));
      eaux->setBoundary(bdPointer(0));
      eaux->setAttribute(eaPointer(0));
      m_ledges.pop_front();
    }

    while (!m_lfaces.empty()) {
      fcPointer faux = m_lfaces.front();
      faux->setAttribute(faPointer(0));
      faux->setHalfEdge(hePointer(0));
      faux->setComplex(Pointer(0));
      faux->setDualVertex(dvPointer(0));
      m_lfaces.pop_front();
    }

    while (m_lbounds.empty()) {
      bdPointer baux = m_lbounds.front();
      baux->setBoundaryEdge(egPointer(0));
      m_lbounds.pop_front();
    }

    m_dual = dgPointer(0);
  }

  // -----------------------------------------------------------------
  // Method: createVertex() 
  // 
  // Creates a new vertex and adds it to the cell complex
  //
  // -----------------------------------------------------------------
  tComplex2::vtPointer
  tComplex2::createVertex(const Point2& p) 
  {
    // Creates a new vertex
    vtPointer newv = tVertex2::New();
    newv->setPoint(p);

    // Insert it into list of vertices of this complex
    m_lverts.push_back(newv);

    // Verifies if this is the first vertex
    if (m_lverts.size() == 1) {
      // Initializes the bounding box of this cell complex
      m_bb = tBBox2::New();
      m_bb->setPMin(p);
      m_bb->setPMax(p);
    }
    else {
      // Update bounding box
      Point2 p1 = m_bb->getPMin();
      Point2 p2 = m_bb->getPMax();

      double xmin = p1.x();
      double ymin = p1.y();
      double xmax = p2.x();
      double ymax = p2.y();

      if (xmin > p.x()) xmin = p.x();
      if (ymin > p.y()) ymin = p.y();
      if (xmax < p.x()) xmax = p.x();
      if (ymax < p.y()) ymax = p.y();

      m_bb->setPMin(Point2(xmin,ymin));
      m_bb->setPMax(Point2(xmax,ymax));
    }

    return newv;
  }

  // -----------------------------------------------------------------
  // Method: createDualVertex()
  // 
  // Creates a new dual vertex and inserts it into the dual graph
  //
  // -----------------------------------------------------------------
  void
  tComplex2::createDualVertex(const fcPointer& f) 
  {
    // Create a new dual vertex
    dvPointer newv = tDualGraphVertex::New();

    // Initialize the values of the new dual vertex
    newv->setFace(f);
    newv->setOwner(m_dual);

    // Make face f point to its dual vertex in the dual graph
    f->setDualVertex(newv);

    // Add vertex to dual graph
    m_dual->addVert(newv);
  }

  // -----------------------------------------------------------------
  // Method: deleteDualVertex()
  // 
  // Removes a vertex from the dual graph
  //
  // -----------------------------------------------------------------
  void
  tComplex2::deleteDualVertex(const fcPointer& f) 
  {
    // Get dual vertex
    dvPointer v = f->getDualVertex();

    // Remove vertex from dual graph
    m_dual->delVert(v);

    // Face "f" is not pointing to any vertex
    f->setDualVertex(dvPointer(0));
  }

  // -----------------------------------------------------------------
  // Method: createDualVertex()
  // 
  // Creates a new edge in the dual graph of this cell complex
  //
  // -----------------------------------------------------------------
  void
  tComplex2::createDualEdge(
			    const fcPointer& f1, 
			    const fcPointer& f2
			    ) 
  {
    // get dual vertex of f1 and f2
    dvPointer v1 = f1->getDualVertex();
    dvPointer v2 = f2->getDualVertex();

    // add edge to dual graph
    m_dual->addEdge(v1,v2);
  }

  // -----------------------------------------------------------------
  // Method: getNextStarVert()
  // 
  // Returns the next half-edge in the star of the initial vertex of a
  // given half-edge.
  //
  // -----------------------------------------------------------------
  tComplex2::hePointer
  tComplex2::getNextStarVert(const hePointer& he)
  {
    hePointer heaux = he->getEdge()->getMate(he);

    if (heaux.isNotNull()) return heaux->getNext();
    return heaux;
  }

  // -----------------------------------------------------------------
  // Method: getPrevStarVert()
  // 
  // Returns the previous half-edge in  the star of the initial vertex
  // of a given half-edge.
  //
  // -----------------------------------------------------------------
  tComplex2::hePointer
  tComplex2::getPrevStarVert(const hePointer& he)
  {
    hePointer heaux = he->getPrev();

    if (heaux.isNotNull()) return heaux->getEdge()->getMate(heaux);
    return heaux;
  }

  // -----------------------------------------------------------------
  // Method: getNextBoundEdge()
  // 
  // Returns the boundary edge following a given boundary edge.
  //
  // -----------------------------------------------------------------
  tComplex2::egPointer
  tComplex2::getNextBoundEdge(const egPointer& e)
  {
    hePointer he1 = e->getHalfEdge()->getNext();
    hePointer he2 = he1->getEdge()->getMate(he1);
    while (he2.isNotNull()) {
      he1 = he2->getNext();
      he2 = he1->getEdge()->getMate(he1); 
    }

    return he1->getEdge();
  }

  // -----------------------------------------------------------------
  // Method: getPrevBoundEdge()
  // 
  // Returns the boundary edge preceding a given boundary edge.
  //
  // -----------------------------------------------------------------
  tComplex2::egPointer
  tComplex2::getPrevBoundEdge(const egPointer& e)
  {
    hePointer he1 = e->getHalfEdge()->getPrev();
    hePointer he2 = he1->getEdge()->getMate(he1);
    while (he2.isNotNull()) {
      he1 = he2->getPrev();
      he2 = he1->getEdge()->getMate(he1); 
    }

    return he1->getEdge();
  }

  // -----------------------------------------------------------------
  // Method: isOnConstrainedEdge()
  // 
  // Returns true if a given vertex belongs to a constrained edge
  //
  // -----------------------------------------------------------------
  bool 
  tComplex2::isOnConstrainedEdge(const vtPointer& v) const
    throw (tExceptionObject)
  {
    if (v->getHalfEdge()->getFace()->getComplex().getPointee() != this) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Query vertex does not belong to this complex";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    hePointer he1 = v->getHalfEdge();
    hePointer he2 = he1;
    do {
      if (he2->getEdge()->isConstrained()) return true;
      he2 = he2->getEdge()->getMate(he2)->getNext();
    } while (he1 != he2);

    return false;
  }

  // -----------------------------------------------------------------
  // Method: getVertexStar()
  // 
  // Computes a polygon corresponding to the star of a vertex.
  //
  // -----------------------------------------------------------------
  vector<Point2>
  tComplex2::getVertexStar(const vtPointer& v) 
    throw (tExceptionObject)
  {
    if (v->getHalfEdge()->getFace()->getComplex().getPointee() != this) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Query vertex does not belong to this complex";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    list<Point2> lpoints;

    hePointer he1 = v->getHalfEdge();
    hePointer he2 = he1;
    do {
      hePointer he3 = he2->getNext();
      while (he3 != he2->getPrev()) {
	lpoints.push_back(he3->getVertex()->getPoint());
	he3 = he3->getNext();
      }
      he2 = he3->getEdge()->getMate(he3);

      if (he2.isNull()) {
	lpoints.push_back(he3->getVertex()->getPoint());
	break;
      }

    } while (he2 != he1);

    if (he1->getEdge()->getMate(he1).isNotNull() && he2.isNull()) {
      he2 = he1->getEdge()->getMate(he1);
      do {
	hePointer he3 = he2->getPrev();
	while (he3 != he2->getNext()) {
	  lpoints.push_front(he3->getVertex()->getPoint());
	  he3 = he3->getPrev();
	}
	he2 = he3->getEdge()->getMate(he3);
      } while (he2.isNotNull());
    }

    vector<Point2> vp(lpoints.size());
    unsigned i = 0;
    while (!lpoints.empty()) {
      vp[i] = lpoints.front();
      lpoints.pop_front();
      ++i;
    }

    return vp;
  }

  // -----------------------------------------------------------------
  // Method mkFBEV
  //
  // Creates  a new  face with  only one  vertex, edge,  and boundary,
  // which is used to create an isolated face.
  //
  // -----------------------------------------------------------------
  tComplex2::fcPointer
  tComplex2::mkFBEV(const vtPointer& v)
  {
    // Creates boundary, face and edge
    bdPointer newbd = tBoundary2::New();
    fcPointer newf = tFace2::New();
    egPointer newe = tEdge2::New();

    // Inserts the new elements into the corresponding lists
    m_lbounds.push_back(newbd);
    m_lfaces.push_back(newf);
    m_ledges.push_back(newe);
  
    // Creates the first half-edge of the new edge
    hePointer newhe = tHalfEdge2::New();
    
    // Initialize first half-edge
    newhe->setVertex(v);
    newhe->setEdge(newe);
    newhe->setFace(newf);
    newhe->setNext(newhe);
    newhe->setPrev(newhe);
  
    // Updates vertex
    v->setHalfEdge(newhe);
    v->setBoundHalfEdge(newhe);

    // Initializes new edge
    newe->setHalfEdge1(newhe);
    newe->setBoundary(newbd);

    // Initializes new face
    newf->setHalfEdge(newhe);
    newf->setComplex(Pointer(this));

    // Initializes new boundary
    newbd->setBoundaryEdge(newe);

    // Creates a corresponding vertex in the dual graph
    createDualVertex(newf);

    // Returns the new face
    return newf;
  }


  // -----------------------------------------------------------------
  // Method mkEV()
  //
  // Adds a  new vertex and  a new edge  to an existent face.  The new
  // edge follows  a given edge of  the face with respect  to the face
  // cycle.
  //
  // -----------------------------------------------------------------
  tComplex2::egPointer
  tComplex2::mkEV(
		  const vtPointer& v,
		  const hePointer& he
		  )
  {
    // Creates edge
    egPointer newe = tEdge2::New();

    // Inserts the new edge into its corresponding list
    m_ledges.push_back(newe);

    // Creates the first half-edge of the new edge
    hePointer newhe = tHalfEdge2::New();

    // Initializes first half-edge
    newhe->setVertex(v);
    newhe->setEdge(newe);
    newhe->setFace(he->getFace());

    // Inserts new half-edge in the face cycle
    hePointer henext = he->getNext();
    henext->setPrev(newhe);
    he->setNext(newhe);
    newhe->setNext(henext);
    newhe->setPrev(he);

    // Updates vertex
    v->setHalfEdge(newhe);
    v->setBoundHalfEdge(newhe);

    // Initializes new edge
    newe->setHalfEdge1(newhe);

    // Sets boundary to be the same as that one of "he"
    newe->setBoundary(he->getEdge()->getBoundary());

    // Returns the new edge
    return newe;
  }


  // -----------------------------------------------------------------
  // Method spEmkEV()
  //
  // Splits an edge by adding a new vertex inbetween its endpoints.
  //
  // -----------------------------------------------------------------
  tComplex2::egPointer
  tComplex2::spEmkEV(
		     const vtPointer& v,
		     const hePointer& he
		    )
  {
    // Creates a new edge
    egPointer newe = tEdge2::New();

    // Inserts the new edge into its corresponding list
    m_ledges.push_back(newe);

    // Creates half-edges of the new edge
    hePointer newhe1 = tHalfEdge2::New();
    hePointer newhe2 = tHalfEdge2::New();

    // Initializes first half-edge
    newhe1->setVertex(v);
    newhe1->setEdge(newe);
    newhe1->setFace(he->getFace());

    // Initializes second half-edge
    hePointer hemate = he->getEdge()->getMate(he);
    newhe2->setVertex(hemate->getVertex());
    newhe2->setEdge(newe);
    newhe2->setFace(hemate->getFace());

    // Inserts first half-edge in the face cycle
    hePointer henext = he->getNext();
    henext->setPrev(newhe1);
    he->setNext(newhe1);
    newhe1->setNext(henext);
    newhe1->setPrev(he);

    // Inserts second half-edge in the face cycle
    hePointer heprev = hemate->getPrev();
    heprev->setNext(newhe2);
    hemate->setPrev(newhe2);
    newhe2->setNext(hemate);
    newhe2->setPrev(heprev);

    // Updates vertices
    hemate->setVertex(v);
    newhe2->getVertex()->setHalfEdge(newhe2);
    v->setHalfEdge(newhe1);
    v->setBoundHalfEdge(hePointer(0));

    // Initializes new edge
    newe->setHalfEdge1(newhe1);
    newe->setHalfEdge2(newhe2);
    newe->setBoundary(bdPointer(0));

    // Updates dual graph to include one more edge
    createDualEdge(newhe1->getFace(),newhe2->getFace());

    // Returns the new edge
    return newe;
  }

  // -----------------------------------------------------------------
  // Method mkFE()
  //
  // It creates a new face and a new edge. The new face is incident to
  // an edge of an existent face.  The half-edge of the commom edge in
  // the adjacent face is given as an input parameter.
  //
  // -----------------------------------------------------------------
  tComplex2::fcPointer
  tComplex2::mkFE(const hePointer& he)
  {
    // Creates new edge and new face
    egPointer newe = tEdge2::New();
    fcPointer newf = tFace2::New();

    // Inserts the new elements into the corresponding lists
    m_ledges.push_back(newe);
    m_lfaces.push_back(newf);

    // Gets common face
    fcPointer face = he->getFace();

    // Gets commom edge
    egPointer edge = he->getEdge();

    // Gets commom vertices
    vtPointer v1 = he->getNext()->getVertex();
    vtPointer v2 = he->getVertex();

    // Creates the first half-edge of the new edge
    hePointer newhe1 = tHalfEdge2::New();
    
    // Creates the second half-edge of the commom edge
    hePointer newhe2 = tHalfEdge2::New();
    
    // Initializes first half-edge of the new edge
    newhe1->setVertex(v2);
    newhe1->setEdge(newe);
    newhe1->setFace(newf);
    newhe1->setNext(newhe2);
    newhe1->setPrev(newhe2);

    // Initializes second half-edge of the commom edge
    newhe2->setVertex(v1);
    newhe2->setEdge(edge);
    newhe2->setFace(newf);
    newhe2->setNext(newhe1);
    newhe2->setPrev(newhe1);
  
    // Initializes new edge
    newe->setHalfEdge1(newhe1);
    newe->setBoundary(edge->getBoundary());

    // Updates vertex
    v2->setBoundHalfEdge(newhe1);

    // Initializes new face
    newf->setHalfEdge(newhe1);
    newf->setComplex(Pointer(this));

    // Updates the boundary
    edge->getBoundary()->setBoundaryEdge(newe);

    // Updates commom edge
    edge->setHalfEdge2(newhe2);
    edge->setBoundary(bdPointer(0));

    // Creates a corresponding vertex in the dual graph
    createDualVertex(newf);
  
    // Creates an edge  in the dual graph corresponding  to the common
    // edge between newf and face.
    createDualEdge(face,newf);

    // Returns the new face
    return newf;
  }

  // -----------------------------------------------------------------
  // Method mkFE()
  //
  // Creates a new face and a new edge by splitting an existent face.
  //
  // -----------------------------------------------------------------
  tComplex2::fcPointer
  tComplex2::mkFE(
		  const hePointer& he1, 
		  const hePointer& he2
		  )
    throw (tExceptionObject)
  {
    // If the two  input half-edge are not in the  same face, throw an
    // exception.
    if (he1->getFace() != he2->getFace()) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Attempt to split face using half-edges that do not belong to the same face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // If the half-edges are consecutive, throw an exception
    if ((he1->getNext() == he2) || (he1->getPrev() == he2)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Attempt to split face using consecutive half-edges";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // If the half-edges are the same, throw an exception
    if (he1 == he2) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Attempt to split face using identical half-edges";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Creates new edge and new face
    egPointer newe = tEdge2::New();
    fcPointer newf = tFace2::New();

    // Inserts the new elements into the corresponding lists
    m_ledges.push_back(newe);
    m_lfaces.push_back(newf);

    // Gets common face
    fcPointer face = he1->getFace();

    // Removes dual vertex of "face" from the dual graph
    deleteDualVertex(face);

    // Gets commom vertices
    vtPointer v3 = he1->getVertex();
    vtPointer v4 = he2->getVertex();

    // Creates the first half-edge of the new edge
    hePointer newhe1 = tHalfEdge2::New();

    // Creates the second half-edge of the commom edge
    hePointer newhe2 = tHalfEdge2::New();

    // Initializes first half-edge of the new edge
    newhe1->setVertex(v4);
    newhe1->setEdge(newe);
    newhe1->setFace(face);

    // Initializes second half-edge of the commom edge
    newhe2->setVertex(v3);
    newhe2->setEdge(newe);
    newhe2->setFace(newf);

    // Splits cycle of half-edges
    newhe1->setNext(he1);
    newhe1->setPrev(he2->getPrev());
    newhe2->setNext(he2);
    newhe2->setPrev(he1->getPrev());
    he2->getPrev()->setNext(newhe1);
    he1->getPrev()->setNext(newhe2);
    he2->setPrev(newhe2);
    he1->setPrev(newhe1);
  
    // Initializes new edge
    newe->setHalfEdge1(newhe1);
    newe->setHalfEdge2(newhe2);
    newe->setBoundary(bdPointer(0));

    // Initialize new face
    newf->setHalfEdge(newhe2);
    newf->setComplex(Pointer(this));

    // Updates old face
    face->setHalfEdge(he1);

    // Creates corresponding vertices in the dual graph
    createDualVertex(face);
    createDualVertex(newf);
  
    // Creates all edges corresponding to the adjacent faces to "face"
    // and "newf"  in the dual  graph. Also, update half-edges  in the
    // new face.
    hePointer heaux1 = face->getHalfEdge();
    hePointer heaux2 = heaux1;
    do {
      hePointer heaux3 = heaux2->getEdge()->getMate(heaux2);
      if (heaux3.isNotNull()) {
	createDualEdge(face,heaux3->getFace());
      }
      heaux2 = heaux2->getNext();
    } while (heaux1 != heaux2);

    heaux2 = he2;
    do {
      // Updates  pointer to  new face  of all  half-edges  in "newf",
      // except for "newhe2" that is already set.
      heaux2->setFace(newf);

      heaux1 = heaux2->getEdge()->getMate(heaux2);
      if (heaux1.isNotNull()) {
	createDualEdge(newf,heaux1->getFace());
      }
      heaux2 = heaux2->getNext();
    } while (heaux2 != newhe2);

    // Returns the new face
    return newf;
  }

  // -----------------------------------------------------------------
  // Method spFmkVE()
  //
  // Creates a  new vertice and  a new edge  in the face  containing a
  // given half-edge "he". The new vertex has location at "p", and the
  // new  edge has endpoints  "p" and  "q", where  "q" is  the initial
  // vertex of the half-edge following "he" in its face.
  //
  // -----------------------------------------------------------------
  tComplex2::vtPointer
  tComplex2::spFmkFE(
		     const hePointer& he,
		     const Point2& p
		     ) 
  {
    // If new edge is not on  the boundary, split the face that shares
    // it with the face containing "he".
    egPointer e;
    vtPointer v = createVertex(p);
    if (!he->getEdge()->isOnBoundary()) {
      // Creates new vertex and new edge
      e = spEmkEV(v,he);

      // Gets the mate of he
      hePointer he1 = he->getEdge()->getMate(he);

      // Gets a  half-edge that has an  initial vertex that  is not in
      // the common edge.
      hePointer he2 = he1->getNext()->getNext();
    
      // Splits the  face containing "he1"  and "he2" by  inserting an
      // edge linking their initial vertex.
      mkFE(he1,he2);
    }
    else {
      // Creates new vertex and new edge
      e = mkEV(v,he);
    }

    e->constraintOn();

    return v;
  }

  // -----------------------------------------------------------------
  // Method rmBEV
  //
  // Glues two connected components of  a complex by removing an edge,
  // two vertices and one boundary.
  //
  // -----------------------------------------------------------------
  void
  tComplex2::rmBEV(
		   const hePointer& he1, 
		   const hePointer& he2
		  )
  {
    // Gets   initial  vertex   of  "he1",   "he2",   "he1->next"  and
    // "he2->next".
    vtPointer v1 = he1->getVertex();
    vtPointer v2 = he1->getNext()->getVertex();
    vtPointer v3 = he2->getVertex();
    vtPointer v4 = he2->getNext()->getVertex();

    // Gets the edge containing "he1" and the edge containing "he2"
    egPointer e1 = he1->getEdge();
    egPointer e2 = he2->getEdge();

    // Gets the face containing "he1" and the face containing "he2"
    fcPointer f1 = he1->getFace();
    fcPointer f2 = he2->getFace();

    // Gets the boundary containing  "he1" and the boundary containing
    // "he2".
    bdPointer bd1 = e1->getBoundary();
    bdPointer bd2 = e2->getBoundary();

    // Merge boundary
    //
    // Sets boundary edge pointer in face containing "he2" to point to
    // the boundary containing "he1".
    egPointer eaux = getNextBoundEdge(e2);
    while (eaux != e2) {
      eaux->setBoundary(bd1);
      eaux = getNextBoundEdge(eaux);
    }

    // If  boundary containing  "he1"  points to  the edge  containing
    // "he1",  we  must  make it  point  to  other  edge in  the  same
    // boundary.
    //
    // Gets next edge in the boundary
    if (e1 == bd1->getBoundaryEdge())
      bd1->setBoundaryEdge(getNextBoundEdge(e1));

    // Updates every half-edge with initial vertex "v3" and "v4"
    hePointer heaux = getNextStarVert(v3->getHalfEdge());
    while (heaux.isNotNull()) {
      heaux->setVertex(v2);
      heaux = getNextStarVert(heaux);
    }

    heaux = getPrevStarVert(v3->getHalfEdge());
    while (heaux.isNotNull())  {
      heaux->setVertex(v2);
      heaux = getPrevStarVert(heaux);
    }
  
    v3->getHalfEdge()->setVertex(v2);

    heaux = getNextStarVert(v4->getHalfEdge());
    while (heaux.isNotNull()) {
      heaux->setVertex(v1);
      heaux = getNextStarVert(heaux);
    }

    heaux = getPrevStarVert(v4->getHalfEdge());
    while (heaux.isNotNull())  {
      heaux->setVertex(v1);
      heaux = getPrevStarVert(heaux);
    }

    // Updates boundary half-edge of "v1"
    v1->setBoundHalfEdge(v4->getBoundHalfEdge());

    v4->getHalfEdge()->setVertex(v1);

    // Makes "e1" the edge of "he2"
    he2->setEdge(e1);

    // Makes "he2" the mate of "he1"
    e1->setHalfEdge2(he2);

    // "e1" is no longer a boundary edge
    e1->setBoundary(bdPointer(0));

    // Removes vertices
    m_lverts.remove(v3);
    m_lverts.remove(v4);

    // Removes edge that contained "he2"
    m_ledges.remove(e2);

    // Removes boundary that contained "he2"
    m_lbounds.remove(bd2);

    // Creates an edge  in the dual graph corresponding  to the common
    // edge between "newf" and "face".
    createDualEdge(f1,f2);
  }

  // -----------------------------------------------------------------
  // Method mkGrmEV
  //
  // Glues two faces  of the same connected component,  and removes an
  // edge  and at  most two  vertices.  It  can also  make a  genus by
  // spliting one boundary.
  //
  // -----------------------------------------------------------------
  void
  tComplex2::mkGrmEV(
		     const hePointer& he1, 
		     const hePointer& he2
		     )
  {
    // Gets   initial  vertex   of  "he1",   "he2",   "he1->next"  and
    // "he2->next".
    vtPointer v1 = he1->getVertex();
    vtPointer v2 = he1->getNext()->getVertex();
    vtPointer v3 = he2->getVertex();
    vtPointer v4 = he2->getNext()->getVertex();

    // Gets the edge containing "he1" and the edge containing "he2"
    egPointer e1 = he1->getEdge();
    egPointer e2 = he2->getEdge();

    // Gets the face containing "he1" and the face containing "he2"
    fcPointer f1 = he1->getFace();
    fcPointer f2 = he2->getFace();

    // Gets the boundary containing "he1" and "he2"
    bdPointer bd = e1->getBoundary();

    // Gets edges preceding and succeeding "e1" in "bd"
    egPointer eprev = getPrevBoundEdge(e1);
    egPointer enext = getNextBoundEdge(e1);

    if (eprev == enext) {
      // Boundary has only two edges: "e1" and "e2"
      // We must remove the boundary

      m_lbounds.remove(bd);

      // Update initial vertex of "he1" and "he2"
      he1->getVertex()->setBoundHalfEdge(hePointer(0));
      he2->getVertex()->setBoundHalfEdge(hePointer(0));
    }
    else if (eprev == e2) {
      // "he1" follows "he2" on the boundary "bd"

      // Updates edge pointer of "bd"
      bd->setBoundaryEdge(enext);

      // Updates initial vertex of "he1"
      he1->getVertex()->setBoundHalfEdge(hePointer(0));
    }
    else if (enext == e2) {
      // "he1" precedes "he2" on the boundary "bd"

      // Updates edge pointer of boundary "bd"
      bd->setBoundaryEdge(eprev);

      // Updates initial vertex of "he1->next"
      he1->getNext()->getVertex()->setBoundHalfEdge(hePointer(0));

      // Updates boundary half-edge of "v1"
      v1->setBoundHalfEdge(v4->getBoundHalfEdge());
    }
    else {
      // Boundary "bd" must be split into two ones

      // Creates another boundary
      bdPointer bdaux = tBoundary2::New();

      // Inserts it into list of boundaries of the complex
      m_lbounds.push_back(bdaux);

      // Sets first edge of the new boundary
      bdaux->setBoundaryEdge(enext);

      // Updates edges of the new boundary
      egPointer eaux = enext;
      while (eaux != e2) {
	eaux->setBoundary(bdaux);
	eaux = getNextBoundEdge(eaux);
      }

      // Updates edge pointer of the old boundary
      bd->setBoundaryEdge(eprev);

      // Updates the boundary half-edge of "he1->next"
      v1->setBoundHalfEdge(v4->getBoundHalfEdge());
    }

    // Updates every half-edge with initial vertex "v3" and "v4"
    hePointer heaux = getNextStarVert(v3->getHalfEdge());
    while (heaux.isNotNull()) {
      heaux->setVertex(v2);
      heaux = getNextStarVert(heaux);
    }

    heaux = getPrevStarVert(v3->getHalfEdge());
    while (heaux.isNotNull())  {
      heaux->setVertex(v2);
      heaux = getPrevStarVert(heaux);
    }

    v3->getHalfEdge()->setVertex(v2);

    heaux = getNextStarVert(v4->getHalfEdge());
    while (heaux.isNotNull()) {
      heaux->setVertex(v1);
      heaux = getNextStarVert(heaux);
    }

    heaux = getPrevStarVert(v4->getHalfEdge());
    while (heaux.isNotNull())  {
      heaux->setVertex(v1);
      heaux = getPrevStarVert(heaux);
    }

    v4->getHalfEdge()->setVertex(v1);

    // Makes "e1" the edge of "he2"
    he2->setEdge(e1);

    // Makes "he2" the mate of "he1"
    e1->setHalfEdge2(he2);

    // "e1" is no longer a boundary edge
    e1->setBoundary(bdPointer(0));

    // Removes vertices if needed
    if (v3 != v2) {
      m_lverts.remove(v3);
    }

    if (v4 != v1) {
      m_lverts.remove(v4);
    }

    // Removes edge that contained "he2"
    m_ledges.remove(e2);

    // Creates an edge  in the dual graph corresponding  to the common
    // edge between "newf" and "face".
    createDualEdge(f1,f2);
  }

  // -----------------------------------------------------------------
  // Method mgFrmE
  //
  // Merge two faces by removing their common edge.
  //
  // -----------------------------------------------------------------
  tComplex2::fcPointer
  tComplex2::mgFrmE(const hePointer& he1) throw (tExceptionObject)
  {
    // Gets the edge containing the half-edge "he"
    egPointer e = he1->getEdge();

    // If this edge is a boundary edge then it is incident to only one
    // face.
    if (e->isOnBoundary()) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Edge used to merge faces is a boundary edge";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Gets mate of half-edge "he1"
    hePointer he2 = e->getMate(he1);

    // Checks if both faces have at least two edges
    if ((he1 == he1->getNext()) || (he2 == he2->getNext())) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Attempt to merge face containing only one edge";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Updates initial vertex of both "he1" and "he2"
    he1->getVertex()->setHalfEdge(he2->getNext());
    he2->getVertex()->setHalfEdge(he1->getNext());

    // Updates half-edges in the face containing "he2"
    fcPointer f1 = he1->getFace();
    fcPointer f2 = he2->getFace();
    hePointer he = he2->getNext();
    do {
      he->setFace(f1);
      he = he->getNext();
    } while (he != he2);

    // Updates the dual graph
    deleteDualVertex(f2);
    he = he2->getNext();
    do {
      if (he->getEdge()->getMate(he).isNotNull()) {
	createDualEdge(f1,he->getEdge()->getMate(he)->getFace());
      }
      he = he->getNext();
    } while (he != he2);

    he1->getPrev()->setNext(he2->getNext());
    he2->getNext()->setPrev(he1->getPrev());
    he1->getNext()->setPrev(he2->getPrev());
    he2->getPrev()->setNext(he1->getNext());

    // Updates face "f1"
    f1->setHalfEdge(he1->getNext());
    
    // Set to null all pointer in "he1", "he2", "e" and "f2"
    he1->setPrev(hePointer(0));
    he1->setNext(hePointer(0));
    he1->setVertex(vtPointer(0));
    he1->setEdge(egPointer(0));
    he1->setFace(fcPointer(0));

    he2->setPrev(hePointer(0));
    he2->setNext(hePointer(0));
    he2->setVertex(vtPointer(0));
    he2->setEdge(egPointer(0));
    he2->setFace(fcPointer(0));

    e->setHalfEdge1(hePointer(0));
    e->setHalfEdge2(hePointer(0));
    e->setBoundary(bdPointer(0));

    f2->setHalfEdge(hePointer(0));
    f2->setComplex(Pointer(0));
    f2->setDualVertex(dvPointer(0));

    // Removes  "e"  and  "f2"  from  the list  of  edges  and  faces,
    // respectively, of this cell complex.
    m_ledges.remove(e);
    m_lfaces.remove(f2);

    return f1;
  }

};     // end namespace hedge
