//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    complex2.h
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#ifndef __complex2_h
#define __complex2_h

#include <list>

#include "typedefs.h"
#include "macros.h"
#include "smartptr.h"
#include "dataobj.h"
#include "exceptionobj.h"

#include "bbox2.h"

#include "vertex2.h"
#include "vertex2attrib.h"
#include "halfedge2.h"
#include "edge2.h"
#include "edge2attrib.h"
#include "face2.h"
#include "face2attrib.h"
#include "boundary2.h"
#include "dualgraph.h"

namespace hedge {

  // -----------------------------------------------------------------
  // class tComplex2 
  //
  // This  class implements  a cell  complex in  the  handle-edge data
  // structure.
  //
  // -----------------------------------------------------------------

  class tComplex2 : public tDataObject {
  public:
    // Standard class typedefs
    typedef tComplex2             Self;
    typedef tSmartPtr<Self>       Pointer;
    typedef tSmartPtr<const Self> ConstPointer;

    // Type definitions
    typedef tBBox2::Pointer           bbPointer;
    typedef tDualGraphVertex::Pointer dvPointer;
    typedef tDualGraph::Pointer       dgPointer;
    typedef tBoundary2::Pointer       bdPointer;
    typedef tFace2::Pointer           fcPointer;
    typedef tFace2Attrib::Pointer     faPointer;
    typedef tEdge2::Pointer           egPointer;
    typedef tEdge2Attrib::Pointer     eaPointer;
    typedef tHalfEdge2::Pointer       hePointer;
    typedef tVertex2::Pointer         vtPointer;
    typedef tVertex2Attrib::Pointer   vaPointer;

    typedef list<vtPointer>::const_iterator   vtIterator;
    typedef list<egPointer>::const_iterator   egIterator;
    typedef list<fcPointer>::const_iterator   fcIterator;
    typedef list<bdPointer>::const_iterator   bdIterator;

    // Method for creating an instance of this object
    newMacro(Self);

    // Default constructor
    tComplex2();

    // Virtual destructor
    virtual ~tComplex2();

    // Returns the number of vertices in this cell complex
    int getNumVerts() const
      { return m_lverts.size(); }

    // Returns the number of edges in this cell complex
    int getNumEdges() const
      { return m_ledges.size(); }

    // Returns the number of faces in this cell complex
    int getNumFaces() const
      { return m_lfaces.size(); }

    // Returns the number of holes in this cell complex
    int getNumHoles() const
      {
	return 1 - (this->getNumVerts()) + 
	  (this->getNumEdges()) - (this->getNumFaces());
      }

    // Returns the number of boundaries of this cell complex
    int getNumBounds() const
      { return m_lbounds.size(); }

    // Returns a smart pointer to a bounding box of this cell complex
    bbPointer getBoundingBox() const
      { return m_bb; }

    // Returns a smart pointer to the dual graph of this cell complex
    dgPointer getDualGraph() const
      { return m_dual; }

    // Returns the next half-edge in the star of the initial vertex of
    // a given half-edge.
    hePointer getNextStarVert(const hePointer&);

    // Returns  the previous  half-edge  in the  star  of the  initial
    // vertex of a given half-edge.
    hePointer getPrevStarVert(const hePointer&);

    // Returns the edge following a given edge in its boundary curve
    egPointer getNextBoundEdge(const egPointer&);

    // Returns the edge preceding a given edge in its boundary curve
    egPointer getPrevBoundEdge(const egPointer&);

    // Returns true if a given vertex belongs to a constrained edge
    bool isOnConstrainedEdge(const vtPointer&) const 
      throw (tExceptionObject);

    // Computes a polygon corresponding to the star of a vertex.
    vector<Point2> getVertexStar(const vtPointer&) 
      throw (tExceptionObject);

    // Creates a new vertex and returns it
    vtPointer createVertex(const Point2&);

    // Creates a new face, boundary,  edge and vertex, and returns the
    // new face.
    fcPointer mkFBEV(const vtPointer&);

    // Creates a new edge and a new vertex, and returns the new edge.
    egPointer mkEV(const vtPointer&, const hePointer&);

    // Splits an edge and makes a  new edge and a new vertex, and then
    // returns the new edge.
    egPointer spEmkEV(const vtPointer&, const hePointer&);

    // Creates a new edge and face, and returns the new face
    fcPointer mkFE(const hePointer&);

    // Creates a new edge and face, and returns the new face
    fcPointer mkFE(const hePointer&, const hePointer&) 
      throw (tExceptionObject);

    // Splits a face and makes a new face and edge
    vtPointer spFmkFE(const hePointer&, const Point2&);

    // Removes edge and face
    void rmEF(const hePointer&);

    // Removes boundary, edge and vertex
    void rmBEV(const hePointer&, const hePointer&);

    // Creates a hole and removes edge and vertex 
    void mkGrmEV(const hePointer&, const hePointer&);

    // Merge two faces by removing their common edge
    fcPointer mgFrmE(const hePointer&) throw (tExceptionObject);

    // Methods to allow the use of iterators
    vtIterator vtBegin() { return m_lverts.begin();  }
    egIterator egBegin() { return m_ledges.begin();  }
    fcIterator fcBegin() { return m_lfaces.begin();  }
    bdIterator bdBegin() { return m_lbounds.begin(); }
 
    vtIterator vtEnd() { return m_lverts.end();  }
    egIterator egEnd() { return m_ledges.end();  }
    fcIterator fcEnd() { return m_lfaces.end();  }
    bdIterator bdEnd() { return m_lbounds.end(); }

  private:
    // Purposely not implemented
    tComplex2(const Self&);
    void operator=(const Self&);

    // Creates a vertex in the dual graph
    void createDualVertex(const fcPointer&);

    // Creates an edge in the dual graph
    void createDualEdge(const fcPointer&, const fcPointer&);

    // Removes a vertex from the dual graph
    void deleteDualVertex(const fcPointer&);

    //
    // Private attributes
    //

    // List of all vertices in this complex
    list<vtPointer> m_lverts;

    // List of all edges in this complex
    list<egPointer> m_ledges;

    // List of all faces in this complex
    list<fcPointer> m_lfaces;

    // List of all boundaries in this complex
    list<bdPointer> m_lbounds;

    // Bounding box enclosing this complex
    bbPointer m_bb;

    // Dual graph of this complex
    dgPointer m_dual;
  };


};    // end namespace hedge


#endif   // __complex2_h
