//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    dualgraph.h
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#ifndef __dualgraph_h
#define __dualgraph_h

#include <list>
#include <vector>

#include "typedefs.h"
#include "macros.h"
#include "smartptr.h"
#include "dataobj.h"

#include "edge2.h"
#include "face2.h"

using namespace std;

namespace hedge {

  // -----------------------------------------------------------------
  // Forward definition of classes
  // -----------------------------------------------------------------
  class tComplex2;
  class tDualGraph;

  // -----------------------------------------------------------------
  // Constant definition
  // -----------------------------------------------------------------

  const int DEFAULT_GRAPH_DEGREE = 6;

  // -----------------------------------------------------------------
  // class tDualGraphVertex 
  //
  // This  class implements  vertices  in  the dual  graph  of a  cell
  // complex represented by the handle-edge data structure.
  //
  // -----------------------------------------------------------------

  class tDualGraphVertex : public tDataObject {
  public:
    // Standard class typedefs
    typedef tDualGraphVertex      Self;
    typedef tSmartPtr<Self>       Pointer;
    typedef tSmartPtr<const Self> ConstPointer;

    // Type definitions
    typedef tSmartPtr<tDualGraph> dgPointer;
    typedef tFace2::Pointer       fcPointer;

    // Method for creating an instance of this object
    newMacro(Self);

    // Default constructor
    tDualGraphVertex()
    {
      m_degree = 0;
      m_maxdg = DEFAULT_GRAPH_DEGREE;
      m_marked = false; 
      m_neigh.resize(m_maxdg);
    }

    // Destructor
    ~tDualGraphVertex() {}

    // Returns the current degree of this vertex
    int getDegree() const
      { return m_degree; }

    // Returns the maximum number of vertices that can be connected to
    // this vertex by a graph edge.
    int getMaxDegree() const
      { return m_maxdg; }

    // Returns a smart pointer to the i-th neighbor of this vertex
    Pointer getNeighbor(int i) const
      { return m_neigh[i]; }

    // Returns  a smart  pointer  to the  face  corresponding to  this
    // vertex in the cell complex.
    fcPointer getFace() const
      { return m_face; }

    // Returns  a smart  pointer  to the  dual  graph containing  this
    // vertex.
    dgPointer getDualGraph() const
      { return m_dual; }

    // Returns true if this vertex is "marked"
    bool isMarked() const
      { return m_marked; }

    // Returns true if this vertex is  the dual of a face containing a
    // constrained edge of the cell complex.
    bool isDualofConstrainedEdge(int i) const
      { return m_face->getCommonEdge(m_neigh[i]->getFace())->isConstrained(); }

    // Set this vertex as marked or unmarked 
    void setMark(bool mark)
      { m_marked = mark; }

    // Friend class
    //
    // This is the only class  whose methods can use the private "set"
    // methods of the class tDualGraphVertex.
    //
    friend class tDualGraph;
    friend class tComplex2;

  private:
    // Purposely not implemented
    tDualGraphVertex(const Self&);
    void operator=(const Self&);
    
    // Set a  value to the pointer  to the face  corresponding to this
    // vertex in the cell complex.
    void setFace(const fcPointer& f)
      { m_face = f; }

    // Set a  value to the pointer  to the dual  graph containing this
    // vertex.
    void setOwner(const dgPointer& g)
      { m_dual = g; }

    // Adds a neighbor to this vertex
    void addNeighbor(const Pointer&);

    // Deletes a neighbor of this vertex
    void delNeighbor(const Pointer&);

    //
    // Private attributes
    //    

    // Smart pointer to the dual graph this vertex belongs to
    dgPointer m_dual;

    // Smart  pointer to  the dual  face of  this vertex  in  the cell
    // complex.
    fcPointer m_face;

    // Array of smart pointers to the vertices adjacent to this one in
    // the dual graph.
    vector<Pointer> m_neigh;

    // Maximum number of adjacent vertices
    int m_maxdg;

    // Current vertex degree
    int m_degree;

    // Flag to indicate if this vertex is or not marked
    bool m_marked;
  };


  // -----------------------------------------------------------------
  // class tDualGraph
  //
  // This  class   implements  the  dual  graph  of   a  cell  complex
  // represented by the handle-edge data structure.
  //
  // -----------------------------------------------------------------
  class tDualGraph : public tDataObject {
  public:
    // Standard class typedefs
    typedef tDualGraph            Self;
    typedef tSmartPtr<Self>       Pointer;
    typedef tSmartPtr<const Self> ConstPointer;

    // Type definitions
    typedef tSmartPtr<tComplex2>      cpPointer;
    typedef tDualGraphVertex::Pointer dvPointer;

    // Method for creating an instance of this object
    newMacro(Self);

    // Default constructor
    tDualGraph() {}

    // Destructor
    ~tDualGraph()
      {  m_lverts.clear(); }

    // Returns the number of vertices of this dual graph
    int getNumVerts() const
      {  return m_lverts.size(); }

    // Sets all vertices in this dual graph as unmarked
    void clearMark();

    // Friend class
    //
    // This is the only class  whose methods can use the private "set"
    // methods of the class tDualGraph.
    //
    friend class tComplex2;

  private:
    // Purposely not implemented
    tDualGraph(const Self&);
    void operator=(const Self&);
    
    // Adds a vertex to this dual graph
    void addVert(const dvPointer& v)
      {
	m_lverts.push_back(v); 
      }

    // Removes a vertex from this dual graph
    void delVert(const dvPointer&);

    // Adds an edge to this dual graph
    void addEdge(const dvPointer& v1, const dvPointer& v2)
      { v1->addNeighbor(v2); v2->addNeighbor(v1); }

    // Removes an edge from this dual graph
    void delEdge(const dvPointer& v1, const dvPointer& v2)
      { v1->delNeighbor(v2); v2->delNeighbor(v1); }

    // Sets the value  of the pointer to the cell  complex that is the
    // dual of this dual graph.
    void setDual(const cpPointer& c)
      { m_mesh = c; }

    //
    // Private Attributes
    //

    // List of all vertices in this dual graph
    list<dvPointer> m_lverts;

    // Smart pointer to the cell complex that is the dual of this dual
    // graph.
    cpPointer m_mesh;
  };

};   // end namespace hedge

/*
#ifdef QMPP_MANUAL_INSTANTIATION
#include "dualgraph.cc"
#endif
*/

#endif   // __dualgraph_h










