//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    edge2.h
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#ifndef __edge2_h
#define __edge2_h

#include "typedefs.h"
#include "macros.h"
#include "smartptr.h"
#include "dataobj.h"

#include "vertex2.h"
#include "halfedge2.h"
#include "edge2attrib.h"

namespace hedge {

  // -----------------------------------------------------------------
  // Forward definition of classes
  // -----------------------------------------------------------------
  class tComplex2;
  class tBoundary2;

  // -----------------------------------------------------------------
  // class tEdge2 
  //
  // This class implements edges in the handle-edge data structure.
  //
  // -----------------------------------------------------------------

  class tEdge2 : public tDataObject {
  public:
    // Standard class typedefs
    typedef tEdge2                Self;
    typedef tSmartPtr<Self>       Pointer;
    typedef tSmartPtr<const Self> ConstPointer;

    // Type definitions
    typedef tSmartPtr<tBoundary2> bdPointer;
    typedef tHalfEdge2::Pointer   hePointer;
    typedef tVertex2::Pointer     vtPointer;
    typedef tEdge2Attrib::Pointer eaPointer;

    // Method for creating an instance of this object
    newMacro(Self);

    // Destructor
    ~tEdge2() {}

    // Returns  a smart  pointer to  the initial  vertex of  the first
    // half-edge of this edge.
    vtPointer getVertex1() const
      { 
	if (m_he1.isNotNull()) return m_he1->getVertex();
	return vtPointer(0); 
      }

    // Returns  a smart  pointer to  the initial  vertex of  the first
    // half-edge of this edge. 
    vtPointer getVertex2() const
      {
	if (m_he2.isNotNull()) return m_he2->getVertex();
	return vtPointer(0);
      }

    // Returns a smart pointer to the first half-edge of this edge
    hePointer getHalfEdge() const
      { return m_he1; }

    // Returns a smart pointer to the second half-edge of this edge
    hePointer getMate(const hePointer& he) const
      {
	if (he == m_he1) return m_he2;
	return m_he1;
      }

    // Returns a smart pointer to the boundary containing this edge
    bdPointer getBoundary() const
    { return m_bd; }

    // Returns a smart pointer to the set of attributes of this edge
    eaPointer getAttribute() const 
      { return m_att; }

    // Returns a smart  pointer to the common vertex  of this edge and
    // another given edge.
    vtPointer getCommonVertex(const Pointer&) const;

    // Returns  true if  this  edge is  a  boundary edge.   Otherwise,
    // returns false.
    bool isOnBoundary() const
      { return m_bd.isNotNull(); }

    // Returns true if this  is a constrained edge. Otherwise, returns
    // false.
    bool isConstrained() const
      { return m_constrained; }

    // Set this edge as a constrained edge
    void constraintOn() 
      { m_constrained = true; }

    // Set this edge as an unconstrained edge
    void constraintOff()
      { m_constrained = false; }

    // Set the first half-edge of this edge
    void setHalfEdge1(const hePointer& he)
      { m_he1 = he; }

    // Set the second half-edge of this edge
    void setHalfEdge2(const hePointer& he)
      { m_he2 = he; }

    // Set the boundary of this edge, if any
    void setBoundary(const bdPointer& bd)
      { m_bd = bd; }

    // Set the attributes of this edge
    void setAttribute(const eaPointer& att)
      { m_att = att; }

    // Friend class
    //
    // This is the only class  whose methods can use the private "set"
    // methods of the class tVertex2.
    //
    friend class tComplex2;

  private:
    // Purposely not implemented
    // Default constructor
    tEdge2() : m_constrained(false) {}
    tEdge2(const Self&);
    void operator=(const Self&);

    //
    // Private attributes
    //

    // First halfedge of this edge
    hePointer m_he1;

    // Second halfedge of this edge
    hePointer m_he2;

    // Boundary curve containing this edge
    bdPointer m_bd;

    // Flag to indicate if this edge is or not constrained
    bool m_constrained;

    // Set of attributes of this edge
    eaPointer m_att;
  };

};   // end namespace hedge


#endif   // __edge2_h
