//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    face2.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
// ===================================================================

#include "face2.h"

namespace hedge {

  // -----------------------------------------------------------------
  // Method: getCommonEdge
  // 
  // Returns  a smart  pointer to  the common  edge of  this  face and
  // another given face.
  //
  // -----------------------------------------------------------------
  tFace2::egPointer
  tFace2::getCommonEdge(const tFace2::Pointer& f)
  {
    hePointer he1 = m_he;
    egPointer e   = he1->getEdge();
  
    do {
      hePointer he2 = e->getMate(he1);
      if (he2.isNotNull() && (he2->getFace() == f)) {
	return e;
      }
      else {
	he1 = he1->getNext();
	e   = he1->getEdge();
      }
    } while (he1 != m_he);

    return egPointer(0);
  }

  // -----------------------------------------------------------------
  // Method: getNumVerts
  // 
  // Returns the number of vertices in this face.
  //
  // -----------------------------------------------------------------
  int
  tFace2::getNumVerts() const
  {
    hePointer he = m_he->getNext();

    int count = 1;
    while (he != m_he) {
      ++count;
      he = he->getNext();
    }

    return count;
  }

  // -----------------------------------------------------------------
  // Method: getBarycenter
  // 
  // Computes and returns the barycenter of this face.
  //
  // -----------------------------------------------------------------
  Point2
  tFace2::getBarycenter() const
  {
    hePointer he = m_he->getNext();
    double xc = m_he->getVertex()->getPoint().x();
    double yc = m_he->getVertex()->getPoint().y();
    int count = 1;

    while (he != m_he) {
      xc += he->getVertex()->getPoint().x();
      yc += he->getVertex()->getPoint().y();
      ++count;
      he = he->getNext();
    }

    xc /= double(count);
    yc /= double(count);

    return Point2(xc,yc);
  }

  // -----------------------------------------------------------------
  // Method: isAdjToBoundary()
  // 
  // Finds out if this face contains a boundary edge, and then returns
  // true if it does, and false otherwise.
  //
  // -----------------------------------------------------------------
  bool
  tFace2::isAdjToBoundary() const
  {
    if (m_he->getEdge()->isOnBoundary()) return true;
    else {
      hePointer he = m_he->getNext();
      while (he != m_he) {
	if (he->getEdge()->isOnBoundary()) return true;
	else he = he->getNext();
      }
    }

    return false;
  }

  // -----------------------------------------------------------------
  // Method: isAdjToConstrainedEdge()
  // 
  // Finds  out if  this face  contains a  constrained edge,  and then
  // returns true if it does, and false otherwise.
  //
  // -----------------------------------------------------------------
  bool
  tFace2::isAdjToConstrainedEdge() const
  {
    if (m_he->getEdge()->isConstrained()) return true;
    else {
      hePointer he = m_he->getNext();
      while (he != m_he) {
	if (he->getEdge()->isConstrained()) return true;
	else he = he->getNext();
      }
    }

    return false;
  }

};   // end namespace hedge
