//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    face2.h
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#ifndef __face2_h
#define __face2_h

#include "typedefs.h"
#include "macros.h"
#include "smartptr.h"
#include "dataobj.h"

#include "halfedge2.h"
#include "edge2.h"
#include "face2attrib.h"

namespace hedge {

  // -----------------------------------------------------------------
  // Forward definition of classes
  // -----------------------------------------------------------------
  class tComplex2;
  class tDualGraphVertex;

  // -----------------------------------------------------------------
  // class tFace2 
  //
  // This class implements faces in the handle-edge data structure.
  //
  // -----------------------------------------------------------------

  class tFace2 : public tDataObject {
  public:
    // Standard class typedefs
    typedef tFace2                Self;
    typedef tSmartPtr<Self>       Pointer;
    typedef tSmartPtr<const Self> ConstPointer;

    // Type definitions
    typedef tSmartPtr<tComplex2>         cpPointer;
    typedef tSmartPtr<tDualGraphVertex>  dvPointer;
    typedef tHalfEdge2::Pointer          hePointer;
    typedef tEdge2::Pointer              egPointer;
    typedef tFace2Attrib::Pointer        faPointer;

    // Method for creating an instance of this object
    newMacro(Self);

    // Virtual destructor
    ~tFace2() {}

    // Returns a smart pointer to one half-edge of this face
    hePointer getHalfEdge() const
      { return m_he; }

    // Returns  a smart pointer  to the  cell complex  containing this
    // face.
    cpPointer getComplex() const
      { return m_cp; }

    // Returns a smart pointer to the set of attributes of this face
    faPointer getAttribute() const
      { return m_att; }

    // Returns  a smart pointer  to the  vertex corresponding  to this
    // face in the dual graph of the cell complex.
    dvPointer getDualVertex() const
      { return m_dual; }

    // Returns a  smart pointer  to the common  edge of this  face and
    // another given face.
    egPointer getCommonEdge(const Pointer&);

    // Returns the number of vertices in this face
    int getNumVerts() const;

    // Returns the barycenter of this face
    Point2 getBarycenter() const;

    // Returns true if this  face contains a boundary edge. Otherwise,
    // returns false.
    bool isAdjToBoundary() const;

    // Returns   true   if    this   face   contains   a   constrained
    // edge. Otherwise, returns false.
    bool isAdjToConstrainedEdge() const;

    // Assigns a set of attributes with this face
    void setAttribute(const faPointer& att)
      { m_att = att; }

    // Sets a value  to the smart pointer to one  of the half-edges of
    // this face.
    void setHalfEdge(const hePointer& he)
      { m_he = he; }

    // Sets  a  value  to  the  smart  pointer  to  the  cell  complex
    // containing this face.
    void setComplex(const cpPointer& cp)
      { m_cp = cp; }

    // Sets a  value to the smart  pointer to the dual  vertex of this
    // face.
    void setDualVertex(const dvPointer& dual)
      { m_dual = dual; }

    // Friend class
    //
    // This is the only class  whose methods can use the private "set"
    // methods of the class tFace2.
    //
    friend class tComplex2;

  private:
    // Purposely not implemented
    tFace2() {}
    tFace2(const Self&);
    void operator=(const Self&);

    //
    // Private attributes
    //

    // One of the half-edges in the half-edge cycle of this face
    hePointer m_he;

    // Cell complex containing this face
    cpPointer m_cp;

    // Set of attributes of this face
    faPointer m_att;

    // Vertex corresponding to this face in the dual graph of the cell
    // complex.
    dvPointer m_dual;
  };

};   // end hedge namespace 

#endif   // __face2_h
