//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    reader.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include "reader.h"

namespace qmpp {

  // -----------------------------------------------------------------
  // Method: readVertexFile()
  // 
  // Reads the vertices of a 2D mesh in a ".NODE" file
  //
  // -----------------------------------------------------------------
  void
  tMeshReader::readVertexFile(VertexSet& lverts) throw (tExceptionObject) 
  {
    // Opens the file
    filebuf fb;
    if (fb.open(filename1.c_str(), ios::in) == 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Unable to open input file";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    istream istr(&fb);
    tLexer is(istr, linecount_);

    // Reads first line:
    //
    // <# of vertices><dimension><# of attributes><boundary markers (0 or 1)> 
    //
    int nVerts;
    int dim;
    int nAttr;       // we are not using this information
    int bMarker;

    // Reads the number of vertices in the mesh
    is.get_int(nVerts);
    if (nVerts < 3) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << filename1.c_str() << " contains " << nVerts << " points";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Reads  the dimension  of the  space the  vertices are  in. This
    // number has to be 2 for our purposes.
    is.get_int(dim);
    if (dim != 2) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << "I can only deal with dimension 2";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Reads  a   number  that   indicates  the  presence   of  vertex
    // attributes.   This number  has to  be 0  as we  do  not support
    // attributes yet.
    is.get_int(nAttr);
    if (nAttr != 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << "Sorry, I do not support attributes yet.";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Reads  a  number  that   indicates  the  presence  of  boundary
    // markers. This number  must be either 0 (no) or  1 (yes), but we
    // ignore these markers when reading vertex information.
    is.get_int(bMarker);
    if (bMarker < 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << "Negative value for number of point markers";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (bMarker > 1) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << "Value greater than one for number of point markers";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Resize the input vector so that  it can fit all vertices in the
    // file.
    lverts.resize(nVerts+1);
    double cx, cy;

    //
    // Reads vertex coordinates
    //
    // <vertex id> <x coordinate> <y coordinate> [boundary mark]
  
    int pNumber = -1;
    int nNumber;
    for (int i=0; i<nVerts; i++) {

      // Reads vertex ID
      is.get_int(nNumber);

      // Verifies if vertex numbering starts with 0 or 1
      if ((i == 0) && (nNumber == 1)) {
	pNumber = 0;
      }
    
      // Verifies vertex ID sequence
      if (nNumber < (pNumber + 1)) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "Line " << linecount_ << ": " 
	   << "Vertices are not numbered correctly";
	fb.close();
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }
      else if (nNumber > (pNumber + 1)) {
        lverts.resize(lverts.size() + (nNumber - pNumber - 1));
        //
        //  Offending line (bug caught on 8/30/2005
        //  Bug caught on 8/30/2005 and reported by Milos Ivanovic
        //
	//  lverts.resize(nVerts + (nNumber - pNumber + 1));
        //
      }

      // Updates numbering sequence controller
      pNumber = nNumber;

      // Reads X coordinate of vertex nNumber
      if (!is.get_double(cx)) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "Line " << linecount_ << ": " 
	   << "Problem reading X coordinate of vertex " << nNumber;
	fb.close();
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Reads Y coordinate of vertex nNumber
      if (!is.get_double(cy)) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "Line " << linecount_ << ": " 
	   << "Problem reading Y coordinate of vertex " << nNumber;
	fb.close();
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Creates vertex with location and ID information
      lverts[nNumber] = Point2(cx,cy);

      // Reads boundary marker. Currently,  such an information is not
      // being used.
      if (bMarker == 1) {
	int marker;
	if (!is.get_int(marker)) {
	  stringstream ss (stringstream::in | stringstream::out);
	  ss << "Line " << linecount_ << ": " 
	     << "Problem reading boundary marker";
	  fb.close();
	  throw tExceptionObject(__FILE__,__LINE__,ss.str());
	}
      }
    }

    // Closes file buffer
    fb.close();
  }

  // -----------------------------------------------------------------
  // Method: readEdgeFile()
  // 
  // Reads the constrained edges of a 2D mesh in a ".EDGE" file
  //
  // -----------------------------------------------------------------
  void
  tMeshReader::readEdgeFile(EdgeSet& ledges) throw (tExceptionObject) 
  {
    // Opens the file
    filebuf fb;
    if (fb.open(filename2.c_str(), ios::in) == 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Unable to open input file";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    istream istr(&fb);
    tLexer is(istr, linecount_);

    // Reads first line:
    //
    // <# of edges><boundary markers (0 or 1)> 
    //
    int nEdges;
    int bMarker;

    // Reads the number of edges in the mesh
    is.get_int(nEdges);
    if (nEdges < 3) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << filename1.c_str() << " contains " << nEdges << " edges";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Reads a number that  indicates the presence of boundary markers
    // for the edges. This number must be either 0 (no) or 1 (yes).
    is.get_int(bMarker);
    if (bMarker < 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << "Negative value for number of point markers";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (bMarker > 1) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << "Value greater than one for number of point markers";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    //
    // Reads the code of edge endpoints
    //
    int pNumber = -1;
    int nNumber;
    int cv1, cv2;
    for (int i=0; i<nEdges; i++) {

      // Reads edge number
      is.get_int(nNumber);

      // Verifies if edge numbering starts with 0 or 1
      if ((i == 0) && (nNumber == 1)) {
	pNumber = 0;
      }
    
      // Verifies edge number sequence
      if (nNumber != (pNumber + 1)) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "Line " << linecount_ << ": " 
	   << "Edges are not numbered correctly";
	fb.close();
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Updates edge numbering sequence controller
      pNumber = nNumber;

      // Reads the code of the first endpoint of the edge
      if (!is.get_int(cv1)) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "Line " << linecount_ << ": " 
	   << "Problem reading vertex code of edge first endpoint " << nNumber;
	fb.close();
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Read the code of the second endpoint of the edge
      if (!is.get_int(cv2)) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "Line " << linecount_ << ": " 
	   << "Problem reading vertex code of edge second endpoint " << nNumber;
	fb.close();
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Reads boundary  marker. The boundary  marker of an  edge must
      // also be  0 or 1.  If  it is 1  then the edge is  a constained
      // edge.  Only constrained  edges  are inserted  into the  input
      // parameter set.
      int marker;
      if (bMarker == 1) {
	if (!is.get_int(marker)) {
	  stringstream ss (stringstream::in | stringstream::out);
	  ss << "Line " << linecount_ << ": " 
	     << "Problem reading first marker";
	  fb.close();
	  throw tExceptionObject(__FILE__,__LINE__,ss.str());
	}

	if (marker == 1) {
	  ledges.insert(make_pair(cv1,cv2));
	  ledges.insert(make_pair(cv2,cv1));
	}
      }
    }

    // Closes file buffer
    fb.close();
  }

  // -----------------------------------------------------------------
  // Method: readElemFile()
  // 
  // Reads the faces of a 2D mesh in a ".ELE" file
  //
  // -----------------------------------------------------------------
  void
  tMeshReader::readElemFile(
			    const VertexSet& lverts,
			    FaceSet& lfaces
			    ) throw (tExceptionObject)
  {
    // Opens the file
    filebuf fb;
    if (fb.open(filename3.c_str(), ios::in) == 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Unable to open input file";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    istream istr(&fb);
    tLexer is(istr, linecount_);

    // Reads file first line:
    //
    // <# of elements><# of vertices per element><# of attributes>
    //
    int nElems;
    int nVerts;      // it should match with argument nv
    int nAttrs;      // it has not been used so far

    // Reads the number of faces in the mesh
    is.get_int(nElems);
    if (nElems < 1) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << filename1.c_str() << " contains " << nElems << " elements";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    // Reads the number of vertices in any particular face
    is.get_int(nVerts);

    // Reads a number that  indicates the presence of face attributes.
    // This number must be either 0  (no) or 1 (yes). Currently, we do
    // not support attributes. So, this number has to be 0.
    is.get_int(nAttrs);
    if (nAttrs != 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Line " << linecount_ << ": " 
	 << "Sorry, we do not support face attributes yet";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    //
    // Reads node indices
    //
    int pNumber = -1;
    int eNumber;
    int i, j;
    for (i=0; i<nElems; i++) {
      // Auxiliary variable to store face information
      tPolygon2 face(nVerts);

      // Reads element number
      is.get_int(eNumber);

      // Verifies if element numbering starts with 0 or 1
      if ((i == 0) && (eNumber == 1)) {
	pNumber = 0;
      }
    
      // verifies element number sequence
      if (eNumber != (pNumber + 1)) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "Line " << linecount_ << ": " 
	   << "Elements are not numbered correctly";
	fb.close();
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Updates face numbering sequence controller
      pNumber = eNumber;

      // Reads vertex indices
      for (j=0; j<nVerts; j++) {
	int index;

	if (!is.get_int(index)) {
	  stringstream ss (stringstream::in | stringstream::out);
	  ss << "Line " << linecount_ << ": " 
	     << "Problem reading " << j+1 << "-th node index of element " << eNumber;
	  fb.close();
	  throw tExceptionObject(__FILE__,__LINE__,ss.str());
	}

	// Inserts vertex in face
	face.insert(j,index,lverts[index]);
      }

      // Inserts face into list of faces
      lfaces.push_back(face);
    }

    // Closes file buffer
    fb.close();
  }

};     // end namespace qmpp
