//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    writer.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include "writer.h"

namespace qmpp {

  // -----------------------------------------------------------------
  // Method: writeVertexFile()
  // 
  // Writes a .NODE file containing information about all vertices in
  // a 2D mesh.
  //
  // -----------------------------------------------------------------
  void
  tMeshWriter::writeVertexFile(const VertexSet& lverts) 
    throw (tExceptionObject)
  {
    // Creates an output file
    filebuf fb;
    if (fb.open(filename1.c_str(), ios::out) == 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Unable to create output file " << filename1.c_str();
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    ostream ostr(&fb);

    // Writes first line:
    //
    // <# of vertices> <dimension> <# of attributes> <boundary markers (0 or 1)> 
    //
    int nVert = lverts.size();
    int dim = 2;
    int nAttr = 0;
    int bMarker = 0;

    ostr << nVert << "\t" << dim << "\t" << nAttr << "\t" << bMarker << endl;

    VertexSet::const_iterator pos;
    for (pos = lverts.begin(); pos != lverts.end(); ++pos) {
      int vn = pos->first;
      double x = pos->second.x();
      double y = pos->second.y();

      // Writes vertex information
      //
      // <# vertex> <x coordinate> <y coordinate>
      //
      ostr << vn << "\t" << x << "\t" << y << endl;
    }

    fb.close();
  }

  // -----------------------------------------------------------------
  // Method: writeEdgeFile()
  // 
  // Writes a .EDGE  file containing information about all  edges in a
  // 2D mesh.
  //
  // -----------------------------------------------------------------
  void 
  tMeshWriter::writeEdgeFile(const EdgeSet& ledges) 
    throw (tExceptionObject)
  {
    // Creates an output file
    filebuf fb;
    if (fb.open(filename2.c_str(), ios::out) == 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Unable to create output file " << filename2.c_str();
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    ostream ostr(&fb);

    // Writes first line:
    //
    // <# of edges><boundary markers (0 or 1)> 
    //
    int nEdge = ledges.size();
    int bMarker = 1;

    ostr << nEdge << "\t" << bMarker << endl;

    EdgeSet::const_iterator pos;
    int count = 0;
    for (pos = ledges.begin(); pos != ledges.end(); ++pos) {
      int e1 = pos->first.getEdgeC1();
      int e2 = pos->first.getEdgeC2();
      ++count;
      if (pos->second) {
	ostr << count << "\t" << e1 << "\t" << e2 << "\t" << "1" << endl;
      }
      else {
	ostr << count << "\t" << e1 << "\t" << e2 << "\t" << "0" << endl;
      }
    }

    fb.close();
  }

  // -----------------------------------------------------------------
  // Method: writeElemFile()
  // 
  // Writes a .ELE file containing information about all faces in a 2D
  // mesh.
  //
  // -----------------------------------------------------------------
  void
  tMeshWriter::writeElemFile(const FaceSet& lfaces)
    throw (tExceptionObject)
  {
    // Creates an output file
    filebuf fb;
    if (fb.open(filename3.c_str(), ios::out) == 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Unable to create output file " << filename3.c_str();
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    ostream ostr(&fb);

    // Writes first line:
    //
    // <# of elements><# of vertices per element><# of attributes>
    //
    int nElem = lfaces.size();
    if (nElem == 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "List of polygons is empty";
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    int nVert = lfaces.front().getNumVerts();
    int nAttr = 0;

    ostr << nElem << "\t" << nVert << "\t" << nAttr << endl;

    // Write the elements
    FaceSet::const_iterator li = lfaces.begin();
    unsigned int count = 0;
    for (; li != lfaces.end(); ++li) {
      if ((*li).getNumVerts() != nVert) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "Polygons have distinct number of vertices";
	fb.close();
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Writes element index
      ++count; ostr << count;

      // Finds the (lowest) leftmost vertex
      Point2 qll = (*li).getPoint(0);
      int ill = 0;
      int i;
      for (i=1; i<nVert; i++) {
	Point2 qllaux = (*li).getPoint(i);
	if ((qll.x() > qllaux.x()) || ((qll.x() == qllaux.x()) && (qll.y() > qllaux.y()))) {
	  qll = qllaux;
	  ill = i;
	}
      }
    
      // Writes the index of the element vertices
      for (int i=0; i<nVert; i++) {
	ostr << "\t" << (*li).getVertexId((ill + i) % nVert);
      }

      ostr << endl;
    }

    fb.close();
  }

  // -----------------------------------------------------------------
  // Method: writePolyFile()
  // 
  // Writes a .POLY file  containing information about all contours of
  // a 2D mesh.
  //
  // -----------------------------------------------------------------
  void
  tMeshWriter::writePolyFile(
			     const VertexSet& lverts, 
			     const EdgeSet& ledges
			     )
                             throw (tExceptionObject)
  {

    // Creates the output file
    filebuf fb;
    if (fb.open(filename4.c_str(), ios::out) == 0) {
      stringstream ss (stringstream::in | stringstream::out);
      ss << "Unable to create output file " << filename4.c_str();
      fb.close();
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    ostream ostr(&fb);

    // Writes first line:
    //
    // <# of vertices> <dimension> <# of attributes> <# of boundary markers (0 or 1)>

    int nVert = lverts.size();
    int dim = 2;
    int nAttr = 0;
    int bMarker = 1;

    ostr << nVert << "\t" << dim << "\t" << nAttr << "\t" << bMarker << endl;

    VertexSet::const_iterator pos1;
    for (pos1 = lverts.begin(); pos1 != lverts.end(); ++pos1) {
      int vn = pos1->first;
      double x = pos1->second.x();
      double y = pos1->second.y();

      // Writes vertex information
      //
      // <# vertex> <x coordinate> <y coordinate>
      //
      ostr << vn << "\t" << x << "\t" << y << "\t" << 1 << endl;
    }

    // Creates a list of constrained edges
    list<tEdgeKey> lcedges;

    EdgeSet::const_iterator pos2;
    for (pos2 = ledges.begin(); pos2 != ledges.end(); ++pos2) {
      if (pos2->second) {
	lcedges.push_back(pos2->first);
      }
    }

    // Writes next line
    //
    // <# of segments> <# of boundary markers (0 or 1)>
    //
    ostr << lcedges.size() << "\t" << "1" << endl;

    // Writes the following lines
    //
    // <segment #> <endpoint> <endpoint> [boundary marker]
    //

    list<tEdgeKey>::const_iterator pos3;
    int count = 0;
    for (pos3 = lcedges.begin(); pos3 != lcedges.end(); ++pos3) {
      int e1 = pos3->getEdgeC1();
      int e2 = pos3->getEdgeC2();
      ++count;
      ostr << count << "\t" << e1 << "\t" << e2 << "\t" << "1" << endl;
    }

    // Writes number of holes
    ostr << "0" << endl;

    fb.close();
  }

};     // end namespace qmpp
