//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    control.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include "control.h"
#include <sys/time.h>

namespace qmpp {

  // -----------------------------------------------------------------
  // Method drawMesh()
  //
  // Draws a  2D mesh in an  OpenGL window. Constrained  edges will be
  // displayed with thicker lines.
  //
  // -----------------------------------------------------------------
  void
  tControl::drawMesh()
  {
    if (m_mesh == 0) return;

    cpPointer cp = m_mesh->getCellComplex();

    // Draw edges
    egPointer e;
    hePointer he;
    vtPointer v1;
    vtPointer v2;

    for (EIT ei = cp->egBegin(); ei != cp->egEnd(); ++ei) {
      e = *ei;
      he = e->getHalfEdge();
      v1 = he->getVertex();
      v2 = he->getNext()->getVertex();

      double x1 = v1->getPoint().x();
      double y1 = v1->getPoint().y();
      double x2 = v2->getPoint().x();
      double y2 = v2->getPoint().y();

      if (e->isConstrained()) {
	glLineWidth(2.0);
	glBegin(GL_LINES);
	glColor3f(0.0,1.0,0.0);
	glVertex2f(x1,y1);
	glVertex2f(x2,y2);
	glEnd();
	glLineWidth(1.0);
      }
      else {
	glBegin(GL_LINES);
	glColor3f(0.7,0.7,0.7);
	glVertex2f(x1,y1);
	glVertex2f(x2,y2);
	glEnd();
      }

      glBegin(GL_POINTS);
      glColor3f(1.0,0.0,0.0);
      glVertex2f(x1,y1);
      glVertex2f(x2,y2);
      glEnd();
    }
  }

  // -----------------------------------------------------------------
  // Method buildQuadMesh()
  //
  // Builds a quadrilateral mesh from files .NODE, .EDGE and .ELE.
  //
  // -----------------------------------------------------------------
  void
  tControl::buildQuadMesh() {
    if (m_mesh == 0) {
      m_mesh = new tQuadMesh();

      try {
	m_mesh->buildMesh(m_fname);
      }
      catch (const tExceptionObject& xpt) {
	treatException(xpt);
	exit(0);
      }
      
      cpPointer cp = m_mesh->getCellComplex();

      std::cout << "Quadrangulation has successfully been built." << std::endl;

      std::cout << std::endl;

      std::cout << "Number of vertices  : " << cp->getNumVerts()  << std::endl;
      std::cout << "Number of edges     : " << cp->getNumEdges()  << std::endl;
      std::cout << "Number of faces     : " << cp->getNumFaces()  << std::endl;
      std::cout << "Number of holes     : " << cp->getNumHoles()  << std::endl;
      std::cout << "Number of boundaries: " << cp->getNumBounds() << std::endl;

      std::cout << std::endl;

      std::cout << "Generating statistics on the input quadrangulation..." << std::endl;

      string fname = m_fname + ".stat";

      tQuadStat stat;
      stat.computeStats(fname, cp);

      std::cout << "Statistics on the input quadrangulation has successfully been generated." << std::endl;

      std::cout << std::endl;
    }
  }

  // -----------------------------------------------------------------
  // Method improveMesh()
  //
  // Calls  the methods  "makeTopologicalCorrections" and  "smooth" to
  // improve the quality of the mesh.
  //
  // -----------------------------------------------------------------
  void
  tControl::improveMesh() {
    if (m_mesh && !m_isImproved) {
      struct timeval stime, etime;

      std::cout << "Improving mesh quality by performing topological corrections..." << std::endl;

      // Gets the time before calling the method to improve the mesh
      gettimeofday(&stime,0);

      m_mesh->makeTopologicalCorrections();

      // Gets the time after calling the method to improve the mesh
      gettimeofday(&etime,0);

      long t1 = (1000l * (etime.tv_sec - stime.tv_sec)) 
	+ ((etime.tv_usec - stime.tv_usec) / 1000l);

      unsigned int ntsq = dynamic_cast<tQuadMesh*>(m_mesh)->getNumTSQuads();
      if (ntsq == 1) {
	std::cout << "There was 1 triangular-shaped quadrilateral." << std::endl;
      }
      else if (ntsq > 0) {
	std::cout << "There were " << ntsq << " triangular-shaped quadrilaterals." << std::endl;
      }
      else {
	std::cout << "There was no triangular-shaped quadrilateral." << std::endl;
      }

      std::cout << "Topological corrections performed successfully in " 
		<< t1 << " miliseconds." << std::endl;

      std::cout << std::endl;

      std::cout << "Improving mesh quality by performing angle-based smoothing..." << std::endl;

      // Gets the time before calling the method to improve the mesh
      gettimeofday(&stime,0);

      for (unsigned i=0; i<m_iter; i++) {
	m_mesh->smooth();
      }

      // Gets the time after calling the method to improve the mesh
      gettimeofday(&etime,0);

      long t2 = (1000l * (etime.tv_sec - stime.tv_sec)) + 
	((etime.tv_usec - stime.tv_usec) / 1000l);

      m_isImproved = true;

      cpPointer cp = m_mesh->getCellComplex();

      std::cout << "Angle-based smoothing performed successfully in " 
		<< t2 << " miliseconds." << std::endl;

      std::cout << std::endl;

      std::cout << "Number of vertices  : " << cp->getNumVerts()  << std::endl;
      std::cout << "Number of edges     : " << cp->getNumEdges()  << std::endl;
      std::cout << "Number of faces     : " << cp->getNumFaces()  << std::endl;
      std::cout << "Number of holes     : " << cp->getNumHoles()  << std::endl;
      std::cout << "Number of boundaries: " << cp->getNumBounds() << std::endl;

      std::cout << std::endl;
    }
  }

  // -----------------------------------------------------------------
  // Method saveQuadrangulation()
  //
  // Saves  the  information about  the  quadrilateral  mesh to  files
  // .NODE, .EDGE, .ELE and .POLY.
  //
  // -----------------------------------------------------------------
  void
  tControl::saveQuadrangulation()
  {
    if (m_mesh == 0) return;

    std::cout << "Saving quadrangulation to a file..." << std::endl;

    string fn1 = string(m_fname) + string("-post.node");
    string fn2 = string(m_fname) + string("-post.edge");
    string fn3 = string(m_fname) + string("-post.ele");
    string fn4 = string(m_fname) + string("-post.poly");

    tMeshWriter writer(fn1,fn2,fn3,fn4);

    cpPointer cp = m_mesh->getCellComplex();

    tMeshWriter::VertexSet vset;
    map<hedge::tVertex2*,int> vtmap;
    int cv = 1;
    for (VIT vi = cp->vtBegin(); vi != cp->vtEnd(); vi++) {
      vtPointer v = *vi;
      vtmap.insert(make_pair(v.getPointee(),cv));
      vset.insert(make_pair(cv,Point2(v->getPoint())));
      cv++;
    }

    tMeshWriter::EdgeSet eset;
    for (EIT ei = cp->egBegin(); ei != cp->egEnd(); ei++) {
      egPointer e = *ei;
      vtPointer v1 = e->getHalfEdge()->getVertex();
      vtPointer v2 = e->getHalfEdge()->getNext()->getVertex();
      bool econst = e->isConstrained();

      map<hedge::tVertex2*,int>::iterator p1 = vtmap.find(v1.getPointee());
      map<hedge::tVertex2*,int>::iterator p2 = vtmap.find(v2.getPointee());

      tEdgeKey ce(p1->second,p2->second);
      if (eset.find(ce.reverse()) == eset.end()) {
	eset.insert(make_pair(ce,econst));
      }
    }

    tMeshWriter::FaceSet fset;
    for (FIT fi = cp->fcBegin(); fi != cp->fcEnd(); ++fi) {
      fcPointer f = *fi;

      hePointer he = f->getHalfEdge();
      vtPointer v1 = he->getVertex();
      vtPointer v2 = he->getNext()->getVertex();
      vtPointer v3 = he->getNext()->getNext()->getVertex();
      vtPointer v4 = he->getPrev()->getVertex();

      tPolygon2 p(4);
      map<hedge::tVertex2*,int>::iterator p1 = vtmap.find(v1.getPointee());
      map<hedge::tVertex2*,int>::iterator p2 = vtmap.find(v2.getPointee());
      map<hedge::tVertex2*,int>::iterator p3 = vtmap.find(v3.getPointee());
      map<hedge::tVertex2*,int>::iterator p4 = vtmap.find(v4.getPointee());

      p.insert(0,p1->second,Point2(v1->getPoint()));
      p.insert(1,p2->second,Point2(v2->getPoint()));
      p.insert(2,p3->second,Point2(v3->getPoint()));
      p.insert(3,p4->second,Point2(v4->getPoint()));

      fset.push_back(p);
    }

    try {
      writer.writeVertexFile(vset);
      writer.writeEdgeFile(eset);
      writer.writeElemFile(fset);
      writer.writePolyFile(vset,eset);
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    std::cout << "Quadrangulation has successfully been saved to a file." << std::endl;

    std::cout << std::endl;
  }

  // -----------------------------------------------------------------
  // Method generateStats()
  //
  // 
  //
  // -----------------------------------------------------------------
  void
  tControl::generateStats()
  {
    if ((m_mesh != 0) &&  m_isImproved) {
      std::cout << "Generating statistics on the output quadrangulation..." << std::endl;

      cpPointer cp = m_mesh->getCellComplex();

      string fname = m_fname + "-post.stat";

      tQuadStat stat;
      stat.computeStats(fname, cp);

      std::cout << "Statistics on the output quadrangulation has successfully been generated." << std::endl;
      
      std::cout << std::endl << std::endl;
    }
  }

};   // end namespace qmpp
