//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    mesh2.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
// ===================================================================

#include "mesh2.h"

#include <gmp.h>     // for all GMP and GMP++ functions
#include <gmpxx.h>


namespace qmpp {

  // -----------------------------------------------------------------
  // Method: tMesh()
  // 
  // Default constructor
  //
  // -----------------------------------------------------------------
  tMesh::tMesh()
  {
    m_cp =  hedge::tComplex2::New();
  }

  // -----------------------------------------------------------------
  // Method findVertexInBound()
  //
  // Returns true if a given vertex  belongs to the same boundary as a
  // given edge. Otherwise, it returns false.
  //
  // -----------------------------------------------------------------
  bool
  tMesh::findVertexInBound(
			   int cv, 
			   const tEdgeKey& ce
			   )
    const
  {
    VertexSet::const_iterator vspos;

    vspos = vs.find(cv);
    if (vspos != vs.end()) {
      hePointer he1 = (es.find(ce))->second;
      hePointer he2;
      while ((vspos != vs.end()) && (vspos->first == cv)) {
	he2 = (vspos->second)->getBoundHalfEdge();
	if (he1->getEdge()->getBoundary() == he2->getEdge()->getBoundary()) {
	  return true;
	}
	else {
	  ++vspos;
	}
      }
    }

    return false;
  }

  // -----------------------------------------------------------------
  // Method IsConstrainedEdge()
  //
  // It returns true if a given edge is constrained, and returns 
  // false if it is not.
  //
  // -----------------------------------------------------------------
  bool
  tMesh::isConstrainedEdge(
			   const tEdgeKey& ce,
			   const ConstrainedEdgeSet& ledges
			   )
    const
  {
    int cv1 = ce.getEdgeC1();
    int cv2 = ce.getEdgeC2();

    ConstrainedEdgeSet::const_iterator epos;
    epos = ledges.find(cv1);
    while ((epos != ledges.end()) && (epos->first == cv1)) {
      if (epos->second == cv2) return true;
      epos++;
    }

    epos = ledges.find(cv2);
    while ((epos != ledges.end()) && (epos->first == cv2)) {
      if (epos->second == cv1) return true;
      epos++;
    }

    return false;
  }

  // -----------------------------------------------------------------
  // Method insertGlue()
  //
  // Inserts  two  half-edges  into  the  list  of  half-edges  to  be
  // identified with each other.
  //
  // -----------------------------------------------------------------
  void
  tMesh::insertGlue(
		    const hePointer& he1,
		    const hePointer& he2,
		    GlueEdgeSet& gset
		    )
  {
    tGlueEdge ge;
    
    ge.m_he1 = he1;
    ge.m_he2 = he2;
    gset.push_back(ge);
  }

  // -----------------------------------------------------------------
  // Method glueFaces()
  //
  // Attaches one  face to another along  a common edge.  If the edges
  // are in  the same boundary, the  glueing might create  a genus. If
  // they belong to distinct connected components, the componentes are
  // merged together.
  //
  // -----------------------------------------------------------------
  void
  tMesh::glueFaces(
		   const hePointer& he1,
		   const hePointer& he2
		   )
  {
    // Gets the boundaries containing the half-edges
    bdPointer bd1 = he1->getEdge()->getBoundary();
    bdPointer bd2 = he2->getEdge()->getBoundary();

    // If the  boundaries are distinct,  we merge them.  Otherwise, we
    // will make a genus.
    if (bd1 != bd2) {
      m_cp->rmBEV(he1,he2);
    }
    else {
      m_cp->mkGrmEV(he1,he2);
    }
  }

  // -----------------------------------------------------------------
  // Method isPointOnLine()
  //
  // Returns true if a given point lies on a given line. Otherwise, it
  // returns false.
  //
  // -----------------------------------------------------------------
  bool
  tMesh::isPointOnLine(
		       const Point2& p1,
		       const Point2& p2,
		       const Point2& v
		       )
  {
    if ((p1.x() == v.x()) && (p1.y() == v.y())) return true;
    if ((p2.x() == v.x()) && (p2.y() == v.y())) return true;

    Line2 l(p1,p2);
    return l.oriented_side(v) == CGAL::ON_ORIENTED_BOUNDARY;
  }

  // -----------------------------------------------------------------
  // Method isPointOnLine()
  //
  // Returns true if a given  segment lies on a given line. Otherwise,
  // it returns false.
  //
  // -----------------------------------------------------------------
  bool
  tMesh::isSegInLine(
		     const Point2& p1,
		     const Point2& p2,
		     const Point2& v1,
		     const Point2& v2
		     )
  {
    bool res1 = isPointOnLine(p1,p2,v1);
    bool res2 = isPointOnLine(p1,p2,v2);
    
    return res1 && res2;
  }

  // -----------------------------------------------------------------
  // Method segLineIntersection()
  //
  // Computes  the intersection  point of  a segment  and a  line, and
  // returns true if such a point exists. Otherwise, returns false.
  //
  // -----------------------------------------------------------------
  bool
  tMesh::segLineIntersection(
			     const Point2& p1,
			     const Point2& p2,
			     const Point2& v1,
			     const Point2& v2,
			     Point2& pt
			     )
  {
    // if intersection is the segment v1v2 itself then return false
    if ( isSegInLine( p1 , p2 , v1 , v2 ) ) return false ;

    // test if intersection is the point "v1"
    if ( isPointOnLine( p1 , p2 , v1 ) ) {
      pt = v1 ;
      return true ;
    }
    
    // test if intersection is the point "v2"
    if ( isPointOnLine( p1 , p2 , v2 ) ) {
      pt = v2 ;
      return true ;
    }

    // find intersection point
    mpf_class a( p2.x() - p1.x() , sizeof( double) * 8 ) ;
    mpf_class b( v2.x() - v1.x() , sizeof( double) * 8 ) ;
    mpf_class c( v1.x() - p1.x() , sizeof( double) * 8 ) ;
    mpf_class d( p2.y() - p1.y() , sizeof( double) * 8 ) ;
    mpf_class e( v2.y() - v1.y() , sizeof( double) * 8 ) ;
    mpf_class f( v1.y() - p1.y() , sizeof( double) * 8 ) ;
    
    mpf_class g( 0 , sizeof( double) * 128 ) ;
    
    g = ( d * b ) - ( e * a ) ;
    
    mpf_class zero( 0 , sizeof( double) * 128 ) ;
    
    // if line and segment are parallel then there is no intersection
    if ( cmp( g , zero ) == 0 ) {
      return false ;
    }
    
    mpf_class t( 0 , sizeof( double) * 128 ) ;
    
    t = ( (f * a) - (d * c) ) / g ;
    
    mpf_class one( 1 , sizeof( double) * 128 ) ;
    
    if ( ( cmp(t , zero) < 0 ) || ( cmp(t , one ) > 0 ) ) {
      return false ;
    }
    
    double dt = t.get_d() ; 
    double db = b.get_d() ;
    double de = e.get_d() ;
    
    pt = Point2( v1.x() + db * dt , v1.y() + de * dt ) ;
    
    if ( IS_EQUAL( pt.x() , v1.x() ) && IS_EQUAL( pt.y() , v1.y() ) ) {
      pt = v1 ;
      return true ;
    }
    
    if ( IS_EQUAL( pt.x() , v2.x() ) && IS_EQUAL( pt.y() , v2.y() ) ) {
      pt = v2 ;
      return true ;
    }
    
    return true ;
  }

  // -----------------------------------------------------------------
  // Method myAssign()
  //
  // Gets the intersection point of a line and a segment, if any.
  //
  // -----------------------------------------------------------------
  bool
  tMesh::myAssign(
		  Point2& p ,
		  const Segment2& s ,
		  const Line2& l
		 ) 
  {
    if (  !assign( p , CGAL::intersection( s , l ) ) ) {
      Point2 p1 = s.vertex( 0 ) ;
      Point2 p2 = s.vertex( 1 ) ;
      Point2 q1 = l.point( 0 ) ;
      Point2 q2 = l.point( 1 ) ;

      return segLineIntersection( q1 , q2 , p1 , p2 , p ) ;
    }

    return true ;
  }


  // -------------------------------------------------------------------
  // Method halfSpaceIntersection()
  //
  // Computes the region resulting  from the intersection of a polygon
  // and a half-space.
  //
  // -------------------------------------------------------------------
  void 
  tMesh::halfSpaceIntersection(
			       const list<Point2>& li,
			       const Point2& p1,
			       const Point2& p2,
			       list<Point2>& lo,
			       CGAL::Oriented_side orient
			       )
    throw (tExceptionObject)
  {
    // create an oriented line "l" from "p1" to "p2"
    Line2 l(p1,p2);
  
    // create an iterator for list of vertices of the polygon
    list<Point2>::const_iterator pi = li.begin();
  
    // get the first vertex of the polygon and advance iterator
    Point2 v1 = *pi;
    Point2 vf = v1;
    ++pi;

    // find out  position of  "v1" with respect  to the line  "l" from
    // "p1" to  "p2". Variable  "side1" will hold  true if "v1"  is on
    // line "l" or on its left.
    CGAL::Oriented_side result = l.oriented_side(v1);
    bool side1 = (result == CGAL::ON_ORIENTED_BOUNDARY) || 
      (result != orient) || (v1 == p1) || (v1 == p2);
    bool sidef = side1;
  
    // loop over all segments except for the last one
    for (; pi != li.end(); ++pi) {
      // get next vertex of the input polygon 
      Point2 v2 = *pi;

      // find out position  of "v2" with respect to  the line "l" from
      // "p1" to "p2".  Variable "side2" will hold true  if "v2" is on
      // line "l" or on its left.
      result = l.oriented_side(v2);
      bool side2 = (result == CGAL::ON_ORIENTED_BOUNDARY) || 
	(result != orient) || (v2 == p1) || (v2 == p2);
      
      // if "v1" and "v2" are on  distinct sides of "l", then line "l"
      // intersects the segment "v1v2".
      Point2 p;
      if (side1 != side2) {
	// compute  intersection point  between line  "l"  and segment
	// "v1v2".

	if (!segLineIntersection(p1,p2,v1,v2,p)) {
	  stringstream ss (stringstream::in | stringstream::out);
	  ss << "halfSpaceIntersection(): cannot recover from numerical instability";
	  throw tExceptionObject(__FILE__,__LINE__,ss.str());
	}
	else {
	  if ((!IS_EQUAL(v1.x(),p.x()) || !IS_EQUAL(v1.y(),p.y())) &&
	      (!IS_EQUAL(v2.x(),p.x()) || !IS_EQUAL(v2.y(),p.y()))) {
	    lo.push_back(p);
	  }
	  else if ((v2.x() == p.x()) && (v2.y() == p.y()) && !side2) {
	    side2 = true;
	  }
	}
      }

      // insert "v2" inside output list  of vertices if "v2" is on "l"
      // or on its left side.
      if (side2) lo.push_back(v2);

      // update "v1" and "side1" for the next loop iteration
      v1 = v2;
      side1 = side2;
    }

    // if the  last and first vertices  are on distinct  sides of "l",
    // then line "l" intersects the segment defined by them.
    if (side1 != sidef) {
      Point2 p;

      if (!segLineIntersection(p1,p2,v1,vf,p)) {
	stringstream ss (stringstream::in | stringstream::out);
	ss << "halfSpaceIntersection(): cannot recover from numerical instability";
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }
      else {
	if ((!IS_EQUAL(v1.x(),p.x()) || !IS_EQUAL(v1.y(),p.y())) &&
	    (!IS_EQUAL(vf.x(),p.x()) || !IS_EQUAL(vf.y(),p.y()))) {
	  lo.push_back(p);
	}
	else if ((vf.x() == p.x()) && (vf.y() == p.y()) && !sidef) {
	  sidef = true;
	}
      }
    }

    // if first vertex  is on line "l" or on its  left side, we insert
    // it into the output list of vertices.
    if (sidef) {
      lo.push_back(vf);
    }
  }

};     // end namespace qmpp
