//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    mesh2.h
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#ifndef __mesh2_h
#define __mesh2_h

#include <string>
#include <list>
#include <map>

#include "typedefs.h"
#include "macros.h"
#include "smartptr.h"
#include "dataobj.h"
#include "exceptionobj.h"

#include "edgekey.h"
#include "polygon2.h"
#include "hedge.h"

namespace qmpp {

  // -----------------------------------------------------------------
  // class tGlueEdge
  //
  // This class  represents a pair  of half-edges that  are identified
  // with each other when cell complex faces are glued together.
  //
  // -----------------------------------------------------------------
  class tGlueEdge {
  public:
    // Type definition
    typedef hedge::tHalfEdge2::Pointer hePointer;

    // Default constructor
    tGlueEdge() 
      {
	m_he1 = hePointer(0);
	m_he2 = hePointer(0);
      }

    // Destructor
    ~tGlueEdge() {}

    //
    // Public attributes
    //
    hePointer m_he1;
    hePointer m_he2;
  };


  // -----------------------------------------------------------------
  // class tMesh
  //
  // This is an  abstract class to build a mesh and  store it into the
  // handle-edge data structure.
  //
  // -----------------------------------------------------------------
  class tMesh {
  public:
    // Type definitions
    typedef hedge::tVertex2::Pointer     vtPointer;
    typedef hedge::tHalfEdge2::Pointer   hePointer;
    typedef hedge::tEdge2::Pointer       egPointer;
    typedef hedge::tFace2::Pointer       fcPointer;
    typedef hedge::tBoundary2::Pointer   bdPointer;
    typedef hedge::tComplex2::Pointer    cpPointer;

    typedef multimap<int,vtPointer>      VertexSet;
    typedef multimap<tEdgeKey,hePointer> EdgeSet;

    typedef vector<Point2>     NodeSet;
    typedef multimap<int,int>  ConstrainedEdgeSet;
    typedef list<tPolygon2>    FaceSet;
    typedef list<tGlueEdge>    GlueEdgeSet;

    // Default constructor
    tMesh();

    // Destructor
    virtual ~tMesh() {}

    // Get method
    cpPointer getCellComplex() const
      { return m_cp; }

    // Pure virtual method 
    virtual cpPointer buildMesh(const string&) = 0;

    // Virtual methods to improve mesh quality
    virtual void makeTopologicalCorrections() {};
    virtual void smooth() {};

  protected:
    // Returns true if a given  vertex belongs to the same boundary as
    // a given edge. Otherwise, it returns false.
    bool findVertexInBound(int, const tEdgeKey&) const;

    // Returns true if a given  edge is constrained, and returns false
    // if it is not.
    bool isConstrainedEdge(const tEdgeKey&, const ConstrainedEdgeSet&) 
      const;

    // Inserts  two  half-edges into  the  list  of  half-edges to  be
    // identified with each other.
    void insertGlue(const hePointer&, const hePointer&, GlueEdgeSet&);

    // Attaches one face to another  along a common edge. If the edges
    // are in the same boundary,  the glueing might create a genus. If
    // they belong  to distinct connected  components, the componentes
    // are merged together.
    void glueFaces(const hePointer&, const hePointer&);

        // Mathematical  routines that  can be  used by  the  classes that
    // perform topological corrections and smoothing on the mesh.
    bool isPointOnLine(const Point2&, const Point2&, const Point2&);

    // Returns   true   if  a   given   segment   lies   on  a   given
    // line. Otherwise, it returns false.
    bool isSegInLine(const Point2&, const Point2&, const Point2&,
		     const Point2&);

    // Computes the  intersection point of  a segment and a  line, and
    // returns true if such a point exists. Otherwise, returns false.
    bool segLineIntersection(const Point2&, const Point2&, 
			     const Point2&, const Point2&, Point2&);

    // Gets the intersection point of a line and a segment, if any.
    bool myAssign(Point2&, const Segment2&, const Line2&);

    //
    void halfSpaceIntersection(const list<Point2>&, const Point2&,
			       const Point2&, list<Point2>&,
			       CGAL::Oriented_side) 
      throw (tExceptionObject);

    //
    // Private Attributes
    //

    // Temporary set  of vertices used by buildMesh()  method to store
    // boundary vertices during mesh construction.
    VertexSet vs;

    // Temporary set  of vertices used by buildMesh()  method to store
    // boundary vertices during mesh construction.
    EdgeSet es;

    // Smart pointer to the cell complex built by buildMesh()  method
    cpPointer m_cp;

  };


};   // end namespace qmpp

#endif   // __mesh2_h
