//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    quadmesh2.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
// ===================================================================

#include "quadmesh2.h"

namespace qmpp {

  // -----------------------------------------------------------------
  // Method buildQuadMesh()
  //
  // Stores  a  quadrilateral  mesh  into the  data  structure  handle
  // edge. All  it needs  is the collection  of quadrilaterals  of the
  // mesh in  any given  order. Quadrilaterals should  consistently be
  // oriented.
  //
  // -----------------------------------------------------------------
  tMesh::cpPointer
  tQuadMesh::buildMesh(const std::string& name) 
    throw (tExceptionObject)
  {
    // Reads  files   with  information  about   vertices,  edges  and
    // triangles of the triangular mesh.
    std::string infile1 = std::string(name) + std::string(".node");
    std::string infile2 = std::string(name) + std::string(".edge");
    std::string infile3 = std::string(name) + std::string(".ele");

    // Creates file reader
    tMeshReader reader(infile1, infile2, infile3);

    // Reads files
    NodeSet lv;
    ConstrainedEdgeSet le;
    FaceSet lf;

    try {
      reader.readVertexFile(lv);
      reader.readEdgeFile(le);
      reader.readElemFile(lv,lf);
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    // We do not need the set  of nodes anymore as node information is
    // also in the face set "lf".
    lv.clear();

    // First stage: making faces
    GlueEdgeSet gluelist;
    typedef FaceSet::const_iterator LFI;

    for (LFI fiter = lf.begin(); fiter != lf.end(); ++fiter) {
      // Gets face from iterator
      tQuadrilateral2 f(*fiter);

      // Checks out if face is simple
      if (!f.isSimple()) {
	std::stringstream ss (std::stringstream::in | std::stringstream::out);
	ss << "Quadrilateral is not simple";
	throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      // Checks out orientation
      if (!f.isCCW()) {
	f.reverse();
      }

      // Creates vertex codes for every vertex of face f
      int cv1 = f.getVertexId(0);
      int cv2 = f.getVertexId(1);
      int cv3 = f.getVertexId(2);
      int cv4 = f.getVertexId(3);

      // Gets vertex locations
      Point2 lv1 = f.getPoint(0);
      Point2 lv2 = f.getPoint(1);
      Point2 lv3 = f.getPoint(2);
      Point2 lv4 = f.getPoint(3);

      // Creates edge codes for every edge of face f
      tEdgeKey ce1 = tEdgeKey(cv1,cv2);
      tEdgeKey ce2 = tEdgeKey(cv2,cv3);
      tEdgeKey ce3 = tEdgeKey(cv3,cv4);
      tEdgeKey ce4 = tEdgeKey(cv4,cv1);

      // Searches for existing edges in the boundary edge set
      int foundedge = 0;
      if (es.find(ce1) != es.end()) foundedge += 1;
      if (es.find(ce2) != es.end()) foundedge += 2;
      if (es.find(ce3) != es.end()) foundedge += 4;
      if (es.find(ce4) != es.end()) foundedge += 8;

      // A new  face is created only if  one of its edges  is found in
      // the set of constrained edges  and if the attaching of the new
      // face does not create a singular vertex.
      if ((foundedge == 1) || (foundedge == 2) || (foundedge == 4) || 
	  (foundedge == 8)) {
	int cvaux1, cvaux2;
	Point2 lvaux1, lvaux2;
	tEdgeKey ceaux1, ceaux2, ceaux3, ceaux4;

	if (foundedge == 1) {
	  cvaux1 = cv3;
	  lvaux1 = lv3;
	  cvaux2 = cv4;
	  lvaux2 = lv4;

	  ceaux1 = ce2;
	  ceaux2 = ce3; 
	  ceaux3 = ce4;
	  ceaux4 = ce1;
	}
	else if (foundedge == 2) {
	  cvaux1 = cv4;
	  lvaux1 = lv4;
	  cvaux2 = cv1;
	  lvaux2 = lv1;

	  ceaux1 = ce3;
	  ceaux2 = ce4;
	  ceaux3 = ce1;
	  ceaux4 = ce2;
	}
	else if (foundedge == 4) {
	  cvaux1 = cv1;
	  lvaux1 = lv1;
	  cvaux2 = cv2;
	  lvaux2 = lv2;

	  ceaux1 = ce4;
	  ceaux2 = ce1;
	  ceaux3 = ce2;
	  ceaux4 = ce3;
	}
	else {  // if (foundedge == 8)
	  cvaux1 = cv2;
	  lvaux1 = lv2;
	  cvaux2 = cv3;
	  lvaux2 = lv3;

	  ceaux1 = ce1;
	  ceaux2 = ce2;
	  ceaux3 = ce3;
	  ceaux4 = ce4;
	}

	// If  vertices "cvaux1"  and  "cvaux2" are  not  in the  same
	// boundary as edge "ceaux4.reverse()", we can attach face "f"
	// to the boundary containing "ceaux4.reverse()".
	if ((!findVertexInBound(cvaux1,ceaux4)) && 
	    (!findVertexInBound(cvaux2,ceaux4))) {

	  // Either "cvaux1" and "cvaux2" have not been created yet or
	  // they are not on the same boundary as "ceaux4.reverse()".
	  this->makeQuadFace(f,cvaux1,lvaux1,cvaux2,lvaux2,ceaux4,
	    ceaux1,ceaux2,ceaux3,le);
	}
	else {
	  // Creates an isolated face, which should be attached to "f"
	  // in the second stage.

	  this->makeQuadFaceGlue1(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,
	    foundedge,gluelist,le);
	}
      }
      else if ((foundedge ==  3) || (foundedge ==  5) || 
	       (foundedge ==  9) || (foundedge ==  6) ||
	       (foundedge == 10) || (foundedge == 12)) {
	// Face  "f" has  two edges  in common  with the  current cell
	// complex. So, we make "f" an isolated face, and then we mark
	// the common edges as edges to be glued later.
	this->makeQuadFaceGlue2(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,
	  ce1,ce2,ce3,ce4,foundedge,gluelist,le);
      }
      else if ((foundedge ==  7) || (foundedge == 11) || 
	       (foundedge == 13) || (foundedge == 14)) {
	// New face  has three edges  in common with the  current cell
	// complex. So, we make "f" an isolated face, and then we mark
	// the common edges as edges to be glued later.
	this->makeQuadFaceGlue3(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,
	  ce1,ce2,ce3,ce4,foundedge,gluelist,le);
      }
      else if (foundedge != 0) {
	// New face  has four  edges in common  with the  current cell
	// complex. So, we make "f" an isolated face, and then we mark
	// the common edges as edges to be glued later.
	this->makeQuadFaceGlue4(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,
	  ce1,ce2,ce3,ce4,gluelist,le);
      }
      else { // foundedge == 0
	// New face  has no  vertex in common  with existent  faces or
	// one, two,  three or four  vertices in common  with existent
	// faces. We will make an isolated faces in any case.

	this->makeQuadFace(f,cv1,cv2,cv3,cv4,lv1,lv2,lv3,lv4,ce1,ce2,
			   ce3,ce4,le);
      }
    }

    // Removes all elements from vertex and edge boundary sets
    vs.clear();
    es.clear();

    // Removes all elements from constrained edge set

    // Second stage: glueing faces
    typedef GlueEdgeSet::iterator LGI;
    LGI giter = gluelist.begin();
    while (giter != gluelist.end()) {
      // Gets the half-edges to be identified in the faces to be glued.
      hePointer he1 = (*giter).m_he1;
      hePointer he2 = (*giter).m_he2;

      // Glues faces
      glueFaces(he1,he2);

      // Removes the half-edges from the glue list
      giter = gluelist.erase(giter);
    }

#ifdef DEBUG_MODE
    // Verifies mesh consistency
    std::cout << "Checking mesh consistency..." << std::endl;
    if (!checkMeshConsistency()) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Quadrilateral mesh is not consistent";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    std::cout << "Quadrilateral mesh is consistent." << std::endl;
#endif

    // Returns complex
    return m_cp;
  }

  // -----------------------------------------------------------------
  // Method makeQuadFace()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFace(
			const tQuadrilateral2& f,
			int cv1,
			int cv2,
			int cv3,
			int cv4,
			const Point2& p1,
			const Point2& p2,
			const Point2& p3,
			const Point2& p4,
			const tEdgeKey& ce1,
			const tEdgeKey& ce2,
			const tEdgeKey& ce3,
			const tEdgeKey& ce4,
			const ConstrainedEdgeSet& ledges
			)
  {
    // Creates all three vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
 
    // Inserts the code of the new vertices into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));

    // Inserts new edge codes into boundary edge set
    es.insert(make_pair(ce1.reverse(),he1));
    es.insert(make_pair(ce2.reverse(),he2));
    es.insert(make_pair(ce3.reverse(),he3));
    es.insert(make_pair(ce4.reverse(),he4));

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,ledges)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeQuadFace()
  //
  // Creates a new triangular face  and attaches it to another face in
  // the mesh.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFace(
			const tQuadrilateral2& f,
			int cv1,
			const Point2& p1,
			int cv2,
			const Point2& p2,
			const tEdgeKey& ce,
			const tEdgeKey& ce2,
			const tEdgeKey& ce3,
			const tEdgeKey& ce4,
			const ConstrainedEdgeSet& le
			)
  {
    // Creates two new vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);

    // Gets first half-edge of the commom edge
    hePointer he = (es.find(ce))->second;

    // Makes a new face with a commom edge and a new edge
    egPointer e = ((m_cp->mkFE(he))->getHalfEdge())->getEdge();

    // Gets first half-edge of the new edge
    hePointer he2 = e->getHalfEdge();

    // Creates  edge following  the  common edge  and  gets its  first
    // half-edge.
    hePointer he3 = (m_cp->mkEV(v1,he2))->getHalfEdge();

    // Creates  edge preceding  the  common edge  and  gets its  first
    // half-edge.
    hePointer he4 = (m_cp->mkEV(v2,he3))->getHalfEdge();

    // Inserts the code of the new vertices into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));

    // Inserts new edge codes into boundary edge set
    es.insert(make_pair(ce2.reverse(),he2));
    es.insert(make_pair(ce3.reverse(),he3));
    es.insert(make_pair(ce4.reverse(),he4));

    // Removes commom edge from the boundary edge set
    es.erase(es.find(ce));

    // Sets status of constrained edges
    if (isConstrainedEdge(ce2,le)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,le)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,le)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeTriFaceGlue1()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh and keeps track of  its common edge with another face of
  // the mesh, so that we can glue this connected component later.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh:: makeQuadFaceGlue1(
				const tQuadrilateral2& f,
				int cv1,
				int cv2,
				int cv3,
				int cv4,
				const Point2& p1,
				const Point2& p2,
				const Point2& p3,
				const Point2& p4,
				int cedge,
				GlueEdgeSet& gset,
				const ConstrainedEdgeSet& ledges
			      )
  {
    // Creates all four vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);
    
    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
    
    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));

    EdgeSet::iterator hepos;
    tEdgeKey ce1(cv2,cv1);
    tEdgeKey ce2(cv3,cv2);
    tEdgeKey ce3(cv4,cv3);
    tEdgeKey ce4(cv1,cv4);
    if (cedge == 1) {
      // Inserts new edge codes into boundary edge set
      es.insert(make_pair(ce2,he2));
      es.insert(make_pair(ce3,he3));
      es.insert(make_pair(ce4,he4));

      // keeps track of edges to be glued at second stage
      hepos = es.find(ce1.reverse());
      insertGlue(hepos->second,he1,gset);
    }
    else if (cedge == 2) {
      // Inserts new edge codes into boundary edge set
      es.insert(make_pair(ce1,he1));
      es.insert(make_pair(ce3,he3));
      es.insert(make_pair(ce4,he4));

      // keeps track of edges to be glued at second stage
      hepos = es.find(ce2.reverse());
      insertGlue(hepos->second,he2,gset);
    }
    else if (cedge == 4) {
      // Inserts new edge codes into boundary edge set
      es.insert(make_pair(ce1,he1));
      es.insert(make_pair(ce2,he2));
      es.insert(make_pair(ce4,he4));

      // keeps track of edges to be glued at second stage
      hepos = es.find(ce3.reverse());
      insertGlue(hepos->second,he3,gset);
    }
    else {
      // Inserts new edge codes into boundary edge set
      es.insert(make_pair(ce1,he1));
      es.insert(make_pair(ce2,he2));
      es.insert(make_pair(ce3,he3));

      // keeps track of edges to be glued at second stage
      hepos = es.find(ce4.reverse());
      insertGlue(hepos->second,he4,gset);
    }
  
    // Removes commom edge from the boundary edge set
    es.erase(hepos);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1.reverse(),ledges)) {
      he1->getEdge()->constraintOn();
    }
    if (isConstrainedEdge(ce2.reverse(),ledges)) {
      he2->getEdge()->constraintOn();
    }
    if (isConstrainedEdge(ce3.reverse(),ledges)) {
      he3->getEdge()->constraintOn();
    }
    if (isConstrainedEdge(ce4.reverse(),ledges)) {
      he4->getEdge()->constraintOn();
    }
  }

  // -----------------------------------------------------------------
  // Method makeQuadFaceGlue2()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh and keeps track of its two common edges with other faces
  // of the mesh, so that we can glue this connected component later.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFaceGlue2(
			       const tQuadrilateral2& f,
			       int cv1,
			       int cv2,
			       int cv3,
			       int cv4,
			       const Point2& p1,
			       const Point2& p2,
			       const Point2& p3,
			       const Point2& p4,
			       const tEdgeKey& ce1,
			       const tEdgeKey& ce2,
			       const tEdgeKey& ce3,
			       const tEdgeKey& ce4,
			       int ne,
			       GlueEdgeSet& gset,
			       const ConstrainedEdgeSet& ledges
			       )
  {
    // Creates all four vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
    
    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));
 
    // Inserts new edge code into boundary edge set and keeps track of
    // the two edges to be glued in the second stage.
    EdgeSet::iterator hepos1, hepos2;
    if (ne == 3) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce2);
      es.insert(make_pair(ce3.reverse(),he3));
      es.insert(make_pair(ce4.reverse(),he4));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he2,gset);
    }
    else if (ne == 5) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce3);
      es.insert(make_pair(ce2.reverse(),he2));
      es.insert(make_pair(ce4.reverse(),he4));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he3,gset);
    }
    else if (ne == 9) {
      hepos1 = es.find(ce4);
      hepos2 = es.find(ce1);
      es.insert(make_pair(ce2.reverse(),he2));
      es.insert(make_pair(ce3.reverse(),he3));
      insertGlue(hepos1->second,he4,gset);
      insertGlue(hepos2->second,he1,gset);
    }
    else if (ne == 6) {
      hepos1 = es.find(ce2);
      hepos2 = es.find(ce3);
      es.insert(make_pair(ce1.reverse(),he1));
      es.insert(make_pair(ce4.reverse(),he4));
      insertGlue(hepos1->second,he2,gset);
      insertGlue(hepos2->second,he3,gset);
    }
    else if (ne == 10) {
      hepos1 = es.find(ce2);
      hepos2 = es.find(ce4);
      es.insert(make_pair(ce1.reverse(),he1));
      es.insert(make_pair(ce3.reverse(),he3));
      insertGlue(hepos1->second,he2,gset);
      insertGlue(hepos2->second,he4,gset);
    }
    else {  // if (ne == 12) {
      hepos1 = es.find(ce3);
      hepos2 = es.find(ce4);
      es.insert(make_pair(ce1.reverse(),he1));
      es.insert(make_pair(ce2.reverse(),he2));
      insertGlue(hepos1->second,he3,gset);
      insertGlue(hepos2->second,he4,gset);
    }
    
    // Removes commom edges from the boundary edge set
    es.erase(hepos1);
    es.erase(hepos2);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,ledges)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeQuadFaceGlue3()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh  and keeps  track of its  three common edges  with other
  // faces of the  mesh, so that we can  glue this connected component
  // later.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFaceGlue3(
			       const tQuadrilateral2& f,
			       int cv1,
			       int cv2,
			       int cv3,
			       int cv4,
			       const Point2& p1,
			       const Point2& p2,
			       const Point2& p3,
			       const Point2& p4,
			       const tEdgeKey& ce1,
			       const tEdgeKey& ce2,
			       const tEdgeKey& ce3,
			       const tEdgeKey& ce4,
			       int ne,
			       GlueEdgeSet& gset,
			       const ConstrainedEdgeSet& ledges
			       )
  {
    // Creates all four vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
    
    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));

    // Inserts new edge code into boundary edge set
    EdgeSet::iterator hepos1, hepos2, hepos3;
    if (ne == 7) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce2);
      hepos3 = es.find(ce3);
      es.insert(make_pair(ce4.reverse(),he4));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he2,gset);
      insertGlue(hepos3->second,he3,gset);
    }
    else if (ne == 11) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce2);
      hepos3 = es.find(ce4);
      es.insert(make_pair(ce3.reverse(),he3));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he2,gset);
      insertGlue(hepos3->second,he4,gset);
    }
    else if (ne == 13) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce3);
      hepos3 = es.find(ce4);
      es.insert(make_pair(ce2.reverse(),he2));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he3,gset);
      insertGlue(hepos3->second,he4,gset);
    }
    else {  // if (ne == 14) {
      hepos1 = es.find(ce2);
      hepos2 = es.find(ce3);
      hepos3 = es.find(ce4);
      es.insert(make_pair(ce1.reverse(),he1));
      insertGlue(hepos1->second,he2,gset);
      insertGlue(hepos2->second,he3,gset);
      insertGlue(hepos3->second,he4,gset);
    }

    // Removes commom edges from the boundary edge set
    es.erase(hepos1);
    es.erase(hepos2);
    es.erase(hepos3);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,ledges)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeQuadFaceGlue4()
  //
  // Creates a new quadrilateral face  as a new connected component of
  // the mesh  and keeps  track of its  three common edges  with other
  // faces of the  mesh, so that we can  glue this connected component
  // later.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeQuadFaceGlue4(
			       const tQuadrilateral2& f,
			       int cv1,
			       int cv2,
			       int cv3,
			       int cv4,
			       const Point2& p1,
			       const Point2& p2,
			       const Point2& p3,
			       const Point2& p4,
			       const tEdgeKey& ce1,
			       const tEdgeKey& ce2,
			       const tEdgeKey& ce3,
			       const tEdgeKey& ce4,
			       GlueEdgeSet& gset,
			       const ConstrainedEdgeSet& ledges
			       )
  {
    // Creates all four vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);
    vtPointer v4 = m_cp->createVertex(p4);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();
    hePointer he4 = (m_cp->mkEV(v4,he3))->getHalfEdge();
    
    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));
    vs.insert(make_pair(cv4,v4));

    // Keeps track of edges to be glued in the second stage
    EdgeSet::iterator hepos1, hepos2, hepos3, hepos4;
    hepos1 = es.find(ce1);
    hepos2 = es.find(ce2);
    hepos3 = es.find(ce3);
    hepos4 = es.find(ce4);

    // Inserts new edge code into boundary edge set and keeps track of
    // the two edges to be glued in the second stage.
    insertGlue(hepos1->second,he1,gset);
    insertGlue(hepos2->second,he2,gset);
    insertGlue(hepos3->second,he3,gset);
    insertGlue(hepos4->second,he4,gset);

    // Removes commom edges from the boundary edge set
    es.erase(hepos1);
    es.erase(hepos2);
    es.erase(hepos3);
    es.erase(hepos4);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
    if (isConstrainedEdge(ce4,ledges)) he4->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method checkFaceConsistency()
  //
  // Checks  if  a  given  quadrilateral  face  is  geometrically  and
  // topologically consistent.
  //
  // -----------------------------------------------------------------
  bool
  tQuadMesh::checkFaceConsistency(const fcPointer& f) const
  {
    // Checks face topology
    if (f->getHalfEdge()->getFace() != f) return false;

    hePointer he1 = f->getHalfEdge();
    hePointer he2 = he1->getNext();
    hePointer he3 = he2->getNext();
    hePointer he4 = he3->getNext();

    if (he2->getFace() != f) return false;
    if (he3->getFace() != f) return false;
    if (he4->getFace() != f) return false;

    if (he1->getPrev() != he4) return false;
    if (he2->getPrev() != he1) return false;
    if (he3->getPrev() != he2) return false;
    if (he4->getPrev() != he3) return false;

    egPointer e1 = he1->getEdge();
    egPointer e2 = he2->getEdge();
    egPointer e3 = he3->getEdge();
    egPointer e4 = he4->getEdge();

    if (e1->getMate(he1).isNotNull()) {
      hePointer hm = e1->getMate(he1);
      if (hm->getEdge() != e1) return false;
      if (hm->getVertex() != he1->getNext()->getVertex()) {
	return false;
      }
      if (hm->getNext()->getVertex() != he1->getVertex()) {
	return false;
      }
    }
    else if (!e1->isConstrained() && !e1->isOnBoundary()) {
      return false;
    }

    if (e2->getMate(he2).isNotNull()) {
      hePointer hm = e2->getMate(he2);
      if (hm->getEdge() != e2) return false;
      if (hm->getVertex() != he2->getNext()->getVertex()) {
	return false;
      }
      if (hm->getNext()->getVertex() != he2->getVertex()) {
	return false;
      }
    }
    else if (!e2->isConstrained() && !e2->isOnBoundary()) {
      return false;
    }

    if (e3->getMate(he3).isNotNull()) {
      hePointer hm = e3->getMate(he3);
      if (hm->getEdge() != e3) return false;
      if (hm->getVertex() != he3->getNext()->getVertex()) {
	return false;
      }
      if (hm->getNext()->getVertex() != he3->getVertex()) {
	return false;
      }
    }
    else if (!e3->isConstrained() && !e3->isOnBoundary()) {
      return false;
    }
    
    if (e4->getMate(he4).isNotNull()) {
      hePointer hm = e4->getMate(he4);
      if (hm->getEdge() != e4) return false;
      if (hm->getVertex() != he4->getNext()->getVertex()) {
	return false;
      }
      if (hm->getNext()->getVertex() != he4->getVertex()) {
	return false;
      }
    }
    else if (!e4->isConstrained() && !e4->isOnBoundary()) {
      return false;
    }

    // Checks face geometry
    Point2 p1 = he1->getVertex()->getPoint();
    Point2 p2 = he2->getVertex()->getPoint();
    Point2 p3 = he3->getVertex()->getPoint();
    Point2 p4 = he4->getVertex()->getPoint();

    tQuadrilateral2 q;

    q.insert(0,1,Point2(p1));
    q.insert(1,2,Point2(p2));
    q.insert(2,3,Point2(p3));
    q.insert(3,4,Point2(p4));

    if (!q.isSimple()) {
      return false;
    }

    return true;
  }

  // -----------------------------------------------------------------
  // Method checkMeshConsistency()
  //
  // Checks   if  this   quadrilateral  mesh   is   geometrically  and
  // topologically consistent.
  //
  // -----------------------------------------------------------------
  bool
  tQuadMesh::checkMeshConsistency() const
  {
    // Checks consistency of each face in the mesh
    for (FIT fi = m_cp->fcBegin(); fi != m_cp->fcEnd(); ++fi) {
      if (!checkFaceConsistency(*fi)) return false;
    }

    return true;
  }

  // -----------------------------------------------------------------
  // Method makeTopologicalCorrections()
  //
  // Executes  topological operations  on this  quadrilateral  mesh in
  // order to improve its quality.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::makeTopologicalCorrections()
  {
    eliminateTShapedQuads();
  }

  // -----------------------------------------------------------------
  // Method eliminateTShapedQuads()
  //
  // Eliminates  all triangular-shaped  quadrilaterals  by decomposing
  // them into new ones or combining them with adjacent ones.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::eliminateTShapedQuads()
  {
    LTSFaces lfaces;

    fcPointer f;
    hePointer he1;
    hePointer he2;
    hePointer he3;
    hePointer he4;
    double x1,y1,x2,y2,x3,y3,x4,y4;

    // Builds a list of triangular-shaped faces
    m_tsq = 0;
    for (FIT fi = m_cp->fcBegin(); fi != m_cp->fcEnd(); ++fi) {
      f  = *fi;
      he1 = f->getHalfEdge();
      he2 = he1->getNext();
      he3 = he2->getNext();
      he4 = he3->getNext();

      x1 = he1->getVertex()->getPoint().x();
      y1 = he1->getVertex()->getPoint().y();
      x2 = he2->getVertex()->getPoint().x();
      y2 = he2->getVertex()->getPoint().y();
      x3 = he3->getVertex()->getPoint().x();
      y3 = he3->getVertex()->getPoint().y();
      x4 = he4->getVertex()->getPoint().x();
      y4 = he4->getVertex()->getPoint().y();

      tPolygon2 q(4);
      q.insert(0,1,Point2(x1,y1));
      q.insert(1,2,Point2(x2,y2));
      q.insert(2,3,Point2(x3,y3));
      q.insert(3,4,Point2(x4,y4));

      if ((q.cosangle(0) >= m_COSV) || (q.cosangle(1) >= m_COSV) ||
	  (q.cosangle(2) >= m_COSV) || (q.cosangle(3) >= m_COSV)) {
	lfaces.push_back(f);
	++m_tsq;
      }
    }

    // For each triangular-shaped face,  identifies its >= 170 degrees
    // angle and then  combine with an adjacent face  or subdivide the
    // face to eliminate such an angle.
    while (!lfaces.empty()) {
      f = lfaces.front();
      lfaces.pop_front();

      he1 = f->getHalfEdge();
      he2 = he1->getNext();
      he3 = he2->getNext();
      he4 = he3->getNext();

      x1 = he1->getVertex()->getPoint().x();
      y1 = he1->getVertex()->getPoint().y();
      x2 = he2->getVertex()->getPoint().x();
      y2 = he2->getVertex()->getPoint().y();
      x3 = he3->getVertex()->getPoint().x();
      y3 = he3->getVertex()->getPoint().y();
      x4 = he4->getVertex()->getPoint().x();
      y4 = he4->getVertex()->getPoint().y();

      tPolygon2 q(4);
      q.insert(0,1,Point2(x1,y1));
      q.insert(1,2,Point2(x2,y2));
      q.insert(2,3,Point2(x3,y3));
      q.insert(3,4,Point2(x4,y4));

      if (q.cosangle(0) >= m_COSV) {
	requadrangulate(he4,lfaces);
      }
      else if (q.cosangle(1) >= m_COSV) {
	requadrangulate(he1,lfaces);
      }
      else if (q.cosangle(2) >= m_COSV) {
	requadrangulate(he2,lfaces);
      }
      else if (q.cosangle(3) >= m_COSV) {
	requadrangulate(he3,lfaces);
      }
    }
  }

  // -----------------------------------------------------------------
  // Method requadrangulate()
  //
  // Decomposes a quadrilateral into new  ones by subdividing it or by
  // combining it with adjacent ones.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadrangulate(
			     const hePointer& he,
			     LTSFaces& lfaces
			     )
  {

    try {

      // Checks if quadrilateral can  be combined with the one sharing
      // the edge containing "hr".
      hePointer hr = he->getNext()->getNext();
      if (!hr->getEdge()->isConstrained()) {
	tPolygon2 q = makeRightHexagon(he,hr);

	// Computes the cosine of the angles defined by the vertices 3
	// and 6.
	double cosv3 = q.cosangle(2);
	double cosv5 = q.cosangle(4);
	double cosv6 = q.cosangle(5);
	
	// Finds out if vertices 3 and 6 are convex
	bool cvx3 = !q.isReflex(2);
	bool cvx6 = !q.isReflex(5);
      
	// If vertices 3 and 6 are convex and the angle define by them
	// is not greater than 170  then we can quadrangulate "h" with
	// at most one Steiner point.
	if (cvx3 && cvx6 && (cosv3 < m_COSV) && (cosv6 < m_COSV)) {
	  // Removes  from triangular-shaped  face  list the  adjacent
	  // face used to define an hexagon, if it is there.
	  lfaces.remove(hr->getEdge()->getMate(hr)->getFace());

	  // If the angle at vertex 5  is equal to or greater than 170
	  // degrees, then we can  quadrangulate "h" without using any
	  // Steiner point. Otherwise,  we quadrangulate "h" using one
	  // Steiner point.
	  if (cosv5 >= m_COSV) {
	    requadRight1(he,q);
	  }
	  else {
	    requadRight2(he,q);
	  }
	  return;
	}
	else if (cvx3 && (cosv3 < m_COSV) && (cosv5 < m_COSV)) {
	  // The angle  at vertex  3 is convex  and it is  not greater
	  // than 170 degrees.  Besides, the angle at vertex 5 is less
	  // than 170 degrees
	  
	  // If vertex 2  sees vertex 4 then we  can quadrangulate "h"
	  // using  one Steiner  point.  For keeping  a better  aspect
	  // ratio,  we  test visibility  using  vertex  1 instead  of
	  // vertex 2.
	  Point2 p1 = q.getPoint(0);
	  Point2 p4 = q.getPoint(3);
	  Point2 p6 = q.getPoint(5);
	  Line2 l(p1,p6);
	  if (l.oriented_side(p4) == CGAL::ON_NEGATIVE_SIDE) {
	    // Removes from  triangular-shaped face list  the adjacent
	    // face used to define an hexagon, if it is there.
	    lfaces.remove(hr->getEdge()->getMate(hr)->getFace());
	    
	    requadRight2(he,q);
	    return;
	  }
	}
	else if (cvx6 && (cosv6 < m_COSV)) {
	  // The angle  at Vertex  6 is convex  and it is  not greater
	  // than 170 degrees.
	  
	  // If vertex 2  sees vertex 5 then we  can quadrangulate "h"
	  // using two Steiner points.
	  Point2 p2 = q.getPoint(1);
	  Point2 p3 = q.getPoint(2);
	  Point2 p5 = q.getPoint(4);
	  Line2 l(p2,p3);
	  if (l.oriented_side(p5) == CGAL::ON_POSITIVE_SIDE) {
	    // Removes from  triangular-shaped face list  the adjacent
	    // face used to define an hexagon, if it is there.
	    lfaces.remove(hr->getEdge()->getMate(hr)->getFace());
	    
	    requadRight3(he,q);
	    return;
	  }
	}
	else if (cosv5 < m_COSV) {
	  // The  angles at  vertices 3  and  6 are  greater than  170
	  // degrees  and the  angle  at  vertex 5  is  less than  170
	  // degrees. If the barycenter of the triangle defined by the
	  // vertices 2, 3, and 6 of the hexagon "q" can see vertex 5,
	  // then we can quadrangulate "h" using 2 Steiner points.
	  Point2 p2 = q.getPoint(1);
	  Point2 p3 = q.getPoint(2);
	  Point2 p5 = q.getPoint(4);
	  Point2 p6 = q.getPoint(5);
	  double x = (p6.x() + p2.x() + p3.x()) / 3;
	  double y = (p6.y() + p2.y() + p3.y()) / 3;
	  Point2 b(x,y);
	  Line2 l(b,p6);
	  if (l.oriented_side(p5) == CGAL::ON_NEGATIVE_SIDE) {
	    // Removes from  triangular-shaped face list  the adjacent
	    // face used to define an hexagon, if it is there.
	    lfaces.remove(hr->getEdge()->getMate(hr)->getFace());
	    
	    requadRight4(he,q);
	    return;
	  }
	}
      }

      // Checks if quadrilateral can  be combined with the one sharing
      // the edge containing "hl".
      hePointer hl = he->getPrev();
      if (!hl->getEdge()->isConstrained()) {
	tPolygon2 q = makeLeftHexagon(he,hl);
	
	// Computes the cosine of the angles defined by the vertices 1
	// and 4.
	double cosv1 = q.cosangle(0);
	double cosv5 = q.cosangle(4);
	double cosv4 = q.cosangle(3);
	
	// Finds out if vertices 1 and 4 are convex
	bool cvx1 = !q.isReflex(0);
	bool cvx4 = !q.isReflex(3);
	
	// If vertices  1 and 4 are  convex and the  angles defined at
	// them are not greater than 170 then we can quadrangulate "h"
	// with at most one Steiner point.
	if (cvx1 && cvx4 && (cosv1 < m_COSV) && (cosv4 < m_COSV)) {
	  // Removes  from triangular-shaped  face  list the  adjacent
	  // face used to define an hexagon, if it is there.
	  lfaces.remove(hl->getEdge()->getMate(hl)->getFace());
	  
	  // If the angle at vertex 5  is equal to or greater than 170
	  // degrees, then we can  quadrangulate "h" without using any
	  // Steiner point. Otherwise,  we quadrangulate "h" using one
	  // Steiner point.
	  if (cosv5 >= m_COSV) {
	    requadLeft1(he,q);
	  }
	  else {
	    requadLeft2(he,q);
	  }
	  
	  return;
	}
	else if (cvx1 && (cosv1 < m_COSV) && (cosv5 < m_COSV)) {
	  // The angle  at vertex  1 is convex  and it is  not greater
	  // than 170 degrees.  Besides, the angle at vertex 5 is less
	  // than 170 degrees
	  
	  // If vertex 2  sees vertex 6 then we  can quadrangulate "h"
	  // using  one Steiner  point.  For keeping  a better  aspect
	  // ratio,  we  test visibility  using  vertex  3 instead  of
	  // vertex 2.
	  Point2 p3 = q.getPoint(2);
	  Point2 p4 = q.getPoint(3);
	  Point2 p6 = q.getPoint(5);
	  Line2 l(p3,p4);
	  if (l.oriented_side(p6) == CGAL::ON_POSITIVE_SIDE) {
	    // Removes from  triangular-shaped face list  the adjacent
	    // face used to define an hexagon, if it is there.
	    lfaces.remove(hl->getEdge()->getMate(hl)->getFace());
	    
	    requadLeft2(he,q);
	    return;
	  }
	}
	else if (cvx4 && (cosv4 < m_COSV)) {
	  // The angle  at Vertex  4 is convex  and it is  not greater
	  // than 170 degrees.
	  
	  // If vertex 2  sees vertex 5 then we  can quadrangulate "h"
	  // using two Steiner points.
	  Point2 p1 = q.getPoint(0);
	  Point2 p2 = q.getPoint(1);
	  Point2 p5 = q.getPoint(4);
	  Line2 l(p2,p1);
	  if (l.oriented_side(p5) == CGAL::ON_NEGATIVE_SIDE) {
	    // Removes from  triangular-shaped face list  the adjacent
	    // face used to define an hexagon, if it is there.
	    lfaces.remove(hl->getEdge()->getMate(hl)->getFace());
	    
	    requadLeft3(he,q);
	    return;
	  }
	}
	else if (cosv5 < m_COSV) {
	  // The  angles at  vertices 1  and  4 are  greater than  170
	  // degrees  and the  angle  at  vertex 5  is  less than  170
	  // degrees. If the barycenter of the triangle defined by the
	  // vertices 1, 2, and 4 of the hexagon "q" can see vertex 5,
	  // then we can quadrangulate "h" using 2 Steiner points.
	  Point2 p1 = q.getPoint(0);
	  Point2 p2 = q.getPoint(1);
	  Point2 p4 = q.getPoint(3);
	  Point2 p5 = q.getPoint(4);
	  double x = (p1.x() + p2.x() + p4.x()) / 3;
	  double y = (p1.y() + p2.y() + p4.y()) / 3;
	  Point2 b(x,y);
	  Line2 l(b,p4);
	  if (l.oriented_side(p5) == CGAL::ON_POSITIVE_SIDE) {
	    // Removes  from triangular-shaped  face  list the  adjacent
	    // face used to define an hexagon, if it is there.
	    lfaces.remove(hl->getEdge()->getMate(hl)->getFace());
	    
	    requadLeft4(he,q);
	    return;
	  }
	}
      }

      subdivideQuad(he);

    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }
  }

  // -----------------------------------------------------------------
  // Method makeRightHexagon()
  //
  // Creates an hexagon from two quadrilaterals and then returns it.
  //
  // -----------------------------------------------------------------
  tPolygon2
  tQuadMesh::makeRightHexagon(
			      const hePointer& he1,
			      const hePointer& he2
			      )
  {
    tPolygon2 h(6);

    hePointer heaux = he2->getEdge()->getMate(he2);

    Point2 p1 = he1->getVertex()->getPoint();
    Point2 p2 = he1->getNext()->getVertex()->getPoint();
    Point2 p3 = he2->getVertex()->getPoint();
    Point2 p4 = heaux->getNext()->getNext()->getVertex()->getPoint();
    Point2 p5 = heaux->getPrev()->getVertex()->getPoint();
    Point2 p6 = heaux->getVertex()->getPoint();

    h.insert(0,1,p1);
    h.insert(1,2,p2);
    h.insert(2,3,p3);
    h.insert(3,4,p4);
    h.insert(4,5,p5);
    h.insert(5,6,p6);

    return h;
  }

  // -----------------------------------------------------------------
  // Method makeLeftHexagon()
  //
  // Creates an hexagon from two quadrilaterals and then returns it.
  //
  // -----------------------------------------------------------------
  tPolygon2
  tQuadMesh::makeLeftHexagon(
			     const hePointer& he1,
			     const hePointer& he2
			     )
  {
    tPolygon2 h(6);

    hePointer heaux = he2->getEdge()->getMate(he2);

    Point2 p1 = he1->getVertex()->getPoint();
    Point2 p2 = he1->getNext()->getVertex()->getPoint();
    Point2 p3 = he1->getNext()->getNext()->getVertex()->getPoint();
    Point2 p4 = he1->getPrev()->getVertex()->getPoint();
    Point2 p5 = heaux->getPrev()->getPrev()->getVertex()->getPoint();
    Point2 p6 = heaux->getPrev()->getVertex()->getPoint();

    h.insert(0,1,Point2(p1));
    h.insert(1,2,Point2(p2));
    h.insert(2,3,Point2(p3));
    h.insert(3,4,Point2(p4));
    h.insert(4,5,Point2(p5));
    h.insert(5,6,Point2(p6));

    return h;
  }

  // -----------------------------------------------------------------
  // Method subdivideQuad()
  //
  // Subdivides a  quadrilateral into  5 new ones  by using  4 Steiner
  // points.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::subdivideQuad(const hePointer& he)
    throw (tExceptionObject)
  {

    hePointer he1 = he->getPrev()->getPrev();

    Point2 p1 = he1->getNext()->getVertex()->getPoint();
    Point2 p2 = he1->getNext()->getNext()->getVertex()->getPoint();
    Point2 p3 = he1->getPrev()->getVertex()->getPoint();
    Point2 p4 = he1->getVertex()->getPoint();

    // The four  Steiner points are computed according  to Barry Joe's
    // paper:  Joe,  B.  Quadrilateral  Mesh  Generation in  Polygonal
    // Regions,   Computer-Aided   Design,   Volume  27,   Number   3,
    // pp. 209-222, March 1995.

    double dx, dy, d23, d34;

    // Computes the distance from p2 to p3
    dx  = p3.x() - p2.x();
    dy  = p3.y() - p2.y();
    d23 = sqrt((dx * dx) + (dy * dy));

    // Computes the distance from p3 to p4
    dx  = p3.x() - p4.x();
    dy  = p3.y() - p4.y();
    d34 = sqrt((dx * dx) + (dy * dy));

    double b = d23 / (d23 + d34);
    double a1, a2, a3;

    a1 = (b < 0.5) ? (b / 2) : 0.25;
    a2 = 0.25;
    a3 = ((1 - b) < 0.5) ? ((1 - b) / 2) : 0.25;

    dx = ((1 - a2 - a3) * p4.x()) + (a2 * p1.x()) + (a3 * p2.x());
    dy = ((1 - a2 - a3) * p4.y()) + (a2 * p1.y()) + (a3 * p2.y());
    Point2 sp4(dx,dy);

    dx = (a1 * p4.x()) + ((1 - a1 - a3) * p1.x()) + (a3 * p2.x());
    dy = (a1 * p4.y()) + ((1 - a1 - a3) * p1.y()) + (a3 * p2.y());
    Point2 sp2(dx,dy);

    dx = (a1 * p4.x()) + (a2 * p1.x()) + ((1 - a1 - a2) * p2.x());
    dy = (a1 * p4.y()) + (a2 * p1.y()) + ((1 - a1 - a2) * p2.y());
    Point2 sp3(dx,dy);

    dx = (0.5 * p3.x()) + (0.25 * (sp4.x() + sp3.x()));
    dy = (0.5 * p3.y()) + (0.25 * (sp4.y() + sp3.y()));
    Point2 sp1(dx,dy);

#ifdef DEBUG_MODE
    if (!isAValidFace(sp1,sp4,sp2,sp3)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p2,p3,sp1,sp3)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p3,p4,sp4,sp1)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p4,p1,sp2,sp4)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p1,p2,sp3,sp2)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif
    
    // Creates the new vertices
    vtPointer newv1 = m_cp->createVertex(sp1);
    vtPointer newv2 = m_cp->createVertex(sp2);
    vtPointer newv3 = m_cp->createVertex(sp3);
    vtPointer newv4 = m_cp->createVertex(sp4);

    // Splits face corresponding to the quadrilateral p1p2p3p4 into two
    // triangles, p1p2p3 and p1p3p4.
    m_cp->mkFE(he1->getPrev(),he1->getNext());

    // Splits the new edge p1p3
    m_cp->spEmkEV(newv3,he1->getNext());

    // Splits the new edge p1sp3
    m_cp->spEmkEV(newv2,he1->getNext());

    // Splits the new edge sp3p3
    m_cp->spEmkEV(newv1,he1->getPrev()->getPrev());

    // Splits face  corresponding to the  hexagon p1sp2sp3sp1p3p4 into
    // one triangle, p4p1sp2, and one pentagon p4sp2sp3sp1p3.
    m_cp->mkFE(he1->getNext()->getNext(),he1);

    // Splits the new edge p4sp2
    m_cp->spEmkEV(newv4,he1->getPrev());

    // Splits face corresponding  to the hexagon sp2sp3sp1p3p4sp4 into
    // two quadrilaterals sp2sp3sp1sp4 and sp1p3p4sp4.
    hePointer heaux1 = he1->getPrev();
    heaux1 = heaux1->getEdge()->getMate(heaux1)->getNext();
    hePointer heaux2 = heaux1->getPrev()->getPrev()->getPrev();
    m_cp->mkFE(heaux2,heaux1);

    // Splits face  corresponding to the  hexagon p1p2p3sp1sp3sp2 into
    // two quadrilaterals p1p2sp3sp2 and p2p3sp1sp3.
    heaux1 = he1->getNext()->getEdge()->getMate(he1->getNext())->getPrev();
    heaux2 = heaux1->getNext()->getNext()->getNext();
    m_cp->mkFE(heaux2,heaux1);
  }

  // -----------------------------------------------------------------
  // Method requadRight1()
  //
  // Requadrangulates an hexagon by inserting one diagonal.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadRight1(
			  const hePointer& he,
			  const tPolygon2& q
			  )
     throw (tExceptionObject)
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);
    Point2 p5 = q.getPoint(4);
    Point2 p6 = q.getPoint(5);

#ifdef DEBUG_MODE  
    if (!isAValidFace(p1,p2,p5,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    if (!isAValidFace(p2,p3,p4,p5)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif

    // Merges the quadrilateral faces p1p2p3p4 and p3p4p5p6
    m_cp->mgFrmE(he->getNext()->getNext());

    // Splits face corresponding to  the hexagon p1p2p3p4p5p6 into two
    // quadrilaterals, p1p2p5p6 and p2p3p4p5.
    m_cp->mkFE(he->getNext(),he->getPrev()->getPrev());
  }

  // -----------------------------------------------------------------
  // Method requadRight2()
  //
  // Requadrangulates an hexagon using one Steiner point.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadRight2(
			  const hePointer& he,
			  const tPolygon2& q
			  )
     throw (tExceptionObject)
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);
    Point2 p5 = q.getPoint(4);
    Point2 p6 = q.getPoint(5);

    // Computes location of a Steiner point. This point will be placed
    // in the barycenter of the  region defined by the intersection of
    // the triangle p3p4p6 and two half-spaces.
    list<Point2> li;
    list<Point2> lo;

    li.push_back(p3); li.push_back(p4); li.push_back(p6);

    // Computes  intersection   of  triangle  p3p4p6   and  the  right
    // half-space defined by the line p2p4.
    try {
      halfSpaceIntersection(li,p2,p4,lo,CGAL::ON_NEGATIVE_SIDE);
      li.clear();
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    Point2 sp;
    double x = 0;
    double y = 0;
    int num = 0; 
    while (!lo.empty()) {
      sp = lo.front();
      lo.pop_front();
      x += sp.x();
      y += sp.y();
      ++num;
    }

    sp = Point2(x/num, y/num);

#ifdef DEBUG_MODE
    if (!isAValidFace(p2,p3,p4,sp)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    if (!isAValidFace(sp,p4,p5,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    if (!isAValidFace(p1,p2,sp,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif

    // Creates the new vertex
    vtPointer newv = m_cp->createVertex(sp);

    // Splits face corresponding to the quadrilateral p3p4p5p6 into two
    // triangles, p4p5p6 and p3p4p6.
    hePointer heaux1 = he->getNext()->getNext();
    heaux1 = heaux1->getEdge()->getMate(heaux1);
    hePointer heaux2 = heaux1->getNext()->getNext();
    m_cp->mkFE(heaux1,heaux2);

    // Splits the new edge p4p6
    m_cp->spEmkEV(newv,heaux1->getPrev());

    // Merges quadrilateral faces p1p2p3p4 and p3p4spp6
    m_cp->mgFrmE(he->getNext()->getNext());

    // Splits face  corresponding to the  hexagon p1p2p3spp5p6 into
    // two quadrilaterals, p1p2spp6 and p2p3p4sp.
    m_cp->mkFE(he->getNext(),he->getPrev()->getPrev());
  }

  // -----------------------------------------------------------------
  // Method requadRight3()
  //
  // Requadrangulates an hexagon using two Steiner points.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadRight3(
			  const hePointer& he,
			  const tPolygon2& q
			  )
     throw (tExceptionObject)
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);
    Point2 p5 = q.getPoint(4);
    Point2 p6 = q.getPoint(5);

    // Computes the location of the first Steiner point
    double x = (p6.x() + p2.x() + p3.x()) / 3;
    double y = (p6.y() + p2.y() + p3.y()) / 3;
    Point2 sp1(x,y);

    // Computes the  location of the second Steiner  point. This point
    // will be placed  in the barycenter of the  region defined by the
    // intersection of the triangle p6p3p5 and three half-spaces.
    list<Point2> li;
    list<Point2> lo;

    li.push_back(p6); li.push_back(p3); li.push_back(p5);

    // Computes   intersection  of  triangle   p6p3p5  and   the  left
    // half-space determined by the oriented line sp1p4, and the right
    // half-spaces determined by the oriented lines sp1p5, p2sp1.
    try {
      halfSpaceIntersection(li,sp1,p4,lo,CGAL::ON_NEGATIVE_SIDE);
      li.clear();
      halfSpaceIntersection(lo,sp1,p5,li,CGAL::ON_POSITIVE_SIDE);
      lo.clear();
      halfSpaceIntersection(li,p2,sp1,lo,CGAL::ON_POSITIVE_SIDE);
      li.clear();
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    Point2 sp2;
    x = 0;
    y = 0;
    int num = 0; 
    while (!lo.empty()) {
      sp2 = lo.front();
      lo.pop_front();
      x += sp2.x();
      y += sp2.y();
      ++num;
    }

    sp2 = Point2(x/num, y/num);

#ifdef DEBUG_MODE
    if (!isAValidFace(p1,p2,sp1,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p2,p3,sp2,sp1)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p3,p4,p5,sp2)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(sp1,sp2,p5,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif

    // Creates the new vertices
    vtPointer newv1 = m_cp->createVertex(sp1);
    vtPointer newv2 = m_cp->createVertex(sp2);

    // Splits  face corresponding to  the quadrilateral  p1p2p3p6 into
    // two triangles, p1p2p6 and p2p3p6.
    hePointer heaux1 = he->getNext()->getNext();
    heaux1 = heaux1->getEdge()->getMate(heaux1);
    m_cp->mkFE(he->getPrev(),he->getNext());

    // Splits the new edge p2p6
    m_cp->spEmkEV(newv1,he->getNext());

    // Splits  face corresponding to  the quadrilateral  p6p3p4p5 into
    // two triangles, p6p3p5 and p3p4p5.
    m_cp->mkFE(heaux1->getNext(),heaux1->getPrev());

    // Splits the new edge p5p3
    m_cp->spEmkEV(newv2,heaux1->getNext());

    // Merges the quadrilateral faces p6p3sp2p5 and p6sp1p2p3
    hePointer heaux2 = heaux1->getEdge()->getMate(heaux1)->getNext()->getNext();
    heaux1 = heaux1->getNext()->getNext();
    m_cp->mgFrmE(heaux1->getPrev()->getPrev());

    // Splits  the  hexagon  p6sp1p2p3sp2p5 into  two  quadrilaterals,
    // p6sp1sp2p5 and sp1p2p3sp2.
    m_cp->mkFE(heaux1,heaux2);
  }

  // -----------------------------------------------------------------
  // Method requadRight4()
  //
  // Requadrangulates an hexagon using two Steiner points.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadRight4(
			  const hePointer& he,
			  const tPolygon2& q
			  )
    throw (tExceptionObject)
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);
    Point2 p5 = q.getPoint(4);
    Point2 p6 = q.getPoint(5);

    // Computes the location of the first Steiner point
    double x = (p6.x() + p2.x() + p3.x()) / 3;
    double y = (p6.y() + p2.y() + p3.y()) / 3;
    Point2 sp1(x,y);

    // Computes location of the  second Steiner point. This point will
    // be placed on the diagonal p3p6 of "q".
    Line2 l1(p2,sp1);
    Line2 l2(p2,p5);
    Line2 l3(sp1,p5);
    Segment2 s(p3,p6);

    Point2 pt1;
    if (!myAssign(pt1,s,l1)) {
      pt1 = Point2((p3.x() + p6.x()) / 2, (p3.y() + p6.y()) / 2);
    }

    Point2 pt2;
    if (!myAssign(pt2,s,l2)) {
      pt2 = Point2((p3.x() + p6.x()) / 2, (p3.y() + p6.y()) / 2);
    }

    Point2 pt3;
    if (!myAssign(pt3,s,l3)) {
      pt3 = Point2((p3.x() + p6.x()) / 2, (p3.y() + p6.y()) / 2);
    }

    double d1, d2, d3, dx, dy;

    dx = pt1.x() - p3.x();
    dy = pt1.y() - p3.y();
    d1 = (dx * dx) + (dy * dy);

    dx = pt2.x() - p3.x();
    dy = pt3.y() - p3.y();
    d2 = (dx * dx) + (dy * dy);
 
    dx = pt3.x() - p3.x();
    dy = pt3.y() - p3.y();
    d3 = (dx * dx) + (dy * dy);

    Point2 sp2;
    if ((d1 <= d2) && (d1 <= d3)) {
      sp2 = Point2((pt1.x() + p3.x()) / 2, (pt1.y() + p3.y()) / 2);
    }
    else if (d2 <= d3) {
      sp2 = Point2((pt2.x() + p3.x()) / 2, (pt2.y() + p3.y()) / 2);
    }
    else {
      sp2 = Point2((pt3.x() + p3.x()) / 2, (pt3.y() + p3.y()) / 2);
    }


#ifdef DEBUG_MODE
    if (!isAValidFace(p1,p2,sp1,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p2,p3,sp2,sp1)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p3,p4,p5,sp2)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(sp1,sp2,p5,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif

    // Creates the new vertices
    vtPointer newv1 = m_cp->createVertex(sp1);
    vtPointer newv2 = m_cp->createVertex(sp2);

    // Splits  face corresponding to  the quadrilateral  p1p2p3p6 into
    // two triangles, p1p2p6 and p2p3p6.
    hePointer heaux1 = he->getNext()->getNext();
    heaux1 = heaux1->getEdge()->getMate(heaux1);
    m_cp->mkFE(he->getPrev(),he->getNext());

    // Splits the new edge p2p6
    m_cp->spEmkEV(newv1,he->getNext());

    // Splits  face corresponding to  the quadrilateral  p6p3p4p5 into
    // two triangles, p6p3p5 and p3p4p5.
    m_cp->mkFE(heaux1->getNext(),heaux1->getPrev());

    // Splits the new edge p5p3
    m_cp->spEmkEV(newv2,heaux1->getNext());

    // Merges the quadrilateral faces p6p3sp2p5 and p6sp1p2p3
    hePointer heaux2 = heaux1->getEdge()->getMate(heaux1)->getNext()->getNext();
    heaux1 = heaux1->getNext()->getNext();
    m_cp->mgFrmE(heaux1->getPrev()->getPrev());

    // Splits  the  hexagon  p6sp1p2p3sp2p5 into  two  quadrilaterals,
    // p6sp1sp2p5 and sp1p2p3sp2.
    m_cp->mkFE(heaux1,heaux2);
  }

  // -----------------------------------------------------------------
  // Method requadLeft1()
  //
  // Requadrangulates an hexagon by  inserting one diagonal. It is the
  // symmetric method to requadRight1().
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadLeft1(
			 const hePointer& he,
			 const tPolygon2& q
			 )
    throw (tExceptionObject)
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);
    Point2 p5 = q.getPoint(4);
    Point2 p6 = q.getPoint(5);
    
#ifdef DEBUG_MODE
    if (!isAValidFace(p1,p2,p5,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    if (!isAValidFace(p2,p3,p4,p5)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif

    // Merges the quadrilateral faces p1p2p3p4 and p1p4p5p6
    m_cp->mgFrmE(he->getPrev());

    // Splits face corresponding to  the hexagon p1p2p3p4p5p6 into two
    // quadrilaterals, p1p2p5p6 and p2p3p4p5.
    m_cp->mkFE(he->getNext(),he->getPrev()->getPrev());
  }

  // -----------------------------------------------------------------
  // Method requadLeft2()
  //
  // Requadrangulates an  hexagon using one  Steiner point. It  is the
  // symmetric method to requadRight2().
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadLeft2(
			 const hePointer& he,
			 const tPolygon2& q
			 )
    throw (tExceptionObject)
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);
    Point2 p5 = q.getPoint(4);
    Point2 p6 = q.getPoint(5);

    // Computes location of a Steiner point. This point will be placed
    // in the barycenter of the  region defined by the intersection of
    // the triangle p1p4p6 and two half-spaces.
    list<Point2> li;
    list<Point2> lo;

    li.push_back(p1); li.push_back(p4); li.push_back(p6);

    // Computes  intersection   of  triangle  p1p4p6   and  the  right
    // half-space defined by the line p2p6.
    try {
      halfSpaceIntersection(li,p2,p6,lo,CGAL::ON_POSITIVE_SIDE);
      li.clear();
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    Point2 sp;
    double x = 0;
    double y = 0;
    int num = 0; 
    while (!lo.empty()) {
      sp = lo.front();
      lo.pop_front();
      x += sp.x();
      y += sp.y();
      ++num;
    }

    sp = Point2(x/num, y/num);

#ifdef DEBUG_MODE
    if (!isAValidFace(p2,p3,p4,sp)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    if (!isAValidFace(sp,p4,p5,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    if (!isAValidFace(p1,p2,sp,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif

    // Creates the new vertex
    vtPointer newv = m_cp->createVertex(sp);

    // Splits face corresponding to the quadrilateral p1p4p5p6 into two
    // triangles, p4p5p6 and p1p4p6.
    hePointer heaux1 = he->getPrev();
    heaux1 = heaux1->getEdge()->getMate(heaux1);
    hePointer heaux2 = heaux1->getNext();
    m_cp->mkFE(heaux1->getPrev(),heaux2);

    // Splits the new edge p4p6
    m_cp->spEmkEV(newv,heaux1->getNext());

    // Merges quadrilateral faces p1p2p3p4 and p1p4spp6
    heaux1 = heaux1->getNext()->getNext();
    m_cp->mgFrmE(he->getPrev());

    // Splits face corresponding to  the hexagon p1p2p3p4spp6 into two
    // quadrilaterals, p1p2spp6 and p2p3p4sp.
    m_cp->mkFE(heaux1,he->getNext());
  }

  // -----------------------------------------------------------------
  // Method requadLeft3()
  //
  // Requadrangulates an  hexagon using two Steiner points.  It is the
  // symmetric method to requadRight3().
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadLeft3(
			 const hePointer& he,
			 const tPolygon2& q
			 )
     throw (tExceptionObject)
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);
    Point2 p5 = q.getPoint(4);
    Point2 p6 = q.getPoint(5);

    // Computes the location of the first Steiner point
    double x = (p1.x() + p2.x() + p4.x()) / 3;
    double y = (p1.y() + p2.y() + p4.y()) / 3;
    Point2 sp1(x,y);

    // Computes the  location of the second Steiner  point. This point
    // will be placed  in the barycenter of the  region defined by the
    // intersection of the triangle p1p4p5 and three half-spaces.
    list<Point2> li;
    list<Point2> lo;

    li.push_back(p1); li.push_back(p4); li.push_back(p5);

    // Computes   intersection  of  triangle   p1p4p5  and   the right
    // half-space determined by the oriented line sp1p6, and the right
    // half-spaces determined by the oriented lines sp1p5, p2sp1.
    try {
      halfSpaceIntersection(li,sp1,p6,lo,CGAL::ON_POSITIVE_SIDE);
      li.clear();
      halfSpaceIntersection(lo,sp1,p5,li,CGAL::ON_NEGATIVE_SIDE);
      lo.clear();
      halfSpaceIntersection(li,p2,sp1,lo,CGAL::ON_NEGATIVE_SIDE);
      li.clear();
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    Point2 sp2;
    x = 0;
    y = 0;
    int num = 0; 
    while (!lo.empty()) {
      sp2 = lo.front();
      lo.pop_front();
      x += sp2.x();
      y += sp2.y();
      ++num;
    }

    sp2 = Point2(x/num, y/num);

#ifdef DEBUG_MODE
    if (!isAValidFace(p2,p3,p4,sp1)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p1,p2,sp1,sp2)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p5,p6,p1,sp2)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p5,sp2,sp1,p4)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif

    // Creates the new vertices
    vtPointer newv1 = m_cp->createVertex(sp1);
    vtPointer newv2 = m_cp->createVertex(sp2);

    // Splits  face corresponding to  the quadrilateral  p1p2p3p4 into
    // two triangles, p1p2p4 and p2p3p4.
    hePointer heaux1 = he->getPrev();
    heaux1 = heaux1->getEdge()->getMate(heaux1);
    m_cp->mkFE(he->getNext(),he->getPrev());

    // Splits the new edge p2p4
    m_cp->spEmkEV(newv1,he->getNext());

    // Splits  face corresponding to  the quadrilateral  p4p5p6p1 into
    // two triangles, p4p5p1 and p5p6p1.
    m_cp->mkFE(heaux1,heaux1->getPrev()->getPrev());

    // Splits the new edge p5p1
    m_cp->spEmkEV(newv2,heaux1->getPrev());

    // Merges the quadrilateral faces p4p5sp2p1 and p4p1p2sp1
    hePointer heaux2 = he->getNext()->getNext();
    heaux1 = heaux1->getPrev();
    m_cp->mgFrmE(heaux1->getNext());

    // Splits  the  hexagon  p4p5sp2p1p2sp1 into  two  quadrilaterals,
    // p4p5sp2sp1 and p1p2sp1sp21.
    m_cp->mkFE(heaux1,heaux2);
  }

  // -----------------------------------------------------------------
  // Method requadLeft4()
  //
  // Requadrangulates an  hexagon using two Steiner points.  It is the
  // symmetric method to requadRight4().
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::requadLeft4(
			  const hePointer& he,
			  const tPolygon2& q
			  )
    throw (tExceptionObject)
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);
    Point2 p5 = q.getPoint(4);
    Point2 p6 = q.getPoint(5);

    // Computes the location of the first Steiner point
    double x = (p1.x() + p2.x() + p4.x()) / 3;
    double y = (p1.y() + p2.y() + p4.y()) / 3;
    Point2 sp1(x,y);

    // Computes location of the  second Steiner point. This point will
    // be placed on the diagonal p1p4 of "q".
    Line2 l1(p2,sp1);
    Line2 l2(p2,p5);
    Line2 l3(sp1,p5);
    Segment2 s(p1,p4);

    Point2 pt1;
    if (!myAssign(pt1,s,l1)) {
      pt1 = Point2((p1.x() + p4.x()) / 2, (p1.y() + p4.y()) / 2);
    }

    Point2 pt2;
    if (!myAssign(pt2,s,l2)) {
      pt2 = Point2((p1.x() + p4.x()) / 2, (p1.y() + p4.y()) / 2);
    }

    Point2 pt3;
    if (!myAssign(pt3,s,l3)) {
      pt3 = Point2((p1.x() + p4.x()) / 2, (p1.y() + p4.y()) / 2);
    }

    double d1, d2, d3, dx, dy;

    dx = pt1.x() - p1.x();
    dy = pt1.y() - p1.y();
    d1 = (dx * dx) + (dy * dy);

    dx = pt2.x() - p1.x();
    dy = pt3.y() - p1.y();
    d2 = (dx * dx) + (dy * dy);
 
    dx = pt3.x() - p1.x();
    dy = pt3.y() - p1.y();
    d3 = (dx * dx) + (dy * dy);

    Point2 sp2;
    if ((d1 <= d2) && (d1 <= d3)) {
      dx = 0.5*(pt1.x() - p1.x()) + p1.x();
      dy = 0.5*(pt1.y() - p1.y()) + p1.y();
      sp2 = Point2(dx,dy);
    }
    else if (d2 <= d3) {
      dx = 0.5*(pt2.x() - p1.x()) + p1.x();
      dy = 0.5*(pt2.y() - p1.y()) + p1.y();
      sp2 = Point2(dx,dy);
    }
    else {
      dx = 0.5*(pt3.x() - p1.x()) + p1.x();
      dy = 0.5*(pt3.y() - p1.y()) + p1.y();
      sp2 = Point2(dx,dy);
    }


#ifdef DEBUG_MODE
    if (!isAValidFace(p3,p4,sp1,p2)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p1,p2,sp1,sp2)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p4,p5,sp2,sp1)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if (!isAValidFace(p1,sp2,p5,p6)) {
      std::stringstream ss (std::stringstream::in | std::stringstream::out);
      ss << "Method for correcting topology generated non-convex face";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
#endif

    // Creates the new vertices
    vtPointer newv1 = m_cp->createVertex(sp1);
    vtPointer newv2 = m_cp->createVertex(sp2);

    // Splits  face corresponding to  the quadrilateral  p1p2p3p4 into
    // two triangles, p1p2p4 and p2p3p4.
    hePointer heaux1 = he->getPrev();
    heaux1 = heaux1->getEdge()->getMate(heaux1);
    m_cp->mkFE(he->getNext(),he->getPrev());

    // Splits the new edge p2p4
    m_cp->spEmkEV(newv1,he->getNext());

    // Splits  face corresponding to  the quadrilateral  p6p1p4p5 into
    // two triangles, p5p6p1 and p1p4p5.
    m_cp->mkFE(heaux1->getNext()->getNext(),heaux1);

    // Splits the new edge p1p5
    m_cp->spEmkEV(newv2,heaux1->getPrev());

    // Merges the quadrilateral faces p1sp2p4p5 and p1p2sp2p4
    hePointer heaux2 = he->getNext()->getNext();
    heaux1 = heaux1->getPrev();
    m_cp->mgFrmE(heaux1->getNext());

    // Splits  the  hexagon  p5sp2p1p2sp1p4 into  two  quadrilaterals,
    // p1p2sp1sp2 and p5sp2sp1p4.
    m_cp->mkFE(heaux1,heaux2);
  }

  // -----------------------------------------------------------------
  // Method isAValidFace()
  //
  //
  // -----------------------------------------------------------------
  bool
  tQuadMesh::isAValidFace(
			  const Point2& p1,
			  const Point2& p2,
			  const Point2& p3,
			  const Point2& p4
			  )
  {
    tPolygon2 q(4);

    q.insert(0,1,p1);
    q.insert(1,2,p2);
    q.insert(2,3,p3);
    q.insert(3,4,p4);

    if (!q.isSimple()) {
      return false;
    }

    return q.isConvex();
  }

  // -----------------------------------------------------------------
  // Method smooth()
  //
  // Uses the angle-based method  to smooth a quadrilateral mesh. This
  // method is due to Tian Zhou and Kenji Shimada, and its description
  // can be found in the following paper:
  //
  // Zhou,   T.   and   Shimada,  K.    An  Angle-Based   Approach  to
  // Two-Dimensional Mesh Smooth.  Proceeding of the 9th International
  // Meshing  Roundtable,  Sandia  National Laboratories,  pp.373-384,
  // October 2000.
  //
  // -----------------------------------------------------------------
  void
  tQuadMesh::smooth()
  {
    double x, y, xs, ys, x1, y1, x2, y2, x3, y3, d1, d2, d3, a1, a2, 
      beta, sinb, cosb;

    if (m_cp.isNull()) return;

    // For  each  vertex  of  the  mesh  that is  not  incident  to  a
    // constrained  edge,  computes  its  new position  based  on  the
    // vertices on its star.
    vtPointer v;
    for (VIT vi = m_cp->vtBegin(); vi != m_cp->vtEnd(); ++vi) {
      v = *vi;
      
      // Check if vertex "v" is not incident to a constrained edge. If
      // so, we cannot move "v".
      if (!m_cp->isOnConstrainedEdge(v)) {
	// Gets all vertices in the star of "v"

	// Creates a  vector with the points  in the star  of "v" (not
	// including "v" itself) and another vector with the length of
	// the edges  in the  polygon defined by  the vertices  in the
	// star of "v" (not including "v" itself).
	vector<Point2> vp = m_cp->getVertexStar(v);

	// Computes the new position of "v"
	x = v->getPoint().x();
	y = v->getPoint().y();
	xs = 0;
	ys = 0;
	for (unsigned int i = 0; i < vp.size(); i++) {
	  // Computes  the  angles  formed   by  the  pairs  of  edges
	  // (vv(i),v(i)v(i+1)) and (vv(i),v(i-1)v(i)).
	  if (i == 0) {
	    x1 = vp[vp.size()-1].x();
	    y1 = vp[vp.size()-1].y();
	  }
	  else {
	    x1 = vp[i-1].x();
	    y1 = vp[i-1].y();
	  }
	  x2 = vp[i].x();
	  y2 = vp[i].y();
	  x3 = vp[(i+1) % vp.size()].x();
	  y3 = vp[(i+1) % vp.size()].y();

	  Point2 v1((x1-x2),(y1-y2));
	  Point2 v2((x -x2),(y -y2));
	  Point2 v3((x3-x2),(y3-y2));

	  d1 = sqrt((v1.x() * v1.x()) + (v1.y() * v1.y()));
	  d2 = sqrt((v2.x() * v2.x()) + (v2.y() * v2.y()));
	  d3 = sqrt((v3.x() * v3.x()) + (v3.y() * v3.y()));

	  a1 = acos(((v1.x() * v2.x()) + (v1.y() * v2.y())) / (d1 * d2));
	  a2 = acos(((v3.x() * v2.x()) + (v3.y() * v2.y())) / (d3 * d2));

	  beta = (a1 - a2) / 2;

	  sinb = sin(beta);
	  cosb = cos(beta);

	  xs += x2 + (v2.x() * cosb) - (v2.y() * sinb); 
	  ys += y2 + (v2.x() * sinb) + (v2.y() * cosb);
	}

	// New position  of "v"  is the average  of all  new positions
	// computed before.
	double nx = xs / vp.size();
	double ny = ys / vp.size();

	Polygon2 q;
	for (unsigned int j=0; j < vp.size(); j++) {
	  q.push_back(Point2(vp[j].x(),vp[j].y()));
	}

	if (q.is_simple()) {
	  // If the new point is inside the star of "v" then we try to
	  // move the point;
	  if (q.has_on_bounded_side(Point2(nx,ny))) {
	    v->setPoint(Point2(nx,ny));

	    // If  the  new   point  position  produces  a  non-convex
	    // quadrilateral, we undo the change of position.
	    if (isNotAValidMove(v)) {
	      v->setPoint(Point2(x,y));
	    }
	  }
	}
      }
    }
  }

  // -----------------------------------------------------------------
  // Method isNotAValidMove()
  //
  // Returns true  if the new position  of a vertex causes  one of its
  // incident face to be non-convex.
  //
  // -----------------------------------------------------------------
  bool
  tQuadMesh::isNotAValidMove(const vtPointer& v)
  {
    hePointer he1 = v->getHalfEdge();
    hePointer he2 = he1;
    do {
      tPolygon2 q(4);

      q.insert(0,1,Point2(he2->getVertex()->getPoint()));
      q.insert(1,2,Point2(he2->getNext()->getVertex()->getPoint()));
      q.insert(2,3,Point2(he2->getNext()->getNext()->getVertex()->getPoint()));
      q.insert(3,4,Point2(he2->getPrev()->getVertex()->getPoint()));

      if (!q.isSimple()) return true;
      if (!q.isConvex()) return true;

      if (q.cosangle(0) >= m_COSV) return true;
      if (q.cosangle(1) >= m_COSV) return true;
      if (q.cosangle(2) >= m_COSV) return true;
      if (q.cosangle(3) >= m_COSV) return true;

      he2 = m_cp->getNextStarVert(he2);
    } while (he2 != he1);

    return false;
  }

};     // end namespace qmpp
