//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    quadmesh2.h
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#ifndef __quadmesh2_h
#define __quadmesh2_h

#include <string>
#include <list>
#include <map>

#include "typedefs.h"
#include "macros.h"
#include "smartptr.h"
#include "dataobj.h"
#include "exceptionobj.h"

#include "edgekey.h"
#include "polygon2.h"
#include "quad2.h"
#include "reader.h"
#include "hedge.h"
#include "mesh2.h"

namespace qmpp {

  // -----------------------------------------------------------------
  // class tQuadMesh
  //
  // This class  contains methods to  build a quadrilateral  mesh from
  // .NODE, .POLY and .ELE files containing quadrilateral nodes, edges
  // and  faces, respectively.  The  resulting mesh  is stored  in the
  // handle-edge data structure.
  //
  // -----------------------------------------------------------------
  class tQuadMesh : public tMesh {
  public:
    // Type definitions
    typedef hedge::tComplex2::fcIterator   FIT;
    typedef hedge::tComplex2::vtIterator   VIT;
    typedef list<fcPointer>                LTSFaces;

    // Default constructor
    tQuadMesh() : 
      tMesh(), 
      m_COSV(0.98480775),
      m_tsq(0)
      {}

    // Destructor
    ~tQuadMesh() {}

    // Builds a triangular mesh from .NODE, .EDGE, and .ELE files, and
    // stores   the   resulting  mesh   into   the  handle-edge   data
    // structure.
    cpPointer buildMesh(const string&) throw (tExceptionObject);

    // Makes topological improvements on this mesh
    virtual void makeTopologicalCorrections();

    // Smooths a mesh
    virtual void smooth();

    // Returns the number of triangular-shaped quadrilaterals found in
    // this mesh.
    unsigned int getNumTSQuads() const 
      { return m_tsq; }

  private:
    // Creates a  new quadrilateral face as a  new connected component
    // of the mesh.
    void makeQuadFace(const tQuadrilateral2&, int, int, int, int,
		      const Point2&,const Point2&,const Point2&,
		      const Point2&,const tEdgeKey&,const tEdgeKey&,
		      const tEdgeKey&, const tEdgeKey&,
		      const ConstrainedEdgeSet&
		      );

    // Creates a new  triangular face and attaches it  to another face
    // in the mesh.
    void makeQuadFace(const tQuadrilateral2&, int, const Point2&, int,
		      const Point2&, const tEdgeKey&, const tEdgeKey&,
		      const tEdgeKey&, const tEdgeKey&, 
		      const ConstrainedEdgeSet&
		      );

    // Creates a  new quadrilateral face as a  new connected component
    // of the  mesh and  keeps track of  its common edge  with another
    // face of the mesh, so  that we can glue this connected component
    // later.
    void makeQuadFaceGlue1(const tQuadrilateral2&, int, int, int, int,
			   const Point2&, const Point2&,const Point2&,
			   const Point2&, int, GlueEdgeSet&,
			   const ConstrainedEdgeSet&
			   );

    // Creates a  new quadrilateral face as a  new connected component
    // of the mesh and keeps track  of its two common edges with other
    // faces of the mesh, so that we can glue this connected component
    // later.
    void makeQuadFaceGlue2(const tQuadrilateral2&, int, int, int, int,
			   const Point2&, const Point2&, const Point2&,
			   const Point2&, const tEdgeKey&, 
			   const tEdgeKey&, const tEdgeKey&,
			   const tEdgeKey&, int, GlueEdgeSet&,
			   const ConstrainedEdgeSet&
			   );

    // Creates a  new quadrilateral face as a  new connected component
    // of  the mesh and  keeps track  of its  three common  edges with
    // other faces  of the  mesh, so that  we can glue  this connected
    // component later.
    void makeQuadFaceGlue3(const tQuadrilateral2&, int, int, int, int,
			   const Point2&, const Point2&, const Point2&,
			   const Point2&, const tEdgeKey&,
			   const tEdgeKey&, const tEdgeKey&,
			   const tEdgeKey&, int, GlueEdgeSet&,
			   const ConstrainedEdgeSet&
			   );

    // Creates a  new quadrilateral face as a  new connected component
    // of  the mesh and  keeps track  of its  three common  edges with
    // other faces  of the  mesh, so that  we can glue  this connected
    // component later.
    void makeQuadFaceGlue4(const tQuadrilateral2&, int, int, int, int,
			   const Point2&, const Point2&, const Point2&,
			   const Point2&, const tEdgeKey&, 
			   const tEdgeKey&, const tEdgeKey&,
			   const tEdgeKey&, GlueEdgeSet&,
			   const ConstrainedEdgeSet&
			   );

    
    // Checks  if  a given  quadrilateral  face  is geometrically  and
    // topologically consistent.
    bool checkFaceConsistency(const fcPointer&) const;

    // Checks  if   this  quadrilateral  mesh   is  geometrically  and
    // topologically consistent.
    bool checkMeshConsistency() const;

    // Eliminates  all quadrilaterals that  contains an  angle greater
    // than 170 degrees.
    void eliminateTShapedQuads();

    // Decomposes a  quadrilateral into new ones by  subdividing it or
    // by combining it with adjacent ones.
    void requadrangulate(const hePointer&, LTSFaces&);

    // Creates an hexagon from two quadrilaterals and then returns it.
    tPolygon2 makeRightHexagon(const hePointer&, const hePointer&);

    // Creates an hexagon from two quadrilaterals and then returns it.
    tPolygon2 makeLeftHexagon(const hePointer&, const hePointer&);

    // Subdivides a quadrilateral  into 5 new ones by  using 4 Steiner
    // points.
    void subdivideQuad(const hePointer&) throw (tExceptionObject);

    // Requadrangulates an hexagon by inserting one diagonal.
    void requadRight1(const hePointer&, const tPolygon2&) 
      throw (tExceptionObject);

    // Requadrangulates an hexagon using one Steiner point.
    void requadRight2(const hePointer&, const tPolygon2&) 
      throw (tExceptionObject);

    // Requadrangulates an hexagon using two Steiner points.
    void requadRight3(const hePointer&, const tPolygon2&)
      throw (tExceptionObject);

    // Requadrangulates an hexagon using two Steiner points.
    void requadRight4(const hePointer&, const tPolygon2&)
      throw (tExceptionObject);

    // Requadrangulates an  hexagon by  inserting one diagonal.  It is
    // the symmetric method to requadRight1().
    void requadLeft1(const hePointer&, const tPolygon2&)
      throw (tExceptionObject);

    // Requadrangulates an hexagon using  one Steiner point. It is the
    // symmetric method to requadRight2().
    void requadLeft2(const hePointer&, const tPolygon2&) 
      throw (tExceptionObject);

    // Requadrangulates an hexagon using two Steiner points. It is the
    // symmetric method to requadRight3().
    void requadLeft3(const hePointer&, const tPolygon2&)
      throw (tExceptionObject);

    // Requadrangulates an  hexagon using  two Steiner points.   It is
    // the symmetric method to requadRight4().
    void requadLeft4(const hePointer&, const tPolygon2&)
      throw (tExceptionObject);

    bool isAValidFace(const Point2&,const Point2&,
		      const Point2&,const Point2&);

    // Returns true if the new position  of a vertex causes one of its
    // incident face to be non-convex.
    bool isNotAValidMove(const vtPointer&);

    //
    // private attributes
    //

    // cosine of an angle of 10 degrees
    const double m_COSV;

    // counter of number of triangular-shaped quadrilaterals
    unsigned int m_tsq;
  };


};   // end namespace qmpp

#endif   // __quadmesh2_h
