//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    quadstat.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
// ===================================================================

#include "quadstat.h"
#include <cmath>
#include <iostream>
#include <fstream>

namespace qmpp {

  // -----------------------------------------------------------------
  // Method computeStats()
  //
  // Computes statistical information about a quadrilateral mesh.
  //
  // -----------------------------------------------------------------
  void
  tQuadStat::computeStats(const string& fn, const cpPointer& cp)
    throw (tExceptionObject)
  {
    // Gets number of faces of this complex
    unsigned int nfaces = cp->getNumFaces();

    // Creates vectors  to store  the area and  the quality  factor of
    // each quadrilateral in the mesh
    vector<double> va(nfaces);
    vector<double> vu(nfaces);

    // Loops  over  all  quadrilateral  and computes  their  area  and
    // quality factor.
    fcPointer f;
    hePointer he;
    double totarea = 0;
    double totu = 0;
    unsigned int i = 0;
    for (FIT fi = cp->fcBegin(); fi != cp->fcEnd(); ++fi) {
      f  = *fi;

      // Gets one half-edge of this quadrilateral
      he = f->getHalfEdge();

      // Gets the four point of the quadrilateral
      Point2 p1 = he->getVertex()->getPoint();
      Point2 p2 = he->getNext()->getVertex()->getPoint();
      Point2 p3 = he->getNext()->getNext()->getVertex()->getPoint();
      Point2 p4 = he->getPrev()->getVertex()->getPoint();

      // Creates a quadrilateral
      tQuadrilateral2 q;

      q.insert(0,1,p1);
      q.insert(1,2,p2);
      q.insert(2,3,p3);
      q.insert(3,4,p4);

      // Computes the area of the quadrilateral
      va[i] = q.area();

      // Computes the quality factor of this quadrilateral
      vu[i] = computeQualityFactor(q);

      // Adds up the area
      totarea += va[i];

      // Adds up quality factor
      totu += vu[i];

      // Computes minimum  and maximum  area, and minimum  and maximum
      // quality factor value.
      if (i == 0) {
	m_minarea = m_maxarea = va[i];
	m_minu = m_maxu = vu[i];
      }
      else {
	if (va[i] < m_minarea) m_minarea = va[i];
	if (va[i] > m_maxarea) m_maxarea = va[i];
	if (vu[i] < m_minu) m_minu = vu[i];
	if (vu[i] > m_maxu) m_maxu = vu[i];
      }

      ++i;
    }

    // Computes the average area
    m_avgarea = totarea / nfaces;

    // Computes the average quality factor
    m_avgu = totu / nfaces;

    // Computes standard deviation and  range of values of the quality
    // factor.
    double totsdvarea = 0;
    double totsdvu = 0;
    m_u1 = m_u2 = m_u3 = m_u4 = m_u5 = 0;
    for (i=0; i<nfaces; i++) {
      totsdvarea += ((va[i] - m_avgarea) * (va[i] - m_avgarea));
      totsdvu    += ((vu[i] -    m_avgu) * (vu[i] -    m_avgu));

      if (vu[i] < 5) ++m_u1;
      else if (vu[i] < 15) ++m_u2;
      else if (vu[i] < 25) ++m_u3;
      else if (vu[i] < 35) ++m_u4;
      else ++m_u5;
    }

    m_sdvarea = sqrt(totsdvarea / nfaces);
    m_sdvu    = sqrt(totsdvu    / nfaces);

    // Writes mesh statistics to a file
    filebuf fb;
    if (fb.open(fn.c_str(), ios::out) == 0) {
      throw new tExceptionObject("Unable to create output file");
    }

    std::ostream ostr(&fb);

    ostr << "Number of vertices  : " << cp->getNumVerts()  << std::endl;
    ostr << "Number of edges     : " << cp->getNumEdges()  << std::endl;
    ostr << "Number of faces     : " << cp->getNumFaces()  << std::endl;
    ostr << "Number of holes     : " << cp->getNumHoles()  << std::endl;
    ostr << "Number of boundaries: " << cp->getNumBounds() << std::endl;

    ostr << "minimum area: " << m_minarea << std::endl;
    ostr << "maximum area: " << m_maxarea << std::endl;
    ostr << "average area: " << m_avgarea << std::endl;
    ostr << "standard deviation: " << m_sdvarea << std::endl;

    ostr << "minimum angle quality factor: " << m_minu << std::endl;
    ostr << "maximum angle quality factor: " << m_maxu << std::endl;
    ostr << "average angle quality factor: " << m_avgu << std::endl;
    ostr << "standard deviation: " << m_sdvu << std::endl;

    ostr << "Number of elements with angle quality factor >=  0 and <   5: " << m_u1 << std::endl;
    ostr << "Number of elements with angle quality factor >=  5 and <  10: " << m_u2 << std::endl;
    ostr << "Number of elements with angle quality factor >= 10 and <  25: " << m_u3 << std::endl;
    ostr << "Number of elements with angle quality factor >= 25 and <  35: " << m_u4 << std::endl;
    ostr << "Number of elements with angle quality factor >= 35 and <= 45: " << m_u5 << std::endl;

    fb.close();
  }

  // -----------------------------------------------------------------
  // Method computeStats()
  //
  // Computes   the  value  of   the  quality   factor  for   a  given
  // quadrilateral. This value is a real number between 0 and 45.
  //
  // -----------------------------------------------------------------
  double
  tQuadStat::computeQualityFactor(const tQuadrilateral2& q) const
  {
    Point2 p1 = q.getPoint(0);
    Point2 p2 = q.getPoint(1);
    Point2 p3 = q.getPoint(2);
    Point2 p4 = q.getPoint(3);

    double a1 = computeMinAngle(p1,p2,p4);
    double a2 = computeMinAngle(p2,p3,p4);
    if (a1 > a2) a1 = a2;

    a2 = computeMinAngle(p1,p2,p3);
    if (a1 > a2) a1 = a2;

    a2 = computeMinAngle(p1,p3,p4);
    if (a1 > a2) a1 = a2;

    return a1;
  }

  // -----------------------------------------------------------------
  // Method computeMinAngle()
  //
  // Computes the minimum angle of a triangle.
  //
  // -----------------------------------------------------------------
  double
  tQuadStat::computeMinAngle(
			     const Point2& p1,
			     const Point2& p2,
			     const Point2& p3
			     )
    const
  {
    double a1 = computeAngle(p1,p2,p3);
    double a2 = computeAngle(p2,p3,p1);
    double a3 = computeAngle(p3,p1,p2);

    if ((a1 < a2) && (a1 < a3)) return a1;
    else if (a2 < a3) return a2;

    return a3;
  }

  // -----------------------------------------------------------------
  // Method computeAngle()
  //
  // Computes the angle defined by two adjacent line segments.
  //
  // -----------------------------------------------------------------
  double
  tQuadStat::computeAngle(
			  const Point2& p1,
			  const Point2& p2,
			  const Point2& p3
			  )
    const
  {
    double dx = p1.x() - p2.x();
    double dy = p1.y() - p2.y();
    double d1 = sqrt((dx * dx) + (dy * dy));
    double x1 = dx / d1;
    double y1 = dy / d1;

    dx = p3.x() - p2.x();
    dy = p3.y() - p2.y();
    double d2 = sqrt((dx * dx) + (dy * dy));
    double x2 = dx / d2;
    double y2 = dy / d2;

    double a = (acos((x1 * x2) + (y1 * y2)) * 180) / MY_PI;

    return a;
  }


};     // end namespace qmpp
