//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    quadstat.h
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#ifndef __quadstat_h
#define __quadstat_h

#include <string.h>

#include "typedefs.h"
#include "macros.h"
#include "smartptr.h"
#include "exceptionobj.h"

#include "quad2.h"
#include "hedge.h"
#include "stat.h"

namespace qmpp {

  // -----------------------------------------------------------------
  // class tQuadStat
  //
  // This  class  contains  methods   to  measure  the  quality  of  a
  // quadrilateral mesh.
  //
  // -----------------------------------------------------------------
  class tQuadStat : public tStat {
  public:
    // Default constructor
    tQuadStat() : 
      tStat(),
      m_minarea(0),
      m_maxarea(0),
      m_avgarea(0),
      m_sdvarea(0),
      m_minu(0),
      m_maxu(0),
      m_avgu(0),
      m_sdvu(0),
      m_u1(0),
      m_u2(0),
      m_u3(0),
      m_u4(0),
      m_u5(0)
      {}

    // Destructor
    ~tQuadStat() {}

    // Computes statistical information about a quadrilateral mesh
    virtual void computeStats(const string&, const cpPointer&) 
      throw (tExceptionObject);

  private:
    // Computes  the   value  of  the  quality  factor   for  a  given
    // quadrilateral. This value is a real number between 0 and 45.
    double computeQualityFactor(const tQuadrilateral2&) const;

    // Computes the minimum angle of a triangle.
    double computeMinAngle(const Point2&, const Point2&, 
			   const Point2&) const;

    // Computes the angle defined by two adjacent line segments.
    double computeAngle(const Point2&, const Point2&,
			const Point2&) const;

    //
    // private attributes
    //

    // Minimum area of a quadrilateral in the mesh
    double m_minarea;

    // Maximum area of a quadrilateral in the mesh
    double m_maxarea;

    // Average area of the quadrilaterals in the mesh
    double m_avgarea;

    // // Standard deviation of quadrilateral area
    double m_sdvarea;

    // Minimum value  of the quality factor "u"  (the minimum possible
    // value is 0).
    double m_minu;

    // Maximum value  of the quality factor "u"  (the maximum possible
    // value is 45)
    double m_maxu;

    // Average value of the quality factor "u"
    double m_avgu;

    // Standard deviation of the quality factor "u"
    double m_sdvu;

    // Number of quadrilaterals with quality factor >= 0 and < 5
    unsigned int m_u1;

    // Number of quadrilaterals with quality factor >= 5 and < 15
    unsigned int m_u2;

    // Number of quadrilaterals with quality factor >= 15 and < 25
    unsigned int m_u3;

    // Number of quadrilaterals with quality factor >= 25 and < 35
    unsigned int m_u4;

    // Number of quadrilaterals with quality factor >= 35 and <= 45
    unsigned int m_u5;
  };


};   // end namespace qmpp


#endif   // __quadstat_h
