//  ===================================================================
//
//  Program:   Quadrilateral Mesh Post-Processor
//  Module:    trimesh2.cc
//  Language:  C++
//  Date:      March 2003
//  Version:   1.0.0
//
//  Copyright (c) 2002, 2003
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
// ===================================================================

#include "trimesh2.h"

namespace qmpp {

  // -----------------------------------------------------------------
  // Method buildTriMesh()
  //
  // Stores a triangular mesh into the data structure handle edge. All
  // it needs is the collection of  triangles of the mesh in any given
  // order. Triangles should consistently be oriented.
  //
  // -----------------------------------------------------------------
  tTriMesh::cpPointer
  tTriMesh::buildMesh(const std::string& name)
  {
    // Reads  files   with  information  about   vertices,  edges  and
    // triangles of the triangular mesh.
    std::string infile1 = std::string(name) + std::string(".node");
    std::string infile2 = std::string(name) + std::string(".edge");
    std::string infile3 = std::string(name) + std::string(".ele");

    // Creates file reader
    tMeshReader reader(infile1, infile2, infile3);

    // Reads files
    NodeSet lv;
    ConstrainedEdgeSet le;
    FaceSet lf;

    try {
      reader.readVertexFile(lv);
      reader.readEdgeFile(le);
      reader.readElemFile(lv,lf);
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }

    // We do not need the set  of nodes anymore as node information is
    // also in the face set "lf".
    lv.clear();

    // First stage: making faces
    GlueEdgeSet gluelist;
    typedef FaceSet::const_iterator LFI;

    for (LFI fiter = lf.begin(); fiter != lf.end(); ++fiter) {
      // Gets face from iterator
      tTriangle2 f(*fiter);

      // Checks out orientation
      if (!f.isCCW()) {
	f.reverse();
      }

      // Creates vertex codes for every vertex of face f
      int cv1 = f.getVertexId(0);
      int cv2 = f.getVertexId(1);
      int cv3 = f.getVertexId(2);

      // Gets vertex locations
      Point2 lv1 = f.getPoint(0);
      Point2 lv2 = f.getPoint(1);
      Point2 lv3 = f.getPoint(2);

      // Creates edge codes for every edge of face f
      tEdgeKey ce1 = tEdgeKey(cv1,cv2);
      tEdgeKey ce2 = tEdgeKey(cv2,cv3);
      tEdgeKey ce3 = tEdgeKey(cv3,cv1);

      // Searches for existing edges in the boundary edge set
      int foundedge = 0;
      if (es.find(ce1) != es.end()) foundedge += 1;
      if (es.find(ce2) != es.end()) foundedge += 2;
      if (es.find(ce3) != es.end()) foundedge += 4;

      // A new  face is created only if  one of its edges  is found in
      // the set of constrained edges  and if the attaching of the new
      // face does not create a singular vertex.
      if ((foundedge == 1) || (foundedge == 2) || (foundedge == 4)) {
	int cvaux;
	Point2 lvaux;
	tEdgeKey ceaux1, ceaux2, ceaux3;

	if (foundedge == 1) {
	  cvaux = cv3;
	  lvaux = lv3;
	  ceaux1 = ce2;
	  ceaux2 = ce3; 
	  ceaux3 = ce1;
	}
	else if (foundedge == 2) {
	  cvaux = cv1;
	  lvaux = lv1;
	  ceaux1 = ce3;
	  ceaux2 = ce1;
	  ceaux3 = ce2;
	}
	else {  // foundedge == 4
	  cvaux = cv2;
	  lvaux = lv2;
	  ceaux1 = ce1;
	  ceaux2 = ce2;
	  ceaux3 = ce3;
	}

	// If  vertex "cvaux"  is not  in  the same  boundary as  edge
	// "ceaux3.reverse()", we can attach  face "f" to the boundary
	// containing "ceaux3.reverse()".
	if (!findVertexInBound(cvaux,ceaux3)) {
	  // Either "cvaux" has  not been created yet or  it is not on
	  // the boundary of "ceaux3.reverse()".

	  this->makeTriFace(f,cvaux,lvaux,ceaux3,ceaux1,ceaux2,le);
	}
	else {
	  // Creates an isolated face, which should be attached to "f"
	  // in the second stage.

	  this->makeTriFaceGlue1(f,cv1,cv2,cv3,lv1,lv2,lv3,ceaux3,
				 ceaux1,ceaux2,gluelist,le);
	}
      }
      else if (foundedge != 0) {   // foundedge == 3, 5, 6 or 7

	// if new  face has two  common edges with existent  faces, we
	// must attach  it to one of  the faces. In  the second stage,
	// the new face is glued to the other one.
	if (foundedge == 3) {
	  // edges "ce1" and "ce2" are common edges
	  this->makeTriFaceGlue2(f,cv1,cv2,cv3,lv1,lv2,lv3,ce1,ce2,ce3,
				 3,gluelist,le);
	}
	else if (foundedge == 5) {
	  // edges "ce1" and "ce3" are common edges
	  this->makeTriFaceGlue2(f,cv1,cv2,cv3,lv1,lv2,lv3,ce1,ce2,ce3,
				 2,gluelist,le);
	}
	else if (foundedge == 6) {
	  // edges "ce2" and "ce3" are common edges
	  this->makeTriFaceGlue2(f,cv1,cv2,cv3,lv1,lv2,lv3,ce1,ce2,ce3,
				 1,gluelist,le);
	}
	else {
	  // all edges in common
	  this->makeTriFaceGlue3(f,cv1,cv2,cv3,lv1,lv2,lv3,ce1,ce2,ce3,
				 gluelist,le);
	}
      }
      else { // foundedge == 0
	// new face  has no  vertex in common  with existent  faces or
	// one, two or three vertices in common with existent faces.

	// we will create isolated faces in any case.

	this->makeTriFace(f,cv1,cv2,cv3,lv1,lv2,lv3,ce1,ce2,ce3,le);
      }
    }

    // Removes all elements from vertex and edge boundary sets
    vs.clear();
    es.clear();

    // Removes all elements from constrained edge set

    // Second stage: glueing faces
    typedef GlueEdgeSet::iterator LGI;
    LGI giter = gluelist.begin();
    while (giter != gluelist.end()) {
      // Gets the half-edges to be identified in the faces to be glued.
      hePointer he1 = (*giter).m_he1;
      hePointer he2 = (*giter).m_he2;

      // Glues faces
      glueFaces(he1,he2);

      // Removes the half-edges from the glue list
      giter = gluelist.erase(giter);
    }

    // Returns complex
    return m_cp;
  }

  // -----------------------------------------------------------------
  // Method makeTriFace()
  //
  // Creates a new triangular face as a new connected component of the
  // mesh.
  //
  // -----------------------------------------------------------------
  void
  tTriMesh::makeTriFace(
			const tTriangle2& f,
			int cv1,
			int cv2,
			int cv3,
			const Point2& p1,
			const Point2& p2,
			const Point2& p3,
			const tEdgeKey& ce1,
			const tEdgeKey& ce2,
			const tEdgeKey& ce3,
			const ConstrainedEdgeSet& ledges
			)
  {
    // Creates all three vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();

    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));

    // Inserts new edge codes into boundary edge set
    es.insert(make_pair(ce1.reverse(),he1));
    es.insert(make_pair(ce2.reverse(),he2));
    es.insert(make_pair(ce3.reverse(),he3));

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) {
      he3->getEdge()->constraintOn();
    }
  }

  // -----------------------------------------------------------------
  // Method makeTriFace()
  //
  // Creates a new triangular face  and attaches it to another face in
  // the mesh.
  //
  // -----------------------------------------------------------------
  void
  tTriMesh::makeTriFace(
			const tTriangle2& f,
			int cv,
			const Point2& p,
			const tEdgeKey& ce,
			const tEdgeKey& ce2,
			const tEdgeKey& ce3,
			const ConstrainedEdgeSet& le
			)
  {
    // Creates all three vertices
    vtPointer v = m_cp->createVertex(p);

    // Gets first half-edge of the commom edge
    hePointer he = (es.find(ce))->second;

    // Makes a new face with a commom edge and a new edge
    egPointer e = ((m_cp->mkFE(he))->getHalfEdge())->getEdge();

    // Gets first half-edge of the new edge
    hePointer he2 = e->getHalfEdge();

    // Creates another edge and get its first half-edge
    hePointer he3 = (m_cp->mkEV(v,he2))->getHalfEdge();

    // Inserts new vertice code into boundary vertex set
    vs.insert(make_pair(cv,v));

    // Inserts new edge codes into boundary edge set
    es.insert(make_pair(ce2.reverse(),he2));
    es.insert(make_pair(ce3.reverse(),he3));

    // Removes commom edge from the boundary edge set
    es.erase(es.find(ce));

    // Sets status of constrained edges
    if (isConstrainedEdge(ce2,le)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,le)) he3->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeTriFaceGlue1()
  //
  // Creates a new triangular face as a new connected component of the
  // mesh and keeps track of its  common edge with another face of the
  // mesh, so that we can glue this connected component later.
  //
  // -----------------------------------------------------------------
  void
  tTriMesh:: makeTriFaceGlue1(
			      const tTriangle2& f,
			      int cv1,
			      int cv2,
			      int cv3,
			      const Point2& p1,
			      const Point2& p2,
			      const Point2& p3,
			      const tEdgeKey& ce,
			      const tEdgeKey& ce2,
			      const tEdgeKey& ce3,
			      GlueEdgeSet& gset,
			      const ConstrainedEdgeSet& ledges
			      )
  {
    // Creates all three vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();

    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));

    // Inserts new edge codes into boundary edge set
    es.insert(make_pair(ce2.reverse(),he2));
    es.insert(make_pair(ce3.reverse(),he3));

    // keeps track of edges to be glued at second stage
    EdgeSet::iterator hepos = es.find(ce);
    insertGlue(hepos->second,he1,gset);
  
    // Removes commom edge from the boundary edge set
    es.erase(hepos);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce,ledges))  he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
  }

  // -----------------------------------------------------------------
  // Method makeTriFaceGlue2()
  //
  // Creates a new triangular face as a new connected component of the
  // mesh and keeps track of its  two common edges with other faces of
  // the mesh, so that we can glue this connected component later.
  //
  // -----------------------------------------------------------------
  void
  tTriMesh::makeTriFaceGlue2(
			     const tTriangle2& f,
			     int cv1,
			     int cv2,
			     int cv3,
			     const Point2& p1,
			     const Point2& p2,
			     const Point2& p3,
			     const tEdgeKey& ce1,
			     const tEdgeKey& ce2,
			     const tEdgeKey& ce3,
			     int ne,
			     GlueEdgeSet& gset,
			     const ConstrainedEdgeSet& ledges
			     )
  {
    // Creates all three vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);

    // Makes the face and keep track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();

    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));

    // Inserts new edge code into boundary edge set and keeps track of
    // the two edges to be glued in the second stage.
    EdgeSet::iterator hepos1, hepos2;
    if (ne == 1) {
      hepos1 = es.find(ce2);
      hepos2 = es.find(ce3);
      es.insert(make_pair(ce1.reverse(),he1));
      insertGlue(hepos1->second,he2,gset);
      insertGlue(hepos2->second,he3,gset);
    }
    else if (ne == 2) {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce3);
      es.insert(make_pair(ce2.reverse(),he2));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he3,gset);
    }
    else {
      hepos1 = es.find(ce1);
      hepos2 = es.find(ce2);
      es.insert(make_pair(ce3.reverse(),he3));
      insertGlue(hepos1->second,he1,gset);
      insertGlue(hepos2->second,he2,gset);
    }

    // Removes commom edges from the boundary edge set
    es.erase(hepos1);
    es.erase(hepos2);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
  }


  // -----------------------------------------------------------------
  // Method makeTriFaceGlue3()
  //
  // Creates a new triangular face as a new connected component of the
  // mesh and keeps  track of its three common  edges with other faces
  // of the mesh, so that we can glue this connected component later.
  //
  // -----------------------------------------------------------------
  void
  tTriMesh::makeTriFaceGlue3(
			     const tTriangle2& f,
			     int cv1,
			     int cv2,
			     int cv3,
			     const Point2& p1,
			     const Point2& p2,
			     const Point2& p3,
			     const tEdgeKey& ce1,
			     const tEdgeKey& ce2,
			     const tEdgeKey& ce3,
			     GlueEdgeSet& gset,
			     const ConstrainedEdgeSet& ledges
			     )
  {
    // Creates all three vertices
    vtPointer v1 = m_cp->createVertex(p1);
    vtPointer v2 = m_cp->createVertex(p2);
    vtPointer v3 = m_cp->createVertex(p3);

    // Makes the face and keeps track of its edges
    fcPointer newf = m_cp->mkFBEV(v1);
    hePointer he1 = newf->getHalfEdge();
    hePointer he2 = (m_cp->mkEV(v2,he1))->getHalfEdge();
    hePointer he3 = (m_cp->mkEV(v3,he2))->getHalfEdge();

    // Inserts new vertice codes into boundary vertex set
    vs.insert(make_pair(cv1,v1));
    vs.insert(make_pair(cv2,v2));
    vs.insert(make_pair(cv3,v3));

    // Inserts new edge code into boundary edge set
    EdgeSet::iterator hepos1, hepos2, hepos3;
 
    // Keeps track of edges to be glued in the second stage
    hepos1 = es.find(ce1);
    hepos2 = es.find(ce2);
    hepos3 = es.find(ce3);

    insertGlue(hepos1->second,he1,gset);
    insertGlue(hepos2->second,he2,gset);
    insertGlue(hepos3->second,he3,gset);

    // Removes commom edges from the boundary edge set
    es.erase(hepos1);
    es.erase(hepos2);
    es.erase(hepos3);

    // Sets status of constrained edges
    if (isConstrainedEdge(ce1,ledges)) he1->getEdge()->constraintOn();
    if (isConstrainedEdge(ce2,ledges)) he2->getEdge()->constraintOn();
    if (isConstrainedEdge(ce3,ledges)) he3->getEdge()->constraintOn();
  }

};     // end namespace qmpp
